
// ---------------------------------------------------------------------------------

void AnimEngine() {
  // an animation engine which runs off scripts loaded into memory from disc or
  // downloaded from the MATRIX app
  // check for delay caused by a Pause command
  if (!AnimPLAY) {return;}
  if (AnimWait > 0) {if ((millis() - PauseTimer) < AnimWait) {return;} else {AnimWait = 0;}}

  // if not in Verbose mode then only send pointer values, provided we are not
  // running under the scheduler
  if (RepON) {
    if ((!Verbose) && (SchCnt < 1)) {Serial.println("##" + String(AnimPnt) + ".");}
  }
  AnimErr = 0; AnimInc = 1; AnimIf = true;
  String zTxt = "";
  String zS$ = "";
  int32_t zF0;  // 1st frame pointer
  int32_t zF1;  // 2nd frame pointer
  int zH; int zW; bool zNew = false;
  int zX0; int zX1;int zX2;int zY0; int zY1;int zY2;
  
//  Serial.println("AnimPnt "+ String(AnimPnt));
  switch (AnCmd[AnimPnt]) {
    case 0x0000: // EoF so stop the Engine in its tracks, if not scheduled
      if (SchCnt > 0) {
        SchPnt++; if (SchPnt >= SchCnt) {SchPnt = 0;}
        ClearAnimArray();
        readSDAnim(getNum4Digits(AnimSch[SchPnt]));     // loads /Anim/AnXXXX.txt
        if (ErrFlg) {AnimErr = 12;}                     // Anim file read failure
        else {setAnimMode(1); AnimPLAY = true; return;} // restart the animation engine
      } else {
        Serial.println("EoF encountered. STOP!");
        AnimSTOP = true;
      } break;

    case 0x0001: break; // blank line ---- or title, so ignore

    case 0x0002: break; // comment line, so ignore

    case 0x4150: // AnimSchAdd - add an AnXXXX.txt file to the Anim scheduler
      if (SchCnt < SchDepth) {
        AnimSch[SchCnt] = AnD0[AnimPnt]; SchCnt++; SchPnt = 0;
      } else {AnimErr = 11;}  // array depth exceeded
      if (Verbose) {Serial.println("AnimSchAdd An" + getNum4Digits(AnD0[AnimPnt]));}
      break;

    case 0x4151: // AnimSchESC - load next Anim file in scheduler list, if pre-loaded
      // if no schedule has been set then this command is ignored as the
      // animation is assumed to be running stand-alone
      if (Verbose) {Serial.println(F("AnimSchESC - exit scheduled animation"));}
      if (SchCnt > 0) {
        SchPnt++; if (SchPnt >= SchCnt) {SchPnt = 0;}
        ClearAnimArray();
        readSDAnim(getNum4Digits(AnimSch[SchPnt]));     // loads /Anim/AnXXXX.txt
        if (ErrFlg) {AnimErr = 12;}                     // Anim file read failure
        else {setAnimMode(1); AnimPLAY = true; return;} // restart the animation engine
      } break;

    case 0x4152: // AnimSchRun - invoke the Anim scheduler, starting with XXXX
      if (Verbose) {Serial.println("AnimSchRun - invoke the Anim scheduler");}
      if (SchCnt > 0) {
        SchPnt = AnD0[AnimPnt]; // get pointer to 1st AnXXXX file
        if (SchPnt >= SchCnt) {SchPnt = 0;}
        setAnimMode(1);
        ClearAnimArray();
        readSDAnim(getNum4Digits(AnimSch[SchPnt]));     // loads /Anim/AnXXXX.txt
        if (ErrFlg) {AnimErr = 12;}                     // Anim file read failure
        else {AnimPLAY = true; return;} // restart the animation engine
      } else {AnimErr = 13;}
      break;

    case 0x4153: // AnimSchClr - clear Anim scheduler and reset it
      ClearAnimSchedule();
      if (Verbose) {Serial.println("AnimSchClr - Anim scheduler is reset");}
      break;
      
    case 0x4201: // BckgndDim - dim the background by a dim factor
      // dim the current foreground frames pixels by a factor
      Gfx_DimBckgnd(1,PixWxPixH,AnD0[AnimPnt]);
      if (Verbose) {Serial.println("BckgndDim " + String(AnD0[AnimPnt]));}
      break;
      
    case 0x4206: // BckgndDimE2 - dim all background buffer pixels by a half
      // dim the current foreground frames pixels by a factor
      Gfx_DimBckgndE2(1,PixWxPixH);
      if (Verbose) {Serial.println("BckgndDimE2");}
      break;

    case 0x4202: // BckgndDimTo - dim the background towards a target RGB
      // dim the backgound frames pixels by a factor towards a target RGB value
      getRGBcols(AnD0[AnimPnt]);
      Gfx_DimBckgndTo(1,PixWxPixH,AnD1[AnimPnt],Pen_Red,Pen_Grn,Pen_Blu);
      if (Verbose) {Serial.println("BckgndDimTo [RGB] " + String(AnD1[AnimPnt]));}
      break;

    case 0x4203: // BckgndLoad - load a file into the backgropund
      BckGndFrame = AnD0[AnimPnt];
      Folder = "/Frames"; Name = "/Fr" + getNum4Digits(BckGndFrame);
//      t0 = micros();
      readSDToLED(2);
//      t1 = micros();
      if (Verbose) {Serial.println("BckgndLoad Fr" + getNum4Digits(AnD0[AnimPnt]));}
//      Serial.println("t1-t0 " + String(t1-t0));
      break;

    case 0x4204: // BckgndNext - load the next background file in the sequence
      // loads the next background in the sequence then cycles round
      zF0 = AnD0[AnimPnt];  // get the 1st foreground pointer
      zF1 = AnD1[AnimPnt];  // get the 2nd foreground pointer
      if (BckGndFrame < zF0) {BckGndFrame = zF0;} // not in current range so set to 1st
      else if (BckGndFrame >= zF1) {BckGndFrame = zF0;} // beyond current range, so reset
      else {BckGndFrame++;}                           // increment pointer
      Folder = "/Frames"; Name = "/Fr" + getNum4Digits(BckGndFrame);
      readSDToLED(2); // load the file into the background
      if (Verbose) {Serial.println("BckgndNext Fr" + getNum4Digits(BckGndFrame));}
      break;

    case 0x4205: // BckgndRot - Up/Right/Down/Left
      // rotation depends on the associated direction value
      switch (AnD0[AnimPnt]) {
        case 0: Gfx_RotUpBck(); zTxt = "Up"; break;     // rotate array Up
        case 1: Gfx_RotRhtBck(); zTxt = "Right"; break; // rotate array Right
        case 2: Gfx_RotDwnBck(); zTxt = "Down"; break;  // rotate array Down
        case 3: Gfx_RotLftBck(); zTxt = "Left"; break;  // rotate array Left
      }
      if (Verbose) {Serial.println("BckgndRot " + zTxt);}
      break;

    case 0x4213: // BckgndWinFlip - rotate XY,XY region in one of 2 directions
      // extract window co-ordinates
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      // rotation depends on the associated direction value
      switch (AnD2[AnimPnt]) {
        case 0: Gfx_FlipBckgndH(zX0,zY0,zX1,zY1); zTxt = "Left-Right"; break;  // flip array window left-right
        case 1: Gfx_FlipBckgndV(zX0,zY0,zX1,zY1); zTxt = "Up-Down"; break;     // flip array window up-down
      }
      if (Verbose) {Serial.println("BckgndWinFlip " + zTxt);}
      break;

    case 0x4215: // BckgndWinRot - rotate a region within the frame in one of 4 directions
      // extract window co-ordinates
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      // rotation depends on the associated direction value
      switch (AnD2[AnimPnt]) {
        case 0: Gfx_RotUpWinBck(zX0,zY0,zX1,zY1); zTxt = "Up"; break;  // rotate array window Up
        case 1: Gfx_RotRhtWinBck(zX0,zY0,zX1,zY1); zTxt = "Right"; break; // rotate array window Right
        case 2: Gfx_RotDwnWinBck(zX0,zY0,zX1,zY1); zTxt = "Down"; break; // rotate array window Down
        case 3: Gfx_RotLftWinBck(zX0,zY0,zX1,zY1); zTxt = "Left"; break; // rotate array window Left
      }
      if (Verbose) {Serial.println("BckgndWinRot " + zTxt);}
      break;

    case 0x42B0: // Bright - set the brightness 0 - 255
      if (Verbose) {Serial.println("Bright " + String(AnD0[AnimPnt]));}
      Bright = AnD0[AnimPnt]; FastLED.setBrightness(Bright);
      BrightNew = Bright; LevelOld = 0;
      break;

    case 0x42B8: // Bright=Var - set the brightness = variable 'A' - 'Z'
      zX0 = AnD0[AnimPnt]; Bright = constrain(Var[zX0],0,255);
      if (Verbose) {Serial.println("Bright=Var '" + String(char(zX0 + 65)) + "' = " + String(Bright));}
      FastLED.setBrightness(Bright); BrightNew = Bright; LevelOld = 0;
      break;

    case 0x4376:  // ClrBckgnd - clear the background frame store
      if (Verbose) {Serial.println("ClrBckgnd");}
      Gfx_FillB(1,PixWxPixH,0,0,0); break;

    case 0x4380:  // ClrBckgndXY - clear the background frame store region X,Y-X,Y
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 = (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 = (AnD1[AnimPnt] & 0xFF) - 127;
      Gfx_FillBReg(zX0,zY0,zX1,zY1,0,0,0); 
      if (Verbose) {Serial.println("ClrBckgndXY " + String(zX0) + "," + String(zY0) + " - " + String(zX1) + "," + String(zY1));}
      break;

    case 0x4377:  // ClrForgnd - clear the foreground frame store
      if (Verbose) {Serial.println("ClrForgnd");}
      Gfx_FillF(1,PixWxPixH,0,0,0); break;

    case 0x4382:  // ClrForgndXY - clear the foreground frame store region X,Y-X,Y
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 = (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 = (AnD1[AnimPnt] & 0xFF) - 127;
      Gfx_FillFReg(zX0,zY0,zX1,zY1,0,0,0); 
      if (Verbose) {Serial.println("ClrForgndXY " + String(zX0) + "," + String(zY0) + " - " + String(zX1) + "," + String(zY1));}
      break;

    case 0x4378:  // ClrFrame - clear the frame store
      if (Verbose) {Serial.println("ClrFrame");}
      Gfx_Fill(1,PixWxPixH,0,0,0); break;

    case 0x4384:  // ClrFrameXY - clear the frame store region X,Y-X,Y
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 = (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 = (AnD1[AnimPnt] & 0xFF) - 127;
      Gfx_FillReg(zX0,zY0,zX1,zY1,0,0,0); 
      if (Verbose) {Serial.println("ClrFrameXY " + String(zX0) + "," + String(zY0) + " - " + String(zX1) + "," + String(zY1));}
      break;
      
    case 0x43A0: // ClrSprite{M} - clear Sprites {M} image store
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask {00000000}
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {zTxt = "1" + zTxt; Gfx_FillSp(zX1,1,256,0,0,0);}
        else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("ClrSprite{M} {" + zTxt + "}");}
      break;

    case 0x4420: // DecVar - decrement variable 'A' - 'Z' by -1
      zX0 = AnD0[AnimPnt]; Var[zX0]--;
      if (Verbose) {Serial.println("DecVar '" + String(char(zX0 + 65)) + "'   " + String(Var[zX0]));}
      break;

    case 0x4450: // Delay - delay the 'Show' by so many frames
      if (Verbose) {Serial.println("Delay " + String(AnD0[AnimPnt]));}
      AnimDel = AnD0[AnimPnt]; break;

    case 0x4452: // Delay=[V] - delay the 'Show' by a variable of frames cycles
      AnimDel = Var[AnD0[AnimPnt]]; if (AnimDel < 1) {AnimDel = 1;}
      if (Verbose) {Serial.println("Delay=[V] " + String(AnimDel));}
      break;

    case 0x4451: // DelayRnd - delay the 'Sow' by a random range of frames
      AnimDel = random((AnD0[AnimPnt] & 0xFF00)>>8,(AnD0[AnimPnt] & 0xFF) + 1);
      if (Verbose) {Serial.println("DelayRnd " + String(AnimDel));}
      break;

    case 0x44B4: // DrawBar[V,V] - draw a filled box using variables and pen colour
      zX0 = (AnD0[AnimPnt] & 0xFF00)>>8; zY0 = AnD0[AnimPnt] & 0xFF;
      zW  = (AnD1[AnimPnt] & 0xFF00)>>8;  zH = AnD1[AnimPnt] & 0xFF;
      Gfx_DrawBar(Var[zX0],Var[zY0],Var[zW],Var[zH]);
      if (Verbose) {Serial.println("DrawBar[V,V] X=" + String(Var[zX0]) + " Y=" + String(Var[zY0]) + " W=" + String(Var[zW]) + " H=" + String(Var[zH]));}
      break;

    case 0x44B0: // DrawBarXYWH - draw a filled box using Gfx and pen colour
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 = (AnD0[AnimPnt] & 0xFF) - 127;
      zW = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zH = (AnD1[AnimPnt] & 0xFF) - 127;
      Gfx_DrawBar(zX0,zY0,zW,zH);
      if (Verbose) {Serial.println("DrawBarXYWH X=" + String(zX0) + " Y=" + String(zY0) + " W=" + String(zW) + " H=" + String(zH));}
      break;

    case 0x44B1: // DrwBarRndXY - draw a box at random X,y with W,H
      zX0 = random(((AnD0[AnimPnt] & 0xFF00)>>8) - 127, (AnD0[AnimPnt] & 0xFF) - 126);
      zY0 = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 127, (AnD1[AnimPnt] & 0xFF) - 126);
      zW = ((AnD2[AnimPnt] & 0xFF00)>>8) - 127;
      zH = (AnD2[AnimPnt] & 0xFF) - 127;
      Gfx_DrawBar(zX0,zY0,zW,zH);
      if (Verbose) {Serial.println("DrawBarRndXY X=" + String(zX0) + " Y=" + String(zY0) + " W=" + String(zW) + " H=" + String(zH));}
      break;

    case 0x44B5: // DrawBox[V,V] - draw a box using variables and pen colour
      zX0 = (AnD0[AnimPnt] & 0xFF00)>>8; zY0 = AnD0[AnimPnt] & 0xFF;
      zW  = (AnD1[AnimPnt] & 0xFF00)>>8;  zH = AnD1[AnimPnt] & 0xFF;
      Gfx_DrawBox(Var[zX0],Var[zY0],Var[zW],Var[zH]);
      if (Verbose) {Serial.println("DrawBox[V,V] X=" + String(Var[zX0]) + " Y=" + String(Var[zY0]) + " W=" + String(Var[zW]) + " H=" + String(Var[zH]));}
      break;

    case 0x44B2: // DrawBoxXYWH - draw a box using Gfx and pen colour
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 = (AnD0[AnimPnt] & 0xFF) - 127;
      zW = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zH = (AnD1[AnimPnt] & 0xFF) - 127;
      Gfx_DrawBox(zX0,zY0,zW,zH);
      if (Verbose) {Serial.println("DrawBoxXYWH X=" + String(zX0) + " Y=" + String(zY0) + " W=" + String(zW) + " H=" + String(zH));}
      break;

    case 0x44B3: // DrwBoxRndXY - draw a box at random X,y with W,H
      zX0 = random(((AnD0[AnimPnt] & 0xFF00)>>8) - 127, (AnD0[AnimPnt] & 0xFF) - 126);
      zY0 = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 127, (AnD1[AnimPnt] & 0xFF) - 126);
      zW = ((AnD2[AnimPnt] & 0xFF00)>>8) - 127;
      zH = (AnD2[AnimPnt] & 0xFF) - 127;
      Gfx_DrawBox(zX0,zY0,zW,zH);
      if (Verbose) {Serial.println("DrawBoxRndXY X=" + String(zX0) + " Y=" + String(zY0) + " W=" + String(zW) + " H=" + String(zH));}
      break;

    case 0x44C0: // DrawLineA-B - draw a line from X0,Y0 to X1,Y1
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 = (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 = (AnD1[AnimPnt] & 0xFF) - 127;
      Gfx_Line(zX0,zY0,zX1,zY1);
      if (Verbose) {Serial.println("DrawLineA-B X0=" + String(zX0) + " Y0=" + String(zY0) + " X1=" + String(zX1) + " Y1=" + String(zY1));}
      break;

    case 0x44C1: // DrawLineRnd - draw a random line within X0,Y0 to X1,Y1
      zX0 = random(((AnD0[AnimPnt] & 0xFF00)>>8) - 127, (AnD0[AnimPnt] & 0xFF) - 126);
      zY0 = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 127, (AnD1[AnimPnt] & 0xFF) - 126);
      zX1 = random(((AnD0[AnimPnt] & 0xFF00)>>8) - 127, (AnD0[AnimPnt] & 0xFF) - 126);
      zY1 = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 127, (AnD1[AnimPnt] & 0xFF) - 126);
      Gfx_Line(zX0,zY0,zX1,zY1);
      if (Verbose) {Serial.println("DrawLineRnd X0=" + String(zX0) + " Y0=" + String(zY0) + " X1=" + String(zX1) + " Y1=" + String(zY1));}
      break;

    case 0x44C3: // DrawLine[V,V] - draw a line with variables from A,B to X,Y
      zX0 = (AnD0[AnimPnt] & 0xFF00)>>8; zY0 = AnD0[AnimPnt] & 0xFF;
      zW  = (AnD1[AnimPnt] & 0xFF00)>>8;  zH = AnD1[AnimPnt] & 0xFF;
      Gfx_Line(Var[zX0],Var[zY0],Var[zW],Var[zH]);
      if (Verbose) {Serial.println("DrawLine[V,V] X=" + String(Var[zX0]) + " Y=" + String(Var[zY0]) + " W=" + String(Var[zW]) + " H=" + String(Var[zH]));}
      break;

    case 0x4586:  // EndIf - terminates 'If..' block
      if (Verbose) {Serial.println("EndIf");}
      break;

    case 0x4589:  // EndLoop - branches to the Loop line
      if (Verbose) {Serial.println("EndLoop");}
      if (AnimLoop < 0) {AnimMode = -1; AnimErr = 1; break;} // ERROR!
      AnimPnt = AnimLoop; break;

    case 0x4658:  // FillBckgndRGB - fill the background with colour
      if (Verbose) {Serial.println("FillBckgndRGB");}
      getRGBcols(AnD0[AnimPnt]); Gfx_FillB(1,PixWxPixH,Pen_Red,Pen_Grn,Pen_Blu); break;

    case 0x4659:  // FillForgndRGB - fill the foreground with colour
      if (Verbose) {Serial.println("FillForgndRGB");}
      getRGBcols(AnD0[AnimPnt]); Gfx_FillF(1,PixWxPixH,Pen_Red,Pen_Grn,Pen_Blu); break;

    case 0x465A:  // FillFrameRGB - fill the frame with colour
      if (Verbose) {Serial.println("FillFrameRGB");}
      getRGBcols(AnD0[AnimPnt]); Gfx_Fill(1,PixWxPixH,Pen_Red,Pen_Grn,Pen_Blu); break;

    case 0x4693: // ForgndDim - dim the foreground by a dim factor
      // dim the current foreground frames pixels by a factor
      Gfx_DimForgnd(1,PixWxPixH,AnD0[AnimPnt]);
      if (Verbose) {Serial.println("ForgndDim " + String(AnD0[AnimPnt]));}
      break;
      
    case 0x469C: // ForgndDimE2 - dim all foreground buffer pixels by a half
      // dim the current foreground frames pixels by a factor
      Gfx_DimForgndE2(1,PixWxPixH);
      if (Verbose) {Serial.println("ForgndDimE2");}
      break;

    case 0x4694: // ForgndDimTo - dim the foregound towards an RGB value by a factor
      // dim the foregound frames pixels by a factor towards a target RGB value
      getRGBcols(AnD0[AnimPnt]);
      Gfx_DimForgndTo(1,PixWxPixH,AnD1[AnimPnt],Pen_Red,Pen_Grn,Pen_Blu);
      if (Verbose) {Serial.println("ForgndDimTo [RGB] " + String(AnD1[AnimPnt]));}
      break;

    case 0x4695:  // ForgndLoad - load a file into the foreground
      ForGndFrame = AnD0[AnimPnt];
      Folder = "/Frames"; Name = "/Fr" + getNum4Digits(ForGndFrame);
//      t0 = micros();
      readSDToLED(1);
//      t1 = micros();
      if (Verbose) {Serial.println("ForgndLoad Fr" + String(AnD0[AnimPnt]));}
      break;

    case 0x4696: // ForgndNext - load the next foreground file in the sequence
      // loads the next foreground in the sequence then cycles round
      zF0 = AnD0[AnimPnt];  // get the 1st foreground pointer
      zF1 = AnD1[AnimPnt];  // get the 2nd foreground pointer
      if (ForGndFrame < zF0) {ForGndFrame = zF0;} // not in current range so set to 1st
      else if (ForGndFrame >= zF1) {ForGndFrame = zF0;} // beyond current range, so reset
      else {ForGndFrame++;}                           // increment pointer
      Folder = "/Frames"; Name = "/Fr" + getNum4Digits(ForGndFrame);
      readSDToLED(1); // load the file into the foreground
      if (Verbose) {Serial.println("ForgndNext Fr" + getNum4Digits(ForGndFrame));}
      break;

    case 0x4697: // ForgndRot - Up/Right/Down/Left
      // rotation depends on the associated direction value
      switch (AnD0[AnimPnt]) {
        case 0: Gfx_RotUpFor(); zTxt = "Up"; break;     // rotate array Up
        case 1: Gfx_RotRhtFor(); zTxt = "Right"; break; // rotate array Right
        case 2: Gfx_RotDwnFor(); zTxt = "Down"; break;  // rotate array Down
        case 3: Gfx_RotLftFor(); zTxt = "Left"; break;  // rotate array Left
      }
      if (Verbose) {Serial.println("ForgndRot " + zTxt);}
      break;

    case 0x469B: // ForgndWinFlip - flip XY,XY region within the foreground in one of 2 directions
      // extract window co-ordinates
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      // rotation depends on the associated direction value
      switch (AnD2[AnimPnt]) {
        case 0: Gfx_FlipForgndH(zX0,zY0,zX1,zY1); zTxt = "Left-Right"; break;  // flip array window left-right
        case 1: Gfx_FlipForgndV(zX0,zY0,zX1,zY1); zTxt = "Up-Down"; break;     // flip array window up-down
      }
      if (Verbose) {Serial.println("ForgndWinFlip " + zTxt);}
      break;

    case 0x469A: // ForgndWinRot - rotate a region within the foreground in one of 4 directions
      // extract window co-ordinates
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      // rotation depends on the associated direction value
      switch (AnD2[AnimPnt]) {
        case 0: Gfx_RotUpWinFor(zX0,zY0,zX1,zY1); zTxt = "Up"; break;  // rotate array window Up
        case 1: Gfx_RotRhtWinFor(zX0,zY0,zX1,zY1); zTxt = "Right"; break; // rotate array window Right
        case 2: Gfx_RotDwnWinFor(zX0,zY0,zX1,zY1); zTxt = "Down"; break; // rotate array window Down
        case 3: Gfx_RotLftWinFor(zX0,zY0,zX1,zY1); zTxt = "Left"; break; // rotate array window Left
      }
      if (Verbose) {Serial.println("ForgndWinRot " + zTxt);}
      break;

    case 0x4698:  // For... - repeat a loop, bound by Next, n-time
      // ensure that the stack is not being exceeded?
      if (ForPnt >= 9) {AnimMode = -1; AnimErr = 2; break;} // ERROR!
      // adjust the stack pointer and store the count and return address
      if (AnD0[AnimPnt] == 0) {AnimMode = -1; AnimErr = 3; break;} // ERROR!
      ForPnt++; ForAdd[ForPnt] = AnimPnt;
      ForCnt[ForPnt] = AnD0[AnimPnt];
      if (Verbose) {Serial.println("For... " + String(ForCnt[ForPnt]));}
      break;

    case 0x4699:  // ForRnd - repeat a loop, bound by Next, A - B random times
      // ensure that the stack is not being exceeded?
      if (ForPnt >= 9) {AnimMode = -1; AnimErr = 2; break;} // ERROR!
      // adjust the stack pointer and store the count and return address
      ForPnt++; ForAdd[ForPnt] = AnimPnt;
      ForCnt[ForPnt] = random((AnD0[AnimPnt] & 0xFF00)>>8,(AnD0[AnimPnt] & 0xFF) + 1);
      if (ForCnt[ForPnt] == 0) {AnimMode = -1; AnimErr = 4; break;} // ERROR!
      if (Verbose) {Serial.println("ForRnd " + String(ForCnt[ForPnt]));}
      break;

    case 0x46B0: // FrameDim - dim the frame by a dim factor
      // dim the current frames pixels by a factor
      Gfx_DimFrame(1,PixWxPixH,AnD0[AnimPnt]);
      if (Verbose) {Serial.println("FrameDim " + String(AnD0[AnimPnt]));}
      break;

    case 0x46B5: // FrameDimE2 - dim the frame by a half
      Gfx_DimFrameE2(1,PixWxPixH);
      if (Verbose) {Serial.println("FrameDimE2");}
      break;

    case 0x46B1: // FrameDimTo - dim the frame toweards an RGB value by a factor
      // dim the current frames pixels by a factor towards a target RGB value
      getRGBcols(AnD0[AnimPnt]);
      Gfx_DimFrameTo(1,PixWxPixH,AnD1[AnimPnt],Pen_Red,Pen_Grn,Pen_Blu);
      if (Verbose) {Serial.println("FrameDimTo [RGB] " + String(AnD1[AnimPnt]));}
      break;

    case 0x46B2: // FrameLoad - load a file into the frame
      AnimFrame = AnD0[AnimPnt];
      Folder = "/Frames"; Name = "/Fr" + getNum4Digits(AnimFrame);
//      t0 = micros();
      readSDToLED(0);
//      t1 = micros();
      if (Verbose) {Serial.println("FrameLoad Fr" + getNum4Digits(AnimFrame));}
//      Serial.println("t1-t0 " + String(t1-t0));
      break;

    case 0x46B3: // FrameNext - load the next frame file in the sequence
      // loads the next frame in the sequence then cycles round
      zF0 = AnD0[AnimPnt];  // get the 1st frame pointer
      zF1 = AnD1[AnimPnt];  // get the 2nd frame pointer
      if (AnimFrame < zF0) {AnimFrame = zF0;} // not in current range so set to 1st
      else if (AnimFrame >= zF1) {AnimFrame = zF0;} // beyond current range, so reset
      else {AnimFrame++;}                           // increment pointer
      Folder = "/Frames"; Name = "/Fr" + getNum4Digits(AnimFrame);
      readSDToLED(0);
      if (Verbose) {Serial.println("FrameNext Fr" + getNum4Digits(AnimFrame));}
      break;

    case 0x46B4: // FrameRot - rotate the frame in one of 4 directions Up/Right/Down/Left
      // rotation depends on the associated direction value
      switch (AnD0[AnimPnt]) {
        case 0: Gfx_RotUp(); zTxt = "Up"; break;  // rotate array Up
        case 1: Gfx_RotRht(); zTxt = "Right"; break; // rotate array Right
        case 2: Gfx_RotDwn(); zTxt = "Down"; break; // rotate array Down
        case 3: Gfx_RotLft(); zTxt = "Left"; break; // rotate array Left
      }
      if (Verbose) {Serial.println("FrameRot " + zTxt);}
      break;

    case 0x46B7: // FrameWinFlip - flip XY,XY region within the foreground in one of 2 directions
      // extract window co-ordinates
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      // rotation depends on the associated direction value
      switch (AnD2[AnimPnt]) {
        case 0: Gfx_FlipFrameH(zX0,zY0,zX1,zY1); zTxt = "Left-Right"; break;  // flip array window left-right
        case 1: Gfx_FlipFrameV(zX0,zY0,zX1,zY1); zTxt = "Up-Down"; break;     // flip array window up-down
      }
      if (Verbose) {Serial.println("FrameWinFlip " + zTxt);}
      break;

    case 0x46B6: // FrameWinRot - rotate a region within the frame in one of 4 directions
      // extract window co-ordinates
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      // rotation depends on the associated direction value
      switch (AnD2[AnimPnt]) {
        case 0: Gfx_RotUpWin(zX0,zY0,zX1,zY1); zTxt = "Up"; break;  // rotate array window Up
        case 1: Gfx_RotRhtWin(zX0,zY0,zX1,zY1); zTxt = "Right"; break; // rotate array window Right
        case 2: Gfx_RotDwnWin(zX0,zY0,zX1,zY1); zTxt = "Down"; break; // rotate array window Down
        case 3: Gfx_RotLftWin(zX0,zY0,zX1,zY1); zTxt = "Left"; break; // rotate array window Left
      }
      if (Verbose) {Serial.println("FrameWinRot " + zTxt);}
      break;

    case 0x4793: // Gosub  - calls a subroutine at a labelled line
      // note that the Animation line numbers start from 1, here we start from 0
      if (GosubPnt < 9) {
        GosubPnt++; Gosub[GosubPnt] = AnimPnt;  // store return address on stack
        AnimPnt = AnD1[AnimPnt] - 1;            // get the branch to line value
        if (AnimPnt <= 0) {AnimErr = 7;}        // missing line reference
      } else {AnimErr = 6;}                     // stack overflow error
      if (Verbose) {Serial.println("Gosub " + String(AnimPnt + 1));}
      break;

    case 0x4794: // GoTo  - branches to a labelled line
      // note that the Animation line numbers start from 1, here we start from 0
      AnimPnt = AnD1[AnimPnt] - 1;              // get the branch to line value
      if (AnimPnt == 0) {AnimErr = 8;}          // missing line reference
      if (Verbose) {Serial.println("GoTo " + String(AnimPnt + 1));}
      break;

    case 0x47A0: // GoToIfScroll  - branches to a labelled line if text scrolling is in progress
      // note that the Animation line numbers start from 1, here we start from 0
      if (ScrollTask == 1) {AnimPnt = AnD1[AnimPnt] - 1;} // get the branch to line value
      if (AnimPnt == 0) {AnimErr = 10;}                   // missing line reference
      if (Verbose) {Serial.println("GoToIfScroll " + String(AnimPnt + 1));}
      break;

    case 0x47A2: // GoToIfSprit{M}  - branches to a labelled line if sprite move is in progress
      // note that the Animation line numbers start from 1, here we start from 0
      // this function uses an 8-bit mask stored in And2[]
      zX0 = AnD2[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {
          // test this sprite for movement
          zTxt = "1" + zTxt;
          if ((SpriteX[zX1] != SpriteTX[zX1]) ||(SpriteY[zX1] != SpriteTY[zX1])) {zNew = true;}
        } else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (zNew) {AnimPnt = AnD1[AnimPnt] - 1;} // get the branch to line value
      if (AnimPnt == 0) {AnimErr = 10;}                   // missing line reference
      if (Verbose) {Serial.println("GoToIfSprit{M} {" + zTxt + "}");}
      break;
      
    case 0x4990: // IcMax - set maximim LED current limit 0.1 - 50A
      // IcMax is x10 and ranges from 1 - 500
      IcMax = AnD0[AnimPnt]; zTxt = String(0.1 * IcMax,1); 
      if (Verbose) {Serial.println("IcMax " + zTxt + " A");}
      break;
      
    case 0x4991: // IcOff - turn IcMax current limit function off
      IcMax = 0; LevelOld = 0; LED[0].setRGB(0,0,0); // turn OFF micros LED
      if (Verbose) {Serial.println("IcOff - IcMax current limit is OFF");}
      break;

    case 0x4948: // IfActSched  - if active schedule, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      if (SchCnt > 0) {AnimIf = true;} // > 0 so do next line
      zTxt = "False, no schedule"; if (AnimIf) {zTxt = "True, schedule is loaded";}
      if (Verbose) {Serial.println("IfActSched - " + zTxt);}
      break;
      
    case 0x4950: // IfNotSprite{M} - if any sprite {M} are not moving, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {
          // test this sprite for movement
          zTxt = "1" + zTxt;
          if ((SpriteX[zX1] == SpriteTX[zX1]) && (SpriteY[zX1] == SpriteTY[zX1])) {
            AnimIf = true;  // 'If' condition is met, so do following lines
          }
        } else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("IfNotSprite{M} {" + zTxt + "}");}
      break;
      
    case 0x4951: // IfSprite{M} - if any sprite {M} are moving, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {
          // test this sprite for movement
          zTxt = "1" + zTxt;
          if ((SpriteX[zX1] != SpriteTX[zX1]) || (SpriteY[zX1] != SpriteTY[zX1])) {
            AnimIf = true;  // 'If' condition is met, so do following lines
          }
        } else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("IfSprite{M} {" + zTxt + "}");}
      break;

    case 0x4953: // IfVar >  - if variable 'A' - 'Z' > value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] > zX1) {AnimIf = true;} // > so do next line
      if (Verbose) {Serial.println("IfVar > '" + String(char(zX0 + 65)) + "' > " + String(zX1));}
      break;

    case 0x4954: // IfVar >= - if variable 'A' - 'Z' >= value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] >= zX1) {AnimIf = true;} // >= so do next line
      if (Verbose) {Serial.println("IfVar >= '" + String(char(zX0 + 65)) + "' >= " + String(zX1));}
      break;

    case 0x4955: // IfVar == - if variable 'A' - 'Z' == value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] == zX1) {AnimIf = true;} // == so do next line
      if (Verbose) {Serial.println("IfVar == '" + String(char(zX0 + 65)) + "' == " + String(zX1));}
      break;

    case 0x4956: // IfVar != - if variable 'A' - 'Z' != value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] != zX1) {AnimIf = true;} // != so do next line
      if (Verbose) {Serial.println("IfVar != '" + String(char(zX0 + 65)) + "' != " + String(zX1));}
      break;

    case 0x4957: // IfVar <= - if variable 'A' - 'Z' <= value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] <= zX1) {AnimIf = true;} // <= so do next line
      if (Verbose) {Serial.println("IfVar <= '" + String(char(zX0 + 65)) + "' <= " + String(zX1));}
      break;

    case 0x4958: // IfVar < - if variable 'A' - 'Z' < value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] < zX1) {AnimIf = true;} // < so do next line
      if (Verbose) {Serial.println("IfVar <  '" + String(char(zX0 + 65)) + "' < " + String(zX1));}
      break;

    case 0x4960: // IfVar > Var - if variable 'A' - 'Z' > Var, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] > Var[zX1]) {AnimIf = true;} // > so do next line
      if (Verbose) {Serial.println("IfVar > Var '" + String(char(zX0 + 65)) + "' > " + String(char(zX1 + 65)));}
      break;

    case 0x4961: // IfVar >= Var - if variable 'A' - 'Z' >= Var, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] >= Var[zX1]) {AnimIf = true;} // >= so do next line
      if (Verbose) {Serial.println("IfVar >= Var '" + String(char(zX0 + 65)) + "' >= " + String(char(zX1 + 65)));}
      break;

    case 0x4962: // IfVar == Var - if variable 'A' - 'Z' == Var, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] == Var[zX1]) {AnimIf = true;} // == so do next line
      if (Verbose) {Serial.println("IfVar == Var '" + String(char(zX0 + 65)) + "' == " + String(char(zX1 + 65)));}
      break;

    case 0x4963: // IfVar != Var - if variable 'A' - 'Z' != Var, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] != Var[zX1]) {AnimIf = true;} // != so do next line
      if (Verbose) {Serial.println("IfVar != Var '" + String(char(zX0 + 65)) + "' != " + String(char(zX1 + 65)));}
      break;

    case 0x4964: // IfVar <= - if variable 'A' - 'Z' <= value, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] <= Var[zX1]) {AnimIf = true;} // <= so do next line
      if (Verbose) {Serial.println("IfVar <= Var '" + String(char(zX0 + 65)) + "' <= " + String(char(zX1 + 65)));}
      break;

    case 0x4965: // IfVar < Var - if variable 'A' - 'Z' < Var, do next line
      AnimIf = false;  // default 'If failed' condition, skip following lines
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
      if (Var[zX0] < Var[zX1]) {AnimIf = true;} // < so do next line
      if (Verbose) {Serial.println("IfVar < Var '" + String(char(zX0 + 65)) + "' < " + String(char(zX1 + 65)));}
      break;

    case 0x49A0: // IncVar - increment variable 'A' - 'Z' by +1
      zX0 = AnD0[AnimPnt]; Var[zX0]++;
      if (Verbose) {Serial.println("IncVar '" + String(char(zX0 + 65)) + "'   " + String(Var[zX0]));}
      break;

    case 0x49A2: // IncVarToVar - increment variable 'A' - 'Z' by +/-1 towards another variable
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt];
           if (Var[zX1] > Var[zX0]) {Var[zX0]++;}
      else if (Var[zX1] < Var[zX0]) {Var[zX0]--;}
      if (Verbose) {Serial.println("IncVarToVar '" + String(char(zX0 + 65)) + "' = " + String(Var[zX0]) + " To '" + String(char(zX1 + 65)) + "' = " + String(Var[zX1]));}
      break;
      
    case 0x4c00: //Label - assigns a text label to a line for branch instructions like GoTo...
      if (Verbose) {Serial.println("Label:");}
      break;

    case 0x4C93: // LoadBckgnd - load background into frame store
      if (Verbose) {Serial.println("LoadBckgnd");}
      Gfx_CopyInBck(); break;

    case 0x4C97: // LoadBckWin[V] - copies background region into frame using variables
       // extract window co-ordinates from variables
      zX0 = (AnD0[AnimPnt] & 0xFF00)>>8; zY0 = AnD0[AnimPnt] & 0xFF;
      zX1 = (AnD1[AnimPnt] & 0xFF00)>>8; zY1 = AnD1[AnimPnt] & 0xFF;
      zX2 = (AnD2[AnimPnt] & 0xFF00)>>8; zY2 = AnD2[AnimPnt] & 0xFF;
      Gfx_CopyInBckWin(Var[zX0],Var[zY0],Var[zX1],Var[zY1],Var[zX2],Var[zY2]);
      if (Verbose) {Serial.println("LoadBckWin[V] " + String(Var[zX0]) + "," + String(Var[zY0]) + " " + String(Var[zX1]) + "," + String(Var[zY1]) + " " + " " + String(Var[zX2]) + "," + String(Var[zY2]) + " ");}
      break;

    case 0x4C95: // LoadBckWinXY - copies background window XY,XY into frame at XY
       // extract window co-ordinates in range -127 to 128
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      zX2 = ((AnD2[AnimPnt] & 0xFF00)>>8) - 127;
      zY2 =  (AnD2[AnimPnt] & 0xFF) - 127;
      Gfx_CopyInBckWin(zX0,zY0,zX1,zY1,zX2,zY2);
      if (Verbose) {Serial.println("LoadBckWinXY " + String(zX0) + "," + String(zY0) + " " + String(zX1) + "," + String(zY1) + " " + " " + String(zX2) + "," + String(zY2) + " ");}
      break;

    case 0x4C94: // LoadForgnd - load foreground into frame store
      if (Verbose) {Serial.println("LoadForgnd");}
      Gfx_CopyInFor(); break;

    case 0x4C98: // LoadForWin[V] - copies foreground region into frame using variables
       // extract window co-ordinates from variables
      zX0 = (AnD0[AnimPnt] & 0xFF00)>>8; zY0 = AnD0[AnimPnt] & 0xFF;
      zX1 = (AnD1[AnimPnt] & 0xFF00)>>8; zY1 = AnD1[AnimPnt] & 0xFF;
      zX2 = (AnD2[AnimPnt] & 0xFF00)>>8; zY2 = AnD2[AnimPnt] & 0xFF;
      Gfx_CopyInForWin(Var[zX0],Var[zY0],Var[zX1],Var[zY1],Var[zX2],Var[zY2]);
      if (Verbose) {Serial.println("LoadForWin[V] " + String(Var[zX0]) + "," + String(Var[zY0]) + " " + String(Var[zX1]) + "," + String(Var[zY1]) + " " + " " + String(Var[zX2]) + "," + String(Var[zY2]) + " ");}
      break;

    case 0x4C96: // LoadForWinXY - copies foreground window XY,XY into frame at XY
       // extract window co-ordinates in range -127 to 128
      zX0 = ((AnD0[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD0[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD1[AnimPnt] & 0xFF) - 127;
      zX2 = ((AnD2[AnimPnt] & 0xFF00)>>8) - 127;
      zY2 =  (AnD2[AnimPnt] & 0xFF) - 127;
      Gfx_CopyInForWin(zX0,zY0,zX1,zY1,zX2,zY2);
      if (Verbose) {Serial.println("LoadForWinXY " + String(zX0) + "," + String(zY0) + " " + String(zX1) + "," + String(zY1) + " " + " " + String(zX2) + "," + String(zY2) + " ");}
      break;

    case 0x4CA0: // Loop - jump back to Loop + 1
      if (Verbose) {Serial.println("Loop");}
      AnimLoop = AnimPnt; break;

    case 0x4E30: // Next - closes a For.. loop, repreats or carries on
      // check for Next without a For...
      if (ForPnt < 0) {AnimMode = -1; AnimErr = 5; break;} // ERROR!
      // if the for counter is > 1 then decrement and loop
      if (ForCnt[ForPnt] > 1) {ForCnt[ForPnt]--; AnimPnt = ForAdd[ForPnt];}
      else {ForCnt[ForPnt] = 0; ForPnt--;}
      if (Verbose) {Serial.println("Next");}
      break;

    case 0x5010: // Pause - sets a delay timer in 1/10ths second, 0.1 - 60.0 sec
      // Pause is x10 and ranges from 1 - 600
      AnimWait = 100 * AnD0[AnimPnt]; zTxt = String(0.1 * AnD0[AnimPnt],1); 
      if (Verbose) {Serial.println("Pause " + zTxt + " sec.");}
      PauseTimer = millis(); break;

    case 0x5030: // PenRGB - sets the pen RGB colour for drawing functions
      getRGBcols(AnD0[AnimPnt]);
      if (Verbose) {Serial.println("PenRGB " + String(Pen_Red) + "  " + String(Pen_Grn) + "  " + String(Pen_Blu));}
      break;

    case 0x5031: // PenRGB=[V] - set pen RGB to variable values
      Pen_Red = Var[AnD0[AnimPnt]]; Pen_Grn = Var[AnD1[AnimPnt]]; Pen_Blu = Var[AnD2[AnimPnt]];
      if (Verbose) {Serial.println("PenRGB=[V] " + String(Pen_Red) + "  " + String(Pen_Grn) + "  " + String(Pen_Blu));}
      break;

    case 0x5032: // PenRndRGB - assigns random RGB colour values to the Pen ink
      Gfx_RndCol(255); Pen_Red = ColRGB[0].r; Pen_Grn = ColRGB[0].g; Pen_Blu = ColRGB[0].b;
      if (Verbose) {Serial.println("PenRndRGB " + String(Pen_Red) + "  " + String(Pen_Grn) + "  " + String(Pen_Grn));}
      break;

    case 0x5228: // ReportON - report ON == true, OFF == false
      if (AnD0[AnimPnt] > 0) {RepON = true; Verbose = true; Serial.println("ReportON = True");}
      else {RepON = false;  Verbose = false;}
      break;
      
    case 0x5231: // Return - returns to the line immediately after the Gosub command
      if (GosubPnt >= 0) {AnimPnt = Gosub[GosubPnt]; GosubPnt--;}
      else {AnimErr = 9;} // no return line value
      if (Verbose) {Serial.println("Return");}
      break;

    case 0x5310: // SetVar= - set variable 'A' - 'Z' = value
      zX0 = AnD0[AnimPnt]; Var[zX0] = AnD1[AnimPnt];
      if (Verbose) {Serial.println("SetVar= '" + String(char(zX0 + 65)) + "' = " + String(AnD1[AnimPnt]));}
      break;

    case 0x5312: // SetVar=Rnd - set variable 'A' - 'Z' = random value
      zX1 = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 128,(AnD1[AnimPnt] & 0xFF) - 127);
      zX0 = AnD0[AnimPnt]; Var[zX0] = zX1;
      if (Verbose) {Serial.println("SetVar=Rnd '" + String(char(zX0 + 65)) + "' = " + String(zX1));}
      break;
      
    case 0x534E: // Show - show the contents of the frame buffer
      if (Verbose) {Serial.println("Show");}
      LEDshow = true; AnimBuild = false; break;

    case 0x539D: // Speed - set the animation speed
      if (AnD0[AnimPnt] < 1) {AnD0[AnimPnt] = 1;} // avoid div-zero
      if (Verbose) {Serial.println("Speed " + String(AnD0[AnimPnt]));}
      FramePeriod = 1000/AnD0[AnimPnt]; break;  // convert fps to ms

    case 0x539F: // Speed=Var - set the Anim speed = variable 'A' - 'Z'
      zX0 = AnD0[AnimPnt]; zX1 = constrain(Var[zX0],1,50);
      FramePeriod = 1000/zX1;
      if (Verbose) {Serial.println("Speedt=Var '" + String(char(zX0 + 65)) + "' = " + String(zX1));}
      break;

    case 0x53A8: // SpriteAtXY  - place sprite N at X,Y
      SpritePnt = AnD0[AnimPnt];  // get the sprite pointer 0 - 7
      SpriteX[SpritePnt] = AnD1[AnimPnt]; // store X value
      SpriteY[SpritePnt] = AnD2[AnimPnt]; // store Y value
      // set target values to match
      SpriteTX[SpritePnt] = SpriteX[SpritePnt]; SpriteTY[SpritePnt] = SpriteY[SpritePnt];
      if (Verbose) {Serial.println("SpriteAtXY " + String(SpritePnt) + " X=" + String(SpriteX[SpritePnt]) + " Y=" + String(SpriteY[SpritePnt]));}
      break;
 
    case 0x53B6: // SpriteAtXY=[V]  - place sprite N at variables for X,Y
      SpritePnt = AnD0[AnimPnt];  // get the sprite pointer 0 - 7
      zX1 = AnD1[AnimPnt]; zY1 = AnD2[AnimPnt]; 
      SpriteX[SpritePnt] = Var[zX1]; // store X value
      SpriteY[SpritePnt] = Var[zY1]; // store Y value
      // set target values to match
      SpriteTX[SpritePnt] = SpriteX[SpritePnt]; SpriteTY[SpritePnt] = SpriteY[SpritePnt];
      if (Verbose) {Serial.println("SpriteAtXY=[V] " + String(SpritePnt) + " X=" + String(SpriteX[SpritePnt]) + " Y=" + String(SpriteY[SpritePnt]));}
      break;
     
    case 0x53AC: // SpriteAtRndXY - places sprite N at random X,Y
      SpritePnt = AnD0[AnimPnt];  // get the sprite pointer 0 - 7
      SpriteX[SpritePnt] = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 128,(AnD1[AnimPnt] & 0xFF) - 127);
      SpriteY[SpritePnt] = random(((AnD2[AnimPnt] & 0xFF00)>>8) - 128,(AnD2[AnimPnt] & 0xFF) - 127);
      SpriteTX[SpritePnt] = SpriteX[SpritePnt]; SpriteTY[SpritePnt] = SpriteY[SpritePnt];
      if (Verbose) {Serial.println("SpriteRndtXY " + String(SpritePnt) + " X=" + String(SpriteX[SpritePnt]) + " Y=" + String(SpriteY[SpritePnt]));}
      break;
      
    case 0x53B1: // SpriteFlipH{M} - flip sprite {M} region, horizontally
      zW = AnD0[AnimPnt];  // get the 8-bit mask
      zX0 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD1[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD2[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD2[AnimPnt] & 0xFF) - 127;
      // ensure that zX1 > zX0
      if (zX0 > zX1) {zH = zX0; zX0 = zX1; zX1 = zH;}
      // ensure that zY1 > zY0
      if (zY0 > zY1) {zH = zY0; zY0 = zY1; zY1 = zH;}
      for (zH = 0; zH <= 7;zH++) {
        if ((zW & 1) > 0) {
          zTxt = "1" + zTxt;
          Gfx_SpriteFlipH(zH,zX0,zY0,zX1,zY1);
        } else {zTxt = "0" + zTxt;}
        zW = zW>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("SpriteFlipH{M} {" + zTxt + "}");}
      break;
      
    case 0x53B2: // SpriteFlipV{M} - flip sprite {M} region, vertically
      zW = AnD0[AnimPnt];  // get the 8-bit mask
      zX0 = ((AnD1[AnimPnt] & 0xFF00)>>8) - 127;
      zY0 =  (AnD1[AnimPnt] & 0xFF) - 127;
      zX1 = ((AnD2[AnimPnt] & 0xFF00)>>8) - 127;
      zY1 =  (AnD2[AnimPnt] & 0xFF) - 127;
      // ensure that zX1 > zX0
      if (zX0 > zX1) {zH = zX0; zX0 = zX1; zX1 = zH;}
      // ensure that zY1 > zY0
      if (zY0 > zY1) {zH = zY0; zY0 = zY1; zY1 = zH;}
      for (zH = 0; zH <= 7;zH++) {
        if ((zW & 1) > 0) {
          zTxt = "1" + zTxt;
          Gfx_SpriteFlipV(zH,zX0,zY0,zX1,zY1);
        } else {zTxt = "0" + zTxt;}
        zW = zW>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("SpriteFlipV{M} {" + zTxt + "}");}
      break;
      
    case 0x53A9: // SpriteLoad - load a file into sprite 0 - 7
      SpritePnt = AnD0[AnimPnt];  // get the sprite pointer 0 - 7
      SpriteNum = AnD1[AnimPnt];  // get image file number
      SpriteFN[SpritePnt] = SpriteNum;  // store the file loaded reference
      Folder = "/Sprite16"; Name = "/Sp" + getNum4Digits(SpriteNum);
      ClearSpriteImage(SpritePnt); readSDToLED(3);
      if (Verbose) {Serial.println("SpriteLoad " + String(SpritePnt) + " Sp" + getNum4Digits(SpriteNum));}
      break;

    case 0x53AA: // SpriteNext - load the next sprite file in the sequence
      SpritePnt = AnD0[AnimPnt];  // get the sprite pointer 0 - 7
      zF0 = AnD1[AnimPnt];        // get the 1st SpNNNNNN pointer
      zF1 = AnD2[AnimPnt];        // get the 2nd SpNNNNNN pointer
      SpriteNum = SpriteFN[SpritePnt];  // get the current file number
      if (SpriteNum < zF0) {SpriteNum = zF0;}       // not in current range so set to 1st
      else if (SpriteNum >= zF1) {SpriteNum = zF0;} // beyond current range, so reset
      else {SpriteNum++;}                           // increment pointer
      SpriteFN[SpritePnt] = SpriteNum;              // store the file loaded reference
      Folder = "/Sprite16"; Name = "/Sp" + getNum4Digits(SpriteNum);
      ClearSpriteImage(SpritePnt); readSDToLED(3);
      if (Verbose) {Serial.println("SpriteNext " + String(SpritePnt) + " Sp" + getNum4Digits(SpriteNum));}
      break;
      
    case 0x53AB: // SpriteOn - turn sprite N ON/OFF true/false
      SpritePnt = AnD0[AnimPnt];  // get the sprite pointer 0 - 7
      if (AnD1[AnimPnt] > 0) {SpriteON[SpritePnt] = true; zTxt = "True";}
      else {SpriteON[SpritePnt] = false; zTxt = "False";}
      if (Verbose) {Serial.println("SpriteOn " + String(SpritePnt) + " " + zTxt);}
      break;
      
    case 0x53AE: // SpriteOn{M} - turn sprites ON/OFF mask 00000000 - 11111111
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {SpriteON[zX1] = true; zTxt = "1" + zTxt;} else {SpriteON[zX1] = false; zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("SpriteOn{M} {" + zTxt + "}");}
      break;
      
    case 0x53B0: // SpriteRndXY{M} - set sprite {0000000} random X,Y targets
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {
          zTxt = "1" + zTxt;
          // get new random target values for each sprite
          SpriteTX[zX1] = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 128,(AnD1[AnimPnt] & 0xFF) - 127);
          SpriteTY[zX1] = random(((AnD2[AnimPnt] & 0xFF00)>>8) - 128,(AnD2[AnimPnt] & 0xFF) - 127);
          if ((SpriteTX[zX1] != SpriteX[zX1]) || (SpriteTY[zX1] != SpriteY[zX1])) {
            // sprite is not at new target co-ordinates, so move it in that direction
            // get increment value as largest X,Y difference
            if (abs(SpriteTX[zX1] - SpriteX[zX1]) > abs(SpriteTY[zX1] - SpriteY[zX1])) {
              SpriteInc[zX1] = abs(SpriteTX[zX1] - SpriteX[zX1]);
            } else {SpriteInc[zX1] = abs(SpriteTY[zX1] - SpriteY[zX1]);}
            SpriteDiv[zX1] = SpriteInc[zX1]; if (SpriteDiv[zX1] < 1) {SpriteDiv[zX1] = 1;}
            // set start points
            SpriteSX[zX1] = SpriteX[zX1]; SpriteSY[zX1] = SpriteY[zX1];
            SpriteInc[zX1]--; int zAdj = SpriteDiv[zX1]/2;
            if (SpriteInc[zX1] < 1) {
              // reached target
              SpriteInc[zX1] = 0; SpriteX[zX1] = SpriteTX[zX1]; SpriteY[zX1] = SpriteTY[zX1];
            } else {
              // still moving towards target
              SpriteX[zX1] = SpriteTX[zX1] - ((((SpriteTX[zX1] - SpriteSX[zX1])*SpriteInc[zX1]) - zAdj)/SpriteDiv[zX1]);
              SpriteY[zX1] = SpriteTY[zX1] - ((((SpriteTY[zX1] - SpriteSY[zX1])*SpriteInc[zX1]) - zAdj)/SpriteDiv[zX1]);
            }
          }
        } else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("SpriteRndXY{M} {" + zTxt + "}");}
      break;

    case 0x53B4: // SpriteRot{M} - rotate sprite {0000000} in one of 4 directions
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {
          zTxt = "1" + zTxt;
          switch(AnD1[AnimPnt]) {
            case 0: Gfx_SpriteRotUp(zX1); zS$ = "Up"; break;   // rotate sprite array Up
            case 1: Gfx_SpriteRotRht(zX1); zS$ = "Right"; break;  // rotate sprite array Right
            case 2: Gfx_SpriteRotDwn(zX1); zS$ = "Down"; break;  // rotate sprite array Down
            case 3: Gfx_SpriteRotLft(zX1); zS$ = "Left"; break;  // rotate sprite array Left
          }
         } else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("SpriteRot{M} {" + zTxt + "} " + zS$);}
      break;
      
    case 0x53AF: // SpriteTo{M} - move sprite {00000000} towards TX,TY
      zX0 = AnD0[AnimPnt];  // get the 8-bit mask
      for (zX1 = 0; zX1 <= 7;zX1++) {
        if ((zX0 & 1) > 0) {
          zTxt = "1" + zTxt;
          // check to see if the sprite needs to move?
          if ((SpriteX[zX1] != SpriteTX[zX1]) || (SpriteY[zX1] != SpriteTY[zX1])) {
            SpriteInc[zX1]--; int zAdj = SpriteDiv[zX1]/2;
            if (SpriteInc[zX1] < 1) {
              // reached target
              SpriteInc[zX1] = 0; SpriteX[zX1] = SpriteTX[zX1]; SpriteY[zX1] = SpriteTY[zX1];
            } else {
              // sprite is still moving towards target
              SpriteX[zX1] = SpriteTX[zX1] - ((((SpriteTX[zX1] - SpriteSX[zX1])*SpriteInc[zX1]) - zAdj)/SpriteDiv[zX1]);
              SpriteY[zX1] = SpriteTY[zX1] - ((((SpriteTY[zX1] - SpriteSY[zX1])*SpriteInc[zX1]) - zAdj)/SpriteDiv[zX1]);
            }
          }
        } else {zTxt = "0" + zTxt;}
        zX0 = zX0>>1; // move to the next bit
      }
      if (Verbose) {Serial.println("SpriteTo{M} {" + zTxt + "}");}
      break;
      
    case 0x53AD: // SpriteToXY - move sprite N towards X,Y
      // if X,Y remains constant then sprite moves towards X,Y with each command
      // if X,Y changes then start a move towards that new X,Y co-ordiante
      SpritePnt = AnD0[AnimPnt];            // get the sprite pointer 0 - 7
      if (SpriteTX[SpritePnt] != AnD1[AnimPnt]) {SpriteTX[SpritePnt] = AnD1[AnimPnt]; zNew = true;}  // store target X value
      if (SpriteTY[SpritePnt] != AnD2[AnimPnt]) {SpriteTY[SpritePnt] = AnD2[AnimPnt]; zNew = true;}  // store target Y value
      if ((SpriteTX[SpritePnt] != SpriteX[SpritePnt]) || (SpriteTY[SpritePnt] != SpriteY[SpritePnt])) {
        // sprite is not at target co-ordinates, so move it in that direction
        // if zNew get increment value as largest X,Y difference
        if (zNew) {
          // get new increment value
          if (abs(SpriteTX[SpritePnt] - SpriteX[SpritePnt]) > abs(SpriteTY[SpritePnt] - SpriteY[SpritePnt])) {
            SpriteInc[SpritePnt] = abs(SpriteTX[SpritePnt] - SpriteX[SpritePnt]);
          } else {SpriteInc[SpritePnt] = abs(SpriteTY[SpritePnt] - SpriteY[SpritePnt]);}
          SpriteDiv[SpritePnt] = SpriteInc[SpritePnt]; if (SpriteDiv[SpritePnt] < 1) {SpriteDiv[SpritePnt] = 1;}
          SpriteSX[SpritePnt] = SpriteX[SpritePnt]; SpriteSY[SpritePnt] = SpriteY[SpritePnt];
        }
        SpriteInc[SpritePnt]--; int zAdj = SpriteDiv[SpritePnt]/2;
        if (SpriteInc[SpritePnt] < 1) {
          // reached target
          SpriteInc[SpritePnt] = 0; SpriteX[SpritePnt] = SpriteTX[SpritePnt]; SpriteY[SpritePnt] = SpriteTY[SpritePnt];
        } else {
          // still moving towards target
          SpriteX[SpritePnt] = SpriteTX[SpritePnt] - ((((SpriteTX[SpritePnt] - SpriteSX[SpritePnt])*SpriteInc[SpritePnt]) - zAdj)/SpriteDiv[SpritePnt]);
          SpriteY[SpritePnt] = SpriteTY[SpritePnt] - ((((SpriteTY[SpritePnt] - SpriteSY[SpritePnt])*SpriteInc[SpritePnt]) - zAdj)/SpriteDiv[SpritePnt]);
        }
      } else {
        // sprite is at target co-ordinates
      }
      if (Verbose) {Serial.println("SpriteToXY " + String(SpritePnt) + " X=" + String(SpriteX[SpritePnt]) + " Y=" + String(SpriteY[SpritePnt]));}
      break;
      
    case 0x53B8: // SpriteToXY=[V] - move sprite N to variables for X,Y
      // if X,Y remains constant then sprite moves towards X,Y with each command
      // if X,Y changes then start a move towards that new X,Y co-ordiante
      SpritePnt = AnD0[AnimPnt];            // get the sprite pointer 0 - 7
      zX1 = AnD1[AnimPnt]; zY1 = AnD2[AnimPnt]; 
      if (SpriteTX[SpritePnt] != Var[zX1]) {SpriteTX[SpritePnt] = Var[zX1]; zNew = true;}  // store target X value
      if (SpriteTY[SpritePnt] != Var[zY1]) {SpriteTY[SpritePnt] = Var[zY1]; zNew = true;}  // store target Y value
      if ((SpriteTX[SpritePnt] != SpriteX[SpritePnt]) || (SpriteTY[SpritePnt] != SpriteY[SpritePnt])) {
        // sprite is not at target co-ordinates, so move it in that direction
        // if zNew get increment value as largest X,Y difference
        if (zNew) {
          // get new increment value
          if (abs(SpriteTX[SpritePnt] - SpriteX[SpritePnt]) > abs(SpriteTY[SpritePnt] - SpriteY[SpritePnt])) {
            SpriteInc[SpritePnt] = abs(SpriteTX[SpritePnt] - SpriteX[SpritePnt]);
          } else {SpriteInc[SpritePnt] = abs(SpriteTY[SpritePnt] - SpriteY[SpritePnt]);}
          SpriteDiv[SpritePnt] = SpriteInc[SpritePnt]; if (SpriteDiv[SpritePnt] < 1) {SpriteDiv[SpritePnt] = 1;}
          SpriteSX[SpritePnt] = SpriteX[SpritePnt]; SpriteSY[SpritePnt] = SpriteY[SpritePnt];
        }
        SpriteInc[SpritePnt]--; int zAdj = SpriteDiv[SpritePnt]/2;
        if (SpriteInc[SpritePnt] < 1) {
          // reached target
          SpriteInc[SpritePnt] = 0; SpriteX[SpritePnt] = SpriteTX[SpritePnt]; SpriteY[SpritePnt] = SpriteTY[SpritePnt];
        } else {
          // still moving towards target
          SpriteX[SpritePnt] = SpriteTX[SpritePnt] - ((((SpriteTX[SpritePnt] - SpriteSX[SpritePnt])*SpriteInc[SpritePnt]) - zAdj)/SpriteDiv[SpritePnt]);
          SpriteY[SpritePnt] = SpriteTY[SpritePnt] - ((((SpriteTY[SpritePnt] - SpriteSY[SpritePnt])*SpriteInc[SpritePnt]) - zAdj)/SpriteDiv[SpritePnt]);
        }
      } else {
        // sprite is at target co-ordinates
      }
      if (Verbose) {Serial.println("SpriteToXY=[V] " + String(SpritePnt) + " X=" + String(SpriteX[SpritePnt]) + " Y=" + String(SpriteY[SpritePnt]));}
      break;

    case 0x5430: // Text - defines Text string
      Text = AnStr[AnimPnt]; ScrollTask = 0;  // load text and reset scroll task pointer
      Word = ""; WordRht = ""; // reset 'Word' function
      if (Verbose) {Serial.println("Text " + Text);}
      break;

    case 0x5431: // Text+ - adds to Text string
      Text+= AnStr[AnimPnt];
      if (Verbose) {Serial.println("Text+ " + AnStr[AnimPnt]);}
      break;

    case 0x5434: // TextAtXY - defines Text position as X,Y
      Text_X = AnD0[AnimPnt]; Text_Y = AnD1[AnimPnt];
      ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
      Gfx_DrawTxt(Text_X,Text_Y);
      if (Verbose) {Serial.println("TextAtXY " + String(Text_X) + "," + String(Text_Y));}
      break;

    case 0x5435: // TextAtXY=[V] - draw Text at variables for X,Y
      zX1 = AnD0[AnimPnt]; zY1 = AnD1[AnimPnt]; 
      Text_X = Var[zX1]; Text_Y = Var[zY1];
      ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
      Gfx_DrawTxt(Text_X,Text_Y);
      if (Verbose) {Serial.println("TextAtXY " + String(Text_X) + "," + String(Text_Y));}
      break;

    case 0x543A: // TextRGB - assigns RGB colour to next Text drawn
      getTxtRGB(AnD0[AnimPnt]);
      ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
      if (Verbose) {Serial.println("TextRGB " + String(Txt_Red) + "  " + String(Txt_Grn) + "  " + String(Txt_Grn));}
      break;

    case 0x543D: // TextRGB=[V] - set Text RGB to variable values
      Txt_Red = Var[AnD0[AnimPnt]]; Txt_Grn = Var[AnD1[AnimPnt]]; Txt_Blu = Var[AnD2[AnimPnt]];
      if (Verbose) {Serial.println("TextRGB=[V] " + String(Txt_Red) + "  " + String(Txt_Grn) + "  " + String(Txt_Blu));}
      break;

    case 0x543C: // TextRndRGB - assigns random RGB colours to next Text drawn
      Gfx_RndCol(255); Txt_Red = ColRGB[0].r; Txt_Grn = ColRGB[0].g; Txt_Blu = ColRGB[0].b;
      if (Verbose) {Serial.println("TextRndRGB " + String(Txt_Red) + "  " + String(Txt_Grn) + "  " + String(Txt_Grn));}
      break;

    case 0x543B: // TextRndXY - defines Text position as random X,Y
      Text_X = random(((AnD0[AnimPnt] & 0xFF00)>>8) - 128,(AnD0[AnimPnt] & 0xFF) - 127);
      Text_Y = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 128,(AnD1[AnimPnt] & 0xFF) - 127);
      ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
      Gfx_DrawTxt(Text_X,Text_Y);
      if (Verbose) {Serial.println("TextRndXY " + String(Text_X) + "," + String(Text_Y));}
      break;

    case 0x5440:  // TextScroll - scrolls text 1 pixel to the left
      Gfx_ScrollTxt();
      if (Verbose) {Serial.println("TextScroll");}
      break;

    case 0x5441:  // TextScrollY - set yext Y for scrolling
      // setting the Text_SY value resets the scroll task
      Text_SY = AnD0[AnimPnt]; ScrollTask = 0;
      if (Verbose) {Serial.println("TextScrollY " + String(Text_SY));}
      break;

    case 0x5660:  // Var+=Var - adds the 2nd variable to the 1st
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] += Var[zX1];
      if (Verbose) {Serial.println("Var+=Var '" + String(char(zX0 + 65)) + "' += " + String(char(zX1 + 65)));}
      break;

    case 0x5662:  // Var =Var - assigns the 2nd variables value to the 1st
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] = Var[zX1];
      if (Verbose) {Serial.println("Var =Var '" + String(char(zX0 + 65)) + "'  = " + String(char(zX1 + 65)));}
      break;

    case 0x5664:  // Var-=Var - subtracts the 2nd variable from the 1st
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] -= Var[zX1];
      if (Verbose) {Serial.println("Var-=Var '" + String(char(zX0 + 65)) + "' -= " + String(char(zX1 + 65)));}
      break;

    case 0x5666:  // Var*=Var - multiplies the 1st variable by the 2nd
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] *= Var[zX1];
      if (Verbose) {Serial.println("Var*=Var '" + String(char(zX0 + 65)) + "' *= " + String(char(zX1 + 65)));}
      break;

    case 0x5668:  // Var/=Var - divides the 1st variable by the 2nd
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      if (Var[zX1] != 0) {
        Var[zX0] *= Var[zX1];
        if (Verbose) {Serial.println("Var/=Var '" + String(char(zX0 + 65)) + "' /= " + String(char(zX1 + 65)));}
      } else {if (Verbose) {Serial.println("Var/=Var ERROR - divide by zero");}}
      break;

    case 0x566A:  // Var&=Var - logically ANDs the 2nd variable with the 1st
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] &= Var[zX1];
      if (Verbose) {Serial.println("Var&=Var '" + String(char(zX0 + 65)) + "' &= " + String(char(zX1 + 65)));}
      break;

    case 0x566C:  // Var|=Var - logically ORs the 2nd variable with the 1st
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] |= Var[zX1];
      if (Verbose) {Serial.println("Var|=Var '" + String(char(zX0 + 65)) + "' |= " + String(char(zX1 + 65)));}
      break;

    case 0x566E:  // Var^=Var - logically XORs the 2nd variable with the 1st
      zX0 = AnD0[AnimPnt]; zX1 = AnD1[AnimPnt]; 
      Var[zX0] ^= Var[zX1];
      if (Verbose) {Serial.println("Var^=Var '" + String(char(zX0 + 65)) + "' ^= " + String(char(zX1 + 65)));}
      break;
      
    case 0x5780: // WordAtXY - draw Word at X,Y
      // if Word exists it is simply drawn on the display, otherwise it is extracted from Text
      if (Word.length() < 1) {getWord();}
      if (Word.length() > 0) {
        // A Word has been extracted so use it
        Text_X = AnD0[AnimPnt]; Text_Y = AnD1[AnimPnt];
        ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
        Gfx_DrawStr(Word,Text_X,Text_Y);
      } else {AnimErr = 14;}
      if (Verbose) {Serial.println("WordAtXY '" + Word + "' " + String(Text_X) + "," + String(Text_Y));}
      break;
      
    case 0x5781: // WordAtXY=[V] - draw a Word at variables for X,Y
      // if Word exists it is simply drawn on the display, otherwise it is extracted from Text
      if (Word.length() < 1) {getWord();}
      if (Word.length() > 0) {
        // A Word has been extracted so use it
        zX1 = AnD0[AnimPnt]; zY1 = AnD1[AnimPnt]; 
        Text_X = Var[zX1]; Text_Y = Var[zY1];
        ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
        Gfx_DrawStr(Word,Text_X,Text_Y);
      } else {AnimErr = 14;}
      if (Verbose) {Serial.println("WordAtXY=[V] '" + Word + "' " + String(Text_X) + "," + String(Text_Y));}
      break;

    case 0x5782: // WordCtrXY - draw a Word centred on X,Y
      // if Word exists it is simply drawn on the display, otherwise it is extracted from Text
      if (Word.length() < 1) {getWord();}
      if (Word.length() > 0) {
        // A Word has been extracted so use it
        zX0 = Gfx_GetTextLen(Word);
        Text_X = AnD0[AnimPnt] - (zX0/2); Text_Y = AnD1[AnimPnt] - (GH/2) - 1;
        ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu);
        Gfx_DrawStr(Word,Text_X,Text_Y);
      } else {AnimErr = 14;}
      if (Verbose) {Serial.println("WordCtrXY '" + Word + "' " + String(Text_X) + "," + String(Text_Y));}
      break;

    case 0x5783: // WordCtrXY=[V] - draw a Word centred on variables X,Y
      // if Word exists it is simply drawn on the display, otherwise it is extracted from Text
      if (Word.length() < 1) {getWord();}
      if (Word.length() > 0) {
        // A Word has been extracted so use it
        zX0 = Gfx_GetTextLen(Word);
        zX1 = AnD0[AnimPnt]; zY1 = AnD1[AnimPnt]; 
        Text_X = Var[zX1] - (zX0/2); Text_Y = Var[zY1] - (GH/2) - 1;
        ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu); Gfx_DrawStr(Word,Text_X,Text_Y);
      } else {AnimErr = 14;}
      if (Verbose) {Serial.println("WordCtrXY=[V] '" + Word + "' " + String(Text_X) + "," + String(Text_Y));}
      break;
      
    case 0x5784: // WordNext - get the next word from WordRht, or cycle
      WordRht.trim();   // ensure this isn't just spaces
      if (WordRht.length() > 0) {getWordNext();}
      else {getWord();} // there is no remaining string, so start again
      if (Verbose) {Serial.println("WordNext '" + Word + "'");}
      break;

    case 0x5786: // WordRndXY - draw a Word at random X,Y
      // if Word exists it is simply drawn on the display, otherwise it is extracted from Text
      if (Word.length() < 1) {getWord();}
      if (Word.length() > 0) {
        // A Word has been extracted so use it
        zX0 = Gfx_GetTextLen(Word);
        zX1 = random(((AnD0[AnimPnt] & 0xFF00)>>8) - 128,(AnD0[AnimPnt] & 0xFF) - 127);
        zY1 = random(((AnD1[AnimPnt] & 0xFF00)>>8) - 128,(AnD1[AnimPnt] & 0xFF) - 127);
        Text_X = zX1 - (zX0/2); Text_Y = zY1 - (GH/2) - 1;
        ColRGB[0].setRGB(Txt_Red,Txt_Grn,Txt_Blu); Gfx_DrawStr(Word,Text_X,Text_Y);
      } else {AnimErr = 14;}
      if (Verbose) {Serial.println("WordRndXY '" + Word + "' " + String(Text_X) + "," + String(Text_Y));}
      break;

    default:
      if (AnCmd[AnimPnt] >= 32769) {break;} // inhibited command line
      if (Verbose) {Serial.println("Cmd ????!");}
      break;
  }
  if (!AnimPLAY) {return;}  // fast exit when AnimPLAY == false
  if (AnimSTOP) {return;}   // fast exit when AnimSTOP == true

  if (AnimIf) {
    // do next line, default = true
    AnimPnt+= AnimInc;
  } else {getEndIf();} 
  if (AnimPnt >= AnimMax) {Serial.println("EOF"); AnimSTOP = true;} // STOP

  // Error handler
  if (AnimErr == 0) {return;} // no errors
  
  // Report error and stop
       if (AnimErr ==  1) {Serial.println("Err - Loop without EndLoop");}
  else if (AnimErr ==  2) {Serial.println("Err - For... stack overflow");}
  else if (AnimErr ==  3) {Serial.println("Err - For... zero times");}
  else if (AnimErr ==  4) {Serial.println("Err - ForRnd zero times");}
  else if (AnimErr ==  5) {Serial.println("Err - Next without a For...");}
  else if (AnimErr ==  6) {Serial.println("Err - Gosub stack overflow");}
  else if (AnimErr ==  7) {Serial.println("Err - Gosub missing line ref");}
  else if (AnimErr ==  8) {Serial.println("Err - GoTo missing line ref");}
  else if (AnimErr ==  9) {Serial.println("Err - missing Return line ref");}
  else if (AnimErr == 10) {Serial.println("Err - GoIfScroll missing line ref");}
  else if (AnimErr == 11) {Serial.println("Err - AnimAdd array depth exceeded");}
  else if (AnimErr == 12) {Serial.println("Err - AnimEsc Anim file read failure");}
  else if (AnimErr == 13) {Serial.println("Err - AnimRun no schedule");}
  else if (AnimErr == 14) {Serial.println("Err - Text has no words");}
  else if (AnimErr == 15) {Serial.println("Err - If without EndIf");}
  AnimBuild = false; LEDshow = false; AnimMode = -1; // STOP
}

// ---------------------------------------------------------------------------------
