
/* SSD1306Wire.h functions, see https://github.com/ThingPulse/esp8266-oled-ssd1306

  disply.clear()        - clear the local pixel buffer
  disply.display()      - write the buffer to the display
  disply.displayOff()   - turn the display OFF
  disply.displayOn()    - turn the display ON
  display.drawCircle(int16_t x, int16_t y, int16_t radius)
  display.drawFastImage(int16_t x, int16_t y, int16_t width, int16_t height, const uint8_t *image)
  display.drawHorizontalLine(int16_t x, int16_t y, int16_t length)
  display.drawLine(int16_t x0, int16_t y0, int16_t x1, int16_t y1)  - draw from x0k,y0 to x1,y1
  display.drawProgressBar(uint16_t x, uint16_t y, uint16_t width, uint16_t height, uint8_t progress)
  display.drawRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.drawString(int16_t x, int16_t y, String text)
  display.drawStringMaxWidth(int16_t x, int16_t y, int16_t maxLineWidth, String text)
  display.drawVerticalLine(int16_t x, int16_t y, int16_t length)
  display.drawXbm(int16_t x, int16_t y, int16_t width, int16_t height, const char* xbm)
  display.fillCircle(int16_t x, int16_t y, int16_t radius)
  display.fillRect(int16_t x, int16_t y, int16_t width, int16_t height)
  display.flipScreenVertically()   - flip the current screen upside down
  display.init()         - initialise the display
  display.invertDisplay()- inverted display mode
  display.end()          - free memory used by the display
  display.displayOn()    - turn the display ON
  display.mirrorScreen() - draw the screen mirrored
  display.normalDisplay()- normal display mode
  display.reconnect()    - connect again through I2C
  display.setBrightness(uint8_t)
  display.setColor(OLEDDISPLAY_COLOR color) - BLACK, WHITE, INVERSE
  display.setContrast(uint8_t contrast, uint8_t precharge = 241, uint8_t comdetect = 64)
  display.setFont(const uint8_t* fontData)
  display.setPixel(int16_t x, int16_t y)  - set a pixel at x,y
  display.setTextAlignment(OLEDDISPLAY_TEXT_ALIGNMENT textAlignment)  - TEXT_ALIGN_LEFT, TEXT_ALIGN_CENTER, TEXT_ALIGN_RIGHT, TEXT_ALIGN_CENTER_BOTH

  W = display.getStringWidth(const char* text, uint16_t length)
  W = display.getStringWidth(String text)
  
  Display size is 128 x 64 (w x h) pixels
  Default fonts are ArialMT at 10, 16 & 24 pixels high
  Text options include:
    2 x 24, + 1 x 10 pnt lines
    1 x 24, + 2 x 16 pnt lines
    1 x 24, + 3 x 10 pnt lines
    4 x 16 pnt lines, no line spaces, decending characters will merge
    3 x 16, + 1 x 10 pnt lines
    6 x 10 pnt lines, no line spaces, decending characters will merge
*/

// --------------------------------------------------------------------------------

void Display_40ms() {
  // called every 40ms from the main loop
  // if DispDel > 0 then we wait for DispDel cycles to complete before update
  if (!BUILD) {return;}   // exit if no display is connected
  if (BootAnim) {return;} // no display when running anim from boot
  if (DispNOP) {return;}  // this function is aborted, no I2C activity
  if (DispLock) {Display_Locked(); return;} // lock the display
  
  if (DispDel > 0) {DispDel--; return;} // don't update display if DispDel > 0
  if (DispClr) {display.clear(); DispClr = false; return;}
  if (DispDisp) {display.display(); DispDisp = false; DispClr = true; return;}

  // we refresh individual displays based on a counter, DispCnt
  // set DispCnt -ve to prevent the display update process
  DispCnt--;
  if (DispCnt < 1) {
    // DispCnt is reset for display updates, so can be different for each
    if (DispBack) {
      // draw a back arrow on the screen
//      display.clear();
//      display.drawLine(48,32,80,16);
//      display.drawLine(48,32,80,48);
//      display.drawLine(80,16,80,48);
//      DispCnt = 10;
    } else {
      if (DispNow > 0) {
        // special display modes used in calibration etc.
        switch(DispNow) {
//          case 1: DispCnt = 3; Display_ServoVals(); break;
        }
      } else {
        if (TEST) {
          // in TEST mode we use a special display
          DispCnt = 2; Display_TESTchars();
        } else {
          // normal menu screens, display depends on menu modes
          // if DispMode == 0 then a normal menu is displayed
          // otherwise a custom display task is run
          switch(DispMode) {
            case 0: DispCnt = 5; Display_Menu(); break;
            case 1: DispCnt = 2; Display_TESTchars(); break;

            case 201: DispCnt = 2; DisplayTaskBright(); break;
            case 202: DispCnt = 2; DisplayTaskSpeed(); break;
          } 
        }
      }
    }
    // the following line puts a rectangle around every display, showing the boundry region
    if (Border) {display.drawRect(0,0,128,64);}
  }
}

// --------------------------------------------------------------------------------

void Display_Init() {
  // initialise the OLED display
  display.init();
  display.clear();
  display.setContrast(255);
  display.flipScreenVertically();
  display.display();
  DispClr = false; DispDisp = false;
}

// --------------------------------------------------------------------------------

void Display_Intro() {
  // first screen to be displayed
  // default fonts are ArialMT_Plain_10, ArialMT_Plain_16 and ArialMT_Plain_24
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 6,24,4,24,6
  display.drawString(64, 6, "Matrix");
  display.drawString(64, 34, "Engine");
  display.display();
  DispClr = true;  // display intro, then clear memory ready for first screen
}

// --------------------------------------------------------------------------------

void Display_Locked() {
  // puts the display in locked mode  
  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  display.normalDisplay();
  display.drawString(64,20,"LOCKED");
  display.drawRect(0,0,128,64);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for return cycle
  DispNOP = true; DispLock = false; // set flags to stop I2C, and clear locking process
}

// --------------------------------------------------------------------------------

void Display_Menu() {
  // displays an active menu system  
  String zP; uint8_t zL; uint8_t zY = 0;
  display.clear();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  display.normalDisplay();
  MenuTL = MenuTS + 3; if (MenuTL > MenuT1) {MenuTL = MenuT1;}  // set the last item to be listed pointer
  for (zL = MenuTS;zL <= MenuTL;zL++) {
    switch (zL) {
      case   0: Display_MenuItem(zL,zY,0,"Main:"); break;
      case   1: Display_MenuItem(zL,zY,1,"Animations"); break;
      case   2: Display_MenuItem(zL,zY,1,"Settings"); break;
      
      case 100: Display_MenuItem(zL,zY,2,"Back"); break;
      case 101: Display_MenuItem(zL,zY,1,"Matrix Animation Editor"); break;
      case 102: Display_MenuItem(zL,zY,1,"Walking Creature"); break;
      case 103: Display_MenuItem(zL,zY,1,"Scroilling ForRnd"); break;
      case 104: Display_MenuItem(zL,zY,1,"Hey You!"); break;
      case 105: Display_MenuItem(zL,zY,1,"UTC carousel"); break;
      case 106: Display_MenuItem(zL,zY,1,"Star Sprites"); break;
      case 107: Display_MenuItem(zL,zY,1,"Boxed in"); break;
      case 108: Display_MenuItem(zL,zY,1,"Draw Rnd Boxes"); break;
      case 109: Display_MenuItem(zL,zY,1,"Draw Rnd Bars"); break;
      case 110: Display_MenuItem(zL,zY,1,"Draw Rnd Lines"); break;
      case 111: Display_MenuItem(zL,zY,1,"STEM Carriers Rewarding"); break;
      case 112: Display_MenuItem(zL,zY,1,"UTC Puts Steam IN STEM"); break;
      case 113: Display_MenuItem(zL,zY,1,"Think UTC for STEM!"); break;
      case 114: Display_MenuItem(zL,zY,1,"Three Sprites Rolling"); break;
      case 115: Display_MenuItem(zL,zY,1,"Blinking Head"); break;
      case 116: Display_MenuItem(zL,zY,1,"Face Sprites"); break;
      case 117: Display_MenuItem(zL,zY,1,"Climb The Ladder"); break;
      case 118: Display_MenuItem(zL,zY,1,"MAT..RIX colliding"); break;

      case 200: Display_MenuItem(zL,zY,2,"Back"); break;
      case 201: Display_MenuItem(zL,zY,1,"Brightness"); break;
      case 202: Display_MenuItem(zL,zY,1,"Speed"); break;
    } zY+= 16;
  }
  DispDisp = true;     // set display  & clear flag for next cycle
}

// --------------------------------------------------------------------------------

void Display_MenuItem(uint8_t zL,uint8_t zY,uint8_t zM, String zT) {
  // draw one line of a menu
  int zX = 10;
  if (zM == 0) {
    // top menu item, no indent
    zX = 0;
    if (MenuPnt == zL) {display.fillRect(0,zY+1,128,16); display.setColor(BLACK);}
  } else if ((zM == 1) && (MenuPnt == zL)){
    // normal menu item
    display.fillRect(0,zY+1,128,16); display.setColor(BLACK);
    display.drawString(0,zY, ">");
  } else if (zM == 2) {
    // top menu item, < back function
    if (MenuPnt == zL) {display.fillRect(0,zY+1,128,16); display.setColor(BLACK);}
    display.drawString(0,zY, "^");
  }
  display.drawString(zX,zY,zT);   // display the text
  display.setColor(WHITE);        // ensure display is normal
//  Serial.println(zT);
}

// --------------------------------------------------------------------------------

void DisplayTaskBright() {
  // called when in special display mode
  if (DispOnce) {
    // initialise task
    DispOnce = false; RotEncCnt = Bright; RotUpd = false;
    BrightLast = Bright;  // remember the current value, in case of cancel
  }
  if (DispSub < 1) {
    // displaying the value whilst the user turns the encoder
    if (RotUpd) {
      if (RotEncCnt > 255) {RotEncCnt = 255;}
      if (RotEncCnt < 4) {RotEncCnt = 4;}
      Bright = RotEncCnt; FastLED.setBrightness(Bright);
      RotUpd = false;
    }
    DisplayText2("Brightness",String(Bright));
    // check for switch action, we look for the falling edge
    if (!sw0State && DispSW0) {
      // SW0 has just gone LOW
      DispSub++; DispYN = false; RotEncCnt = 0;
    }
    DispSW0 = sw0State; // track the state of SW0
  } else {
    // user has pressed SW0 so accept or cancel, Y/N
    if (RotUpd) {
      if (RotEncCnt > 0) {DispYN = true;} else {DispYN = false;}
      RotUpd = false; RotEncCnt = 0;
    }
    DisplayText3YN("Set Brightness",String(Bright),DispYN);
    // check for switch action, we look for the falling edge
    if (!sw0State && DispSW0) {
      // SW0 has just gone LOW, so user has made a choice
      // if DispYN == NO then revert to previous brightness
      if (!DispYN) {Bright = BrightLast; FastLED.setBrightness(Bright);}
      else {BrightDef = Bright;}  // set as new default
      DispMode = 0; // return to menu display
    }
    DispSW0 = sw0State; // track the state of SW0
  }
}

// --------------------------------------------------------------------------------

void DisplayTaskSpeed() {
  // called when in special display mode
  unsigned long zSpd = 1000/FramePeriod;  // get speed in fps
  if (DispOnce) {
    // initialise task
    FramePeriodLast = FramePeriod;
    DispOnce = false; RotEncCnt = zSpd; RotUpd = false;
  }
  if (DispSub < 1) {
    // displaying the value whilst the user turns the encoder
    if (RotUpd) {
      if (RotEncCnt > 50) {RotEncCnt = 50;}
      if (RotEncCnt < 1) {RotEncCnt = 1;}
      FramePeriod = 1000/RotEncCnt; // set the new frame period
      RotUpd = false;
    }
    DisplayText2("Speed",String(1000/FramePeriod) + " fps");
    // check for switch action, we look for the falling edge
    if (!sw0State && DispSW0) {
      // SW0 has just gone LOW
      DispSub++; DispYN = false; RotEncCnt = 0;
    }
    DispSW0 = sw0State; // track the state of SW0
  } else {
    // user has pressed SW0 so accept or cancel, Y/N
    if (RotUpd) {
      if (RotEncCnt > 0) {DispYN = true;} else {DispYN = false;}
      RotUpd = false; RotEncCnt = 0;
    }
    DisplayText3YN("Set Speed",String(1000/FramePeriod) + " fps",DispYN);
    // check for switch action, we look for the falling edge
    if (!sw0State && DispSW0) {
      // SW0 has just gone LOW, so user has made a choice
      // if DispYN == NO then revert to previous brightness
      if (!DispYN) {FramePeriod = FramePeriodLast;}
      else {FramePeriodDef = FramePeriod;} // set as new speed (period) default
      DispMode = 0; // return to menu display
    }
    DispSW0 = sw0State; // track the state of SW0
  }
}

// --------------------------------------------------------------------------------

void Display_TESTchars() {
  // displays the printable charcacters in TEST mode
  // display is only updated when the rotary encoder is turned
  if (!RotUpd) {DispCnt = 1; return;}

  String zT;
  RotUpd = false;
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  if (RotEncCnt < 0) {RotEncCnt = 0;}
  if (RotEncCnt > 252) {RotEncCnt = 252;}
  display.drawString(62, 0, String(RotEncCnt) + " =");
  display.drawString(62,16, String(RotEncCnt + 1) + " =");
  display.drawString(62,32, String(RotEncCnt + 2) + " =");
  display.drawString(62,48, String(RotEncCnt + 3) + " =");
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  zT = char(RotEncCnt); display.drawString(66, 0,zT);
  zT = char(RotEncCnt + 1); display.drawString(66,16,zT);
  zT = char(RotEncCnt + 2); display.drawString(66, 32, zT);
  zT = char(RotEncCnt + 3); display.drawString(66, 48, zT);
  DispDisp = true;     // set display clear flag for next cycle
}

// --------------------------------------------------------------------------------

void DisplayText2(String zT0,String zT1) {
  // display two lines of text of equal height
  display.clear();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 14,16,4,16,14
  display.drawString(64, 14, zT0);
  display.drawString(64, 34, zT1);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
}

// --------------------------------------------------------------------------------

void DisplayText2L1(String zT0,String zT1) {
  // display two lines of text, 1st line is large
  display.clear();
  display.setFont(ArialMT_Plain_24);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 8,24,8,16,8
  display.drawString(64,  8, zT0);
  display.setFont(ArialMT_Plain_16);
  display.drawString(64, 40, zT1);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
}

// --------------------------------------------------------------------------------

void DisplayText3(String zT0,String zT1,String zT2) {
  // display three lines of text of equal height, centre screen
  display.clear();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64, 14, zT0);
  display.drawString(64, 34, zT1);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
}

// --------------------------------------------------------------------------------

void DisplayText3YN(String zT0,String zT1,byte zM) {
  // display two lines of text of equal height, centred, plus YES/NO highlighted
  // if zM == 0 then 'NO' is highlighted
  // if zM == 1 then 'YES' is highlighted
  display.clear();
  display.setFont(ArialMT_Plain_16);
  display.setTextAlignment(TEXT_ALIGN_CENTER);
  // Y64 = 2,16,6,16,6,16,2
  display.drawString(64,  2, zT0);
  display.drawString(64, 24, zT1);
  display.setTextAlignment(TEXT_ALIGN_LEFT);
  if (zM == 0) {display.fillRect(0,46,41,16); display.setColor(BLACK);}
  display.drawString(10, 46, "NO"); display.setColor(WHITE);
  display.setTextAlignment(TEXT_ALIGN_RIGHT);
  if (zM == 1) {display.fillRect(86,46,41,16); display.setColor(BLACK);}
  display.drawString(123, 46, "YES"); display.setColor(WHITE);
  display.display();  // display this immediately
  DispClr = true;     // set display clear flag for next cycle
}

// --------------------------------------------------------------------------------
