
// ---------------------------------------------------------------------------------

void ClearAnimArray() {
  // clears the contents of the animation array
  for (int zL = 0; zL < AnimMax; zL++) {
    AnCmd[zL] = 0;  // animation commands
    AnD0[zL] = 0;   // animation data element 0
    AnD1[zL] = 0;   // animation data element 1
    AnD2[zL] = 0;   // animation data element 2
    AnStr[zL] = ""; // string elements
  }
  AnimMax = 0;
}

// ---------------------------------------------------------------------------------

void ClearAnimSchedule() {
  // reset the Anim scheduler
  for (int zS = 0; zS < SchDepth; zS++) {AnimSch[zS] = 0;}
  SchCnt = 0; SchPnt = -1;
}

// ---------------------------------------------------------------------------------

void ClearSpriteArray() {
  // reset the sprite values
  for (int zP = 0; zP < 8; zP++) {
    ClearSpriteImage(zP); // resets all pixels
    SpriteDiv[zP] = 1;  // divider values
    SpriteFN[zP] = -1;    // loaded file references
    SpriteInc[zP] = 0;    // increment to target values
    SpriteON[zP] = false; // ON/off state
    SpriteSX[zP] = 0;     // start X values
    SpriteSY[zP] = 0;     // start Y values
    SpriteTX[zP] = 0;     // target X value
    SpriteTY[zP] = 0;     // target Y value
    SpriteX[zP] = 0;      // X co-ordinate
    SpriteY[zP] = 0;      // Y co-ordinate
  }
}

// ---------------------------------------------------------------------------------

void ClearSpriteImage(int zS) {
  // clears sprite zS LED data
  for (int zP = 0; zP < 257; zP++) {
    switch(zS) {
      case 0: Sprite0[zP].setRGB(0,0,0); break;
      case 1: Sprite1[zP].setRGB(0,0,0); break;
      case 2: Sprite2[zP].setRGB(0,0,0); break;
      case 3: Sprite3[zP].setRGB(0,0,0); break;
      case 4: Sprite4[zP].setRGB(0,0,0); break;
      case 5: Sprite5[zP].setRGB(0,0,0); break;
      case 6: Sprite6[zP].setRGB(0,0,0); break;
      case 7: Sprite7[zP].setRGB(0,0,0); break;
    }
  }
}

// ---------------------------------------------------------------------------------

void ClearVar() {
  // set variable array to all zeros
  for (int zV = 0;zV < 26; zV++) {Var[zV] = 0;}
}

// ---------------------------------------------------------------------------------

void decodeKey(int zkeyVal) {
  // decodes zkeyVal and excutes commands on '.'
  // on receipt of an ACK we return an ACK to run link at high rate
  if (zkeyVal == CR) {return;}    // ignore carriage return
  if (zkeyVal == ACK) {  // return an ACK
//    Serial.println("ACK");
    ACK_Cnt++;  // track number of ACKs received when new USB connection is made
    Serial.print(char(ACK)); return;
  }
  
  keyChar = char(zkeyVal);
  if (AnimPacket) {decodeRxPkt(); return;}  // in Anim packet mode so decode on the fly
  
  if (keyChar == '.') {doCmd();}  // command terminator
  else {cmdRx+= keyChar;}         // extend the Rx string
}

// ---------------------------------------------------------------------------------

void decodeRxPkt() {
  // decodes the serial data as Anim packets on the fly
  // in this version we only assume one string will be sent in one of the data
  // elements. This is always stored in AnStr[], irrespective of which element it came from
  if (keyChar != DEL) {
    // not a delimeter so simply add it to the recevied message and clear counter
    // we keep adding characters until we receive a DEL
    cmdRx+= keyChar; DELcnt = 0;
  } else {
    // received a delimeter so either decode cmdRx or count delimeters
    if (DELcnt < 1) {
      // this is the 1st delimeter after data, so decode this element if it has size
      DELcnt++;           // increase delimeter count
      if (cmdRx.length() > 0) {
        // data element size is greater than 0 bytes so decode it
        // the first element is assumed to be a command, after that there are 3 data elements
             if (AnimDp == 0) {AnCmd[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,16);
        } else if (AnimDp == 1) {
          // this is the 1st data element
          if (cmdRx.substring(0,1).equals("$")) {
            AnD0[AnimMax] = 0;  // not data so clear array element
            AnStr[AnimMax] = cmdRx.substring(1);  // remove the $ character
//          } else {AnD0[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,16);}
          } else {AnD0[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,10);}  // signed decimal numnber
        } else if (AnimDp == 2) {
          // this is the 2nd data element
          if (cmdRx.substring(0,1).equals("$")) {
            AnD1[AnimMax] = 0;  // not data so clear array element
            AnStr[AnimMax] = cmdRx.substring(1);  // remove the $ character
//          } else {AnD1[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,16);}
          } else {AnD1[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,10);}  // signed decimal numnber
        } else if (AnimDp == 3) {
          // this is the 3rd data element
          if (cmdRx.substring(0,1).equals("$")) {
            AnD2[AnimMax] = 0;  // not data so clear array element
            AnStr[AnimMax] = cmdRx.substring(1);  // remove the $ character
//          } else {AnD2[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,16);}
          } else {AnD2[AnimMax] = (int)strtol(cmdRx.c_str(),NULL,10);}  // signed decimal numnber
        }
      }
      AnimDp++;           // increment the data pointer
      cmdRx = "";         // clear the streamed component
      Serial.print(".");  // send a 'dot'
    } else {
      // this is the second delimeter after data so exit Anim packet mode
      AnimPacket = false; // stop packet decoding
      DELcnt = 0;         // reset delimeter counter
      AnimMax++;          // increase the Anim array counter/pointer
      Serial.println();   // send CR
    }
  }
}

// ---------------------------------------------------------------------------------

void doCmd() {
  // a '.' has been received on the serial port so execute command if valid
  // Commands:
  //
  // AA.+++       - defines the start of an animation message packets
  // AC.          - clears the contents of the animation arrays
  // AE.          - signals end of Anim packet streaming, so report status
  // AH.          - PAUSE/PLAY current animation
  // AN.          - single STEP in animation mode
  // AP.          - PLAY an animation file, preloaded from the PC
  // ARnnnn.      - load and play animation file An0000.txt from SD
  // AS.          - STOP playing an animation, if one was playing
  // BVnn.        - set brightness to value nn.
  // FA.          - write the animation data to /Anim/filename.txt
  // FDn.         - ask micro to report contents of ListType n folder + root
  // FFtttt.      - set folder to text tttt
  // FMtttt.      - set FrameMode reference text tttt
  // FNtttt.      - set filename to text tttt
  // FTtttt.      - set FrameTitle reference text tttt
  // FW.          - write the LED matrix to /folder/filename.txt
  // FX.          - toggle SD file system, begin or close
  // MBrrggbbhh.  - set the matrix pen colour and show point hh
  // MCrrggbb.    - set the matrix pen colour
  // MNhhhh.      - set matrix LED pointer to hhhh and show it
  // MOrrggbbhh.  - set the matrix pen colour and point hh, but don't show it
  // MPhhhh.      - set matrix LED pointer to hhhh only
  // MU.          - update display
  // MX.          - clear matrix contents
  // SFnnnn.      - show frame nnnn stored on SD card.
  // SPss.        - set speed fps to value ss.
  // VBn.         - toggles verbose messaging ON/OFF, 1 = ON, 0 = OFF
  // ZAxxxx.      - delete animation file Anxxxx.txt
  // ZFxxxx.      - delete frame file Frxxxx.txt
  // ZSxxxx.      - delete sprite file Spxxxx.txt
  
  String zRxUC = cmdRx.substring(0,2); zRxUC.toUpperCase();
  cmdMode = zRxUC.charAt(0); cmdType = zRxUC.charAt(1);
  switch (cmdMode) {
    case 'A':
      switch(cmdType) {
        case 'A': // received start of Animpacket streaming command AA.
          AnimPacket = true; AnimDp = 0;
          Serial.print("AA.");
          break;
        case 'C': // clears the contents of the animation arrays
          setAnimEngine();  // reset the engine to stop it playing
          ClearAnimArray(); ClearAnimSchedule();
          Serial.println("AC. animation arrays cleared");
          break;
        case 'E': // signals end of Anim packet streaming, so report status
          Serial.println("\nAE. Received " + String(AnimMax) + " data packets");
          break;
        case 'H': // PAUSE/PLAY current animation
          if (AnimPAUSE) {Serial.println("PAUSE removed");} else {Serial.println("PAUSE applied");}
          AnimPAUSE = !AnimPAUSE;
          break;
        case 'N': // STEP causes an auto pause
          Serial.println("AN. auto PAUSE applied");
          AnimSTEP = true; AnimPAUSE = false; break;
        case 'P': // PLAY a preloaded animation file
          if (AnimMax < 1) {Serial.println("No animation script!"); return;}
          setAnimMode(1);
          AnimPLAY = true; // start the animation engine
          Serial.println("AP. PLAY a preloaded animation file");
          break;
        case 'R': // PLAY an animated file from SD card, if it exists
          ClearAnimArray();
          readSDAnim(cmdRx.substring(2));
          if (ErrFlg) {Serial.println("No animation script!"); return;}
          setAnimMode(1); AnimPLAY = true; // start the animation engine
          Serial.println("AR. PLAY animation file from SD");
          break;
        case 'S': // STOP playing an animation, if one was playing
          AnimSTOP = true;
          Serial.println("All animation STOPPED!");
          break;
          
      } break;
      
    case 'B':
      switch(cmdType) {
        case 'V':
          AnyInt = cmdRx.substring(2).toInt(); Bright = AnyInt;
          FastLED.setBrightness(Bright);  LEDshow = true;
//          Serial.println("BV" + String(Bright));
          break;
      } break;
      
    case 'F':
      switch(cmdType) {
        case 'A': // received a write the animation data to /Anim/filename.txt
          WriteAnimToSD();
          Serial.println("FA");
          break;
        case 'D': // received a folder list command
          Serial.println("LW1."); // switch list on Matrix app to console mode
          cmdVal = cmdRx.substring(2).toInt();
          switch(cmdVal) {
            case -1: listDir(SD,"/",0); break;  // list root folder
            case  0: listDir(SD,"/Anim",0); break;
            case  1: listDir(SD,"/Frames",0); break;
            case  2: listDir(SD,"/Sprite16",0); break;
            case  3: listDir(SD,"/Sprite08",0); break;
          }
          Serial.println("LW0."); // close the Matrix app to console mode
          if (cmdVal >= 0) {
            switch(cmdVal) {
              case  0: Serial.println("LT0."); break; // type /Anim
              case  1: Serial.println("LT1."); break; // type /Frames
              case  2: Serial.println("LT2."); break; // type /Sprite16
            }
            Serial.println("LWS."); // tell the Matrix app to sort the list
          }
          break;
        case 'F': // received filepath folder command FFtttt.
          Folder = cmdRx.substring(2);
          Serial.println("FF" + Folder);
          break;
        case 'M': // received FrameMode text command.
          FrameMode = cmdRx.substring(2);
          Serial.println("FM" + FrameMode);
          break;
        case 'N': // received filepath filename command FNtttt.
          Name = cmdRx.substring(2);
          Serial.println("FN" + Name);
          break;
        case 'T': // received FrameTitle text command.
          FrameTitle = cmdRx.substring(2);
          Serial.println("FT" + FrameTitle);
          break;
        case 'W': // received a write LED matrix to microSD command FW.
          WriteLedToSD();
          Serial.println("FW");
          break;
        case 'X': // received FX. command so close/begin file system
          Serial.println("LW1.");
          if (!SDin) {SD.end(); SDin = true; Serial.println("File system closed.");}
          else {
            initSD(); SDin = ErrSD; Serial.println("SD card mounted.");
          } Serial.println("LW0."); break;
      } break;
      
    case 'M':
      switch(cmdType) {
        case 'B': // recevied LED colour command MBrrggbbhh.
          ColRGB[0].r = (int)strtol(cmdRx.substring(2,4).c_str(),NULL,16);
          ColRGB[0].g = (int)strtol(cmdRx.substring(4,6).c_str(),NULL,16);
          ColRGB[0].b = (int)strtol(cmdRx.substring(6,8).c_str(),NULL,16);
          cmdVal  = (int)strtol(cmdRx.substring(8).c_str(),NULL,16);
          if (cmdVal >= NumLEDs) {break;} // prevent LED[] array pointer being overloaded
          LED[cmdVal] = ColRGB[0]; LEDshow = true;
//          Serial.println("MB" + String(ColRGB[0].r) + "," + String(ColRGB[0].g) + "," + String(ColRGB[0].b) + "," + String(cmdVal));
          break;
        case 'C': // recevied LED colour command MNrrggbb.
          ColRGB[0].r = (int)strtol(cmdRx.substring(2,4).c_str(),NULL,16);
          ColRGB[0].g = (int)strtol(cmdRx.substring(4,6).c_str(),NULL,16);
          ColRGB[0].b = (int)strtol(cmdRx.substring(6,8).c_str(),NULL,16);
//          Serial.println("MC" + String(ColRGB[0].r) + "," + String(ColRGB[0].g) + "," + String(ColRGB[0].b));
          break;
        case 'N': // recevied LED pointer command MNhhhh.
          // set this LED to the current colour and trigger a 'show'
          cmdVal  = (int)strtol(cmdRx.substring(2).c_str(),NULL,16);
          if (cmdVal >= NumLEDs) {break;} // prevent LED[] array pointer being overloaded
          LED[cmdVal] = ColRGB[0]; LEDshow = true;
//          Serial.println("MN" + String(cmdVal));
          break;
        case 'O': // recevied LED colour command MOrrggbbhh.
          ColRGB[0].r = (int)strtol(cmdRx.substring(2,4).c_str(),NULL,16);
          ColRGB[0].g = (int)strtol(cmdRx.substring(4,6).c_str(),NULL,16);
          ColRGB[0].b = (int)strtol(cmdRx.substring(6,8).c_str(),NULL,16);
          cmdVal  = (int)strtol(cmdRx.substring(8).c_str(),NULL,16);
          if (cmdVal >= NumLEDs) {break;} // prevent LED[] array pointer being overloaded
          LED[cmdVal] = ColRGB[0];
//          Serial.println("MB" + String(ColRGB[0].r) + "," + String(ColRGB[0].g) + "," + String(ColRGB[0].b) + "," + String(cmdVal));
          break;
        case 'P': // recevied LED pointer command MPhhhh.
          // set this LED to the current colour only, don't trigger a 'show'
          cmdVal  = (int)strtol(cmdRx.substring(2).c_str(),NULL,16);
          if (cmdVal >= NumLEDs) {break;} // prevent LED[] array pointer being overloaded
          LED[cmdVal] = ColRGB[0];
//          Serial.println("MP" + String(cmdVal));
          break;
        case 'U': // received a Matrix update command MU. so 'show' updates
          LEDshow = true; break;
        case 'X': // received clear matrix command MX. but don't 'show' the effect
          FastLED.clear();
//          Serial.println("MX");
          break;
      } break;
      
    case 'S':
      switch(cmdType) {
        case 'E': // received a show frame command SSnnnn.
          Folder = "/Sprite08"; Name = "/Fr" + cmdRx.substring(2);
          Gfx_Fill(1,768,0,0,0); readSDToLED(0); // load the image from SD into the frame store
          LEDshow = true; AnimBuild = false; break;
        case 'F': // received a show frame command SFnnnn.
          Folder = "/Frames"; Name = "/Fr" + cmdRx.substring(2);
          Gfx_Fill(1,768,0,0,0); readSDToLED(0); // load the image from SD into the frame store
          LEDshow = true; AnimBuild = false; break;
        case 'P': // received speed command SPss.
          AnyInt = cmdRx.substring(2).toInt();
          FramePeriod = 1000/AnyInt;  // convert fps to ms
//          Serial.println("SP" + String(AnyInt));
          break;
        case 'S': // received a show frame command SSnnnn.
          Folder = "/Sprite16"; Name = "/Fr" + cmdRx.substring(2);
          Gfx_Fill(1,768,0,0,0); readSDToLED(0); // load the image from SD into the frame store
          LEDshow = true; AnimBuild = false; break;
      } break;
      
    case 'V':
      switch(cmdType) {
        case 'B': // received Verbose ON/OFF.
          cmdVal = cmdRx.substring(2).toInt();
          if (cmdVal == 2) {
            if (RepON) {RepON = false; Verbose = false;}
            else {RepON = true; Verbose = true;} // toggle reporting
          } else {
            // set verbos mode
            if (cmdVal == 1) {Verbose = true;} else {Verbose = false;}
            RepON = true;
          } break;
      } break;
      
    case 'Z':
      switch(cmdType) {
        case 'A': // received delete Anim file
          cmdVal  = (int)strtol(cmdRx.substring(2).c_str(),NULL,10);
          Serial.println("LW1.");
          AnyStr = "/Anim/An" + getNum4Digits(cmdVal) + ".txt";
          Serial.println("Deleting: " + AnyStr);
          AnyErr = SD.remove(AnyStr);
          if (AnyErr != true) {Serial.println("File Deletinon Failed!");}
          Serial.println("Task complete"); break;
        case 'F': // received delete Frame file
          cmdVal  = (int)strtol(cmdRx.substring(2).c_str(),NULL,10);
          Serial.println("LW1.");
          AnyStr = "/Frames/Fr" + getNum4Digits(cmdVal) + ".txt";
          Serial.println("Deleting: " + AnyStr);
          AnyErr = SD.remove(AnyStr);
          if (AnyErr != true) {Serial.println("File Deletinon Failed!");}
          Serial.println("Task complete"); break;
        case 'S': // received delete Sprite16 file
          cmdVal  = (int)strtol(cmdRx.substring(2).c_str(),NULL,10);
          Serial.println("LW1.");
          AnyStr = "/Sprite16/Sp" + getNum4Digits(cmdVal) + ".txt";
          Serial.println("Deleting: " + AnyStr);
          AnyErr = SD.remove(AnyStr);
          if (AnyErr != true) {Serial.println("File Deletinon Failed!");}
          Serial.println("Task complete"); break;
      } break;
    default:
      Serial.println("Unknown cmd: " + cmdMode + cmdType + String(cmdVal));
      break;
  }
  // now reset the variables
  cmdMode = ' '; cmdType = ' '; cmdVal = 0; cmdRx = "";
}

// ---------------------------------------------------------------------------------

void extendCmdVal(int zVal) {
  // adds a new digit to the right-hand end of cmdVal
  cmdVal = abs(cmdVal * 10) + zVal;
}

// ----------------------------------------------------------------------

void getEndIf() {
  // searches for EndIf after a failed 'If...' statement
  // if more 'If' statements are encountered then they too must have 'EndIf's'
  AnimPnt++; int zIfCnt = 1;
//  Serial.println("Searching for EndIf...");
  while (AnimPnt < AnimMax) {
//    Serial.println(String(AnimPnt) + "  " + String(zIfCnt));
    if ((AnCmd[AnimPnt] >= 0x4948) && (AnCmd[AnimPnt] <= 0x4965)) {zIfCnt++;} // 'If' commands encountered
    if (AnCmd[AnimPnt] == 0x4586) {zIfCnt--;} // EndIf - terminates 'If..' block
    if (zIfCnt == 0) {AnimPnt++; break;}      // closing EndIf found
    AnimPnt++;
  }
  if (AnimPnt == AnimMax) {AnimErr = 15;} // no 'EndIf' found
}

// ----------------------------------------------------------------------

String getHex2(long zN) {
  // return a Hex string for zN as HH
  Hex = String(zN,HEX); Hex.toUpperCase();
  // add leading zero
  while (Hex.length() < 2) {Hex = "0" + Hex;}
  return Hex;
}

// ----------------------------------------------------------------------

String getHex4(long zN) {
  // return a Hex string for zN as either HH or HHHH
  Hex = String(zN,HEX); Hex.toUpperCase();
  // add leading zero
  if (zN < 256) {while (Hex.length() < 2) {Hex = "0" + Hex;}}
  else {while (Hex.length() < 4) {Hex = "0" + Hex;}}
  return Hex;
}

// ----------------------------------------------------------------------

String getHex6(long zN) {
  // return a Hex string for zN as six characters HHHHHH
  Hex = String(zN,HEX); Hex.toUpperCase();
  // add leading zero
  while (Hex.length() < 6) {Hex = "0" + Hex;}
  return Hex;
}

// ----------------------------------------------------------------------

String getHexRGB(int zR,int zG,int zB) {
  // return a Hex string for zRzGzB
  Hex = getHex2(zR) + getHex2(zG) + getHex2(zB);
  return Hex;
}

// ----------------------------------------------------------------------

String getNum4Digits(int zN) {
  // return a 4-digit string representing zN
  String zT = String(zN);
       if (zN <   10) {zT = "000" + zT;}
  else if (zN <  100) {zT = "00" + zT;}
  else if (zN < 1000) {zT = "0" + zT;}
  return zT;
}

// --------------------------------------------------------------------------------

void getRGBcols(int32_t zRGB) {
  // extracts the three colours from a 4 byte integer
  Pen_Red = uint8_t((zRGB & 0xFF0000)>>16);
  Pen_Grn = uint8_t((zRGB & 0x00FF00)>> 8);
  Pen_Blu = uint8_t(zRGB & 0x0000FF);
  ColRGB[0].setRGB(Pen_Red,Pen_Grn,Pen_Blu);
}

// --------------------------------------------------------------------------------

void getTxtRGB(int32_t zRGB) {
  // extracts the three colours from a 4 byte integer
  Txt_Red = uint8_t((zRGB & 0xFF0000)>>16);
  Txt_Grn = uint8_t((zRGB & 0x00FF00)>> 8);
  Txt_Blu = uint8_t(zRGB & 0x0000FF);
}

// --------------------------------------------------------------------------------

void getWord() {
  // a Word has not yet been extracted from Text, so extract one
  Word = Text; Word.trim(); WordRht = "";
  if (Word.length() > 0) {
    // find the first word by looking for a space after it or is it
    // a ~ word space character?
    WordNo = Word.indexOf("~");
    WordSp = Word.indexOf(" ");
    if ((WordSp > 0) && (WordNo >= 0)) {
      // test contains both spaces and ~
      if (WordSp < WordNo) {
        // space comes before ~
        WordRht = Word.substring(WordSp); Word = Word.substring(0,WordSp);
      } else {
        // ~ comes before space
        if (WordNo == 0) {
          // ~ is first, and possibly only charcater, so we use a null word
          if (Word.length() <2) {WordRht = "";} else {WordRht = Word.substring(1);}
          Word = " ";
        } else {
          // ~ is not first character so we use it as a space
          WordRht = Word.substring(WordNo); Word = Word.substring(0,WordNo);
        }
      }
    }
    else if (WordSp > 0) {
      // only spaces no ~
      WordRht = Word.substring(WordSp); Word = Word.substring(0,WordSp);
    }
    else if (WordNo >= 0) {
      // no spaces, use ~ as the delimeter
      if (WordNo == 0) {
        // ~ is the first character
        Word = " "; WordRht = Word.substring(1);
      } else {
        // ~ is embedded
        WordRht = Word.substring(WordNo); Word = Word.substring(0,WordNo);
      }
    }
    else {WordRht = "";}  // only one word in Text
  } else {AnimErr = 14;}  // Text did not contain a valid Word
}

// --------------------------------------------------------------------------------

void getWordNext() {
  // WordRht contains text, so we extract the next word
  WordRht.trim();
  WordNo = WordRht.indexOf("~");
  WordSp = WordRht.indexOf(" ");
  if ((WordSp > 0) && (WordNo >= 0)) {
    // test contains both spaces and ~
    if (WordSp < WordNo) {
      // space comes before ~
      Word = WordRht.substring(0,WordSp); WordRht = WordRht.substring(WordSp);
    } else {
      // ~ comes before space
      if (WordNo == 0) {
        // ~ is first, and possibly only character, so we use a null word
        if (WordRht.length() <2) {WordRht = "";} else {WordRht = WordRht.substring(1);}
        Word = " ";
      } else {
        // ~ is not first character so we use it as a delimeter
        Word = WordRht.substring(0,WordNo); WordRht = WordRht.substring(WordNo);
      }
    }
  }
  else if (WordSp > 0) {
    // only spaces no ~
    Word = WordRht.substring(0,WordSp);WordRht = WordRht.substring(WordSp);
  }
  else if (WordNo >= 0) {
    // no spaces, use ~ as the delimeter
    if (WordNo == 0) {
      // ~ is the first character
      Word = " "; WordRht = WordRht.substring(1);
    } else {
      // ~ is embedded
      Word = WordRht.substring(0,WordNo); WordRht = WordRht.substring(WordNo);
    }
  }
  else {Word = WordRht; WordRht = "";}
}

// --------------------------------------------------------------------------------

void MenuCheck() {
  // called from the main loop every 10ms whenever an encoder update occurs
  // in menu mode the rotary encoder count is used to adjust the menu pointer
  // then a display update is triggered to show this change
  if (DispMode != 0) {
    // not in menu mode, so do a special task, like speed adjustment
    return;
  }

  // in menu mode so adjust the menu heighlight
  if (RotEncCnt < 0) {
    // going up the menu
    if (MenuPnt < MenuT1) {MenuPnt++;}
    if (MenuTS < (MenuPnt - 3)) {MenuTS = MenuPnt - 3;}
  }
  if (RotEncCnt > 0) {
    // going down the menu
    if (MenuPnt > MenuT0) {MenuPnt--;}
    if (MenuTS > MenuPnt) {MenuTS = MenuPnt;}
  }
  RotEncCnt = 0; RotUpd = false;  // clear rotary encoder flags
  DispCnt = 0;                    // initiate a display update
}

// --------------------------------------------------------------------------------

void MenuFunc() {
  // called when rotary encoder switch is pressed to perform menu actions
  // if the display is not in menu mode then switching to that mode is the action
  if (DispMode != 0) {DispMode = 0; return;}  // switch to menu screen
  
  // the value of MenuPnt is used as a selection pointer. Its value is changed by
  // movement of the rotary encoder, and it is limited to a menu range, which is
  // set by a call to MenuSet() for a given level
  switch(MenuPnt) {
    case   0: break; // Top level menu
    case   1: MenuSet(100); break;  // Animations
    case   2: MenuSet(200); break;  // Settings
    
    // Animations menu
    case 100: MenuSet(0); MenuPnt = 1; break; // ^ Back, return to Main menu/Animations
    case 101: RunAnimFile("0010"); break;     // Matrix Animation Editor
    case 102: RunAnimFile("0020"); break;     // Walking Creature
    case 103: RunAnimFile("0030"); break;     // Scroilling ForRnd
    case 104: RunAnimFile("0040"); break;     // Hey You!
    case 105: RunAnimFile("0050"); break;     // UTC carousel
    case 106: RunAnimFile("0060"); break;     // Star Sprites
    case 107: RunAnimFile("0070"); break;     // Boxed in
    case 108: RunAnimFile("0072"); break;     // Draw Rnd Boxes
    case 109: RunAnimFile("0074"); break;     // Draw Rnd Bars
    case 110: RunAnimFile("0076"); break;     // Draw Rnd Lines
    case 111: RunAnimFile("0080"); break;     // STEM Carriers Rewarding
    case 112: RunAnimFile("0100"); break;     // UTC Puts Steam IN STEM
    case 113: RunAnimFile("0110"); break;     // Think UTC for STEM!
    case 114: RunAnimFile("0120"); break;     // Three Sprites Rolling
    case 115: RunAnimFile("0130"); break;     // Blinking Head
    case 116: RunAnimFile("0150"); break;     // Face Sprites
    case 117: RunAnimFile("0160"); break;     // Climb The Carrier Ladder
    case 118: RunAnimFile("0140"); break;     // MAT..RIX colliding

    // Settings
    case 200: MenuSet(0); MenuPnt = 2; break; // ^ Back, return to Main menu/Settings
    case 201: setDispTask(201); break;       // Adjust brightness
    case 202: setDispTask(202); break;       // Adjust speed
  }
}

// --------------------------------------------------------------------------------

void MenuSet(uint8_t zL) {
  // called to initialise a menu structure at a given level
  // we clear the rotary encode and update flag when the menu changes
  // menus can have a range of elements, 1 - n
  // here we list the elements and assign a menu range MenuT0 to MenuT1
  RotEncCnt = 0; RotUpd = false;
  MenuLevel = zL; MenuPnt = 0;
  switch(zL) {
    case   0: 
      // there are 3 items in this top level menu
      MenuT0 = 0; MenuT1 = 2; break; // Top level menu
      // Main menu
      // Animations
      // Settings
    case 100:
      // there are 119 items in the Animation sub-menu
      MenuT0 = 100; MenuT1 = 118; break; // Animations
      // ^ Back
      // Matrix Animation Editor
      // Walking Creature
      // Scroilling ForRnd
      // Hey You!
      // UTC carousel
      // Star Sprites
      // Boxed in
      // Draw Rnd Boxes
      // Draw Rnd Bars
      // Draw Rnd Lines
      // STEM Carriers Rewarding
      // UTC Puts Steam IN STEM
      // Think UTC for STEM!
      // Three Sprites Rolling
      // Blinking Head
      // Face Sprites
      // Clinb The Ladder
      // MAT..RIX colliding
    
    case 200:
      MenuT0 = 200; MenuT1 = 202; break; // Settings
      // ^ Back
      // Brightness
      // Speed
  }
  MenuPnt = MenuT0; // point at first item
  MenuTS = MenuT0;  // set the menu list from pointer
}

// --------------------------------------------------------------------------------

void readSDAnim(String zAnimNum) {
  // load animation file An????.txt from SD
  // if successful ErrFlg == false
  uint8_t SdBite;
  String zTemp;
  String AnimPath = "/Anim/An" + zAnimNum + ".txt"; // construct filepath
  if (Verbose) {Serial.println("Read Filepath = " +  AnimPath);}
  else {Serial.println("Loading: " +  AnimPath);}
  ErrFlg = true; File file = SD.open(AnimPath);
  if(!file){Serial.println("Failed to open file for reading"); return;}

//  if (Verbose) {Serial.println("Reading Anim file: ");}
  // if is assumed that the array is cleared before calling this function
  // data is stored as follows, representing each line in the animation arrays
  // AA.hhhh  - command as hex
  // hhhh     - value of AnD0[], 1st data column
  // hhhh     - value of AnD1[], 2nd data column
  // hhhh     - value of AnD2[], 3rd data column
  // $string  - line string value if one exists
  int zValPnt = 0;
  while(file.available()){
    SdBite = file.read();
    if (SdBite != LF) {zTemp+= char(SdBite);}
    else {
      // string of characters obtained so decode them in sequence
//      Serial.print(zTemp + " ");
      switch (zValPnt) {
        case  0: // look for line header, if not seen ignore the rest until we see one
          if (zTemp.substring(0,3).equals("AA.")) {
            // line header AA. has been detected
            AnCmd[AnimMax] = (int)strtol(zTemp.substring(3).c_str(),NULL,16);
            zValPnt++;
          } zTemp = ""; break;
        case 1: // header was found, read data column 1
          AnD0[AnimMax] = (int)strtol(zTemp.c_str(),NULL,10);  // signed decimal numnber
          zValPnt++; zTemp = ""; break;
        case 2: // read data column 2
          AnD1[AnimMax] = (int)strtol(zTemp.c_str(),NULL,10);  // signed decimal numnber
          zValPnt++; zTemp = ""; break;
        case 3: // read data column 3
          AnD2[AnimMax] = (int)strtol(zTemp.c_str(),NULL,10);  // signed decimal numnber
          zValPnt++; zTemp = ""; break;
        case 4: // last value is a string, if it exists after the '$' header
          if (zTemp.length() > 1) { // it looks like something is there
            AnStr[AnimMax] = zTemp.substring(1);
          } else {AnStr[AnimMax] = "";}
          // line read so increment pointers
//          Serial.println("");
          zValPnt = 0; AnimMax++; zTemp = ""; break;
      }
    }
  }
  file.close(); ErrFlg = false;
//  Serial.println("AnimMax = " + String(AnimMax));
}

// --------------------------------------------------------------------------------

void readSDToLED(int zLvl) {
  // reads bytes of data from a file on the SD and place them in a FastLED buffer
  // this uses an optimised hexadecimal coding method to speed up image loading
  uint8_t SdBite;         // bite read from SD
  uint8_t SdBiteLast;     // previous SdBite
  int PixPnt = 0;         // pixel pointer
  uint8_t zDataType = 0;  // type of data received, 0 = RGB, 1 = pixel
  String zTemp;           // string accumulated from SdBite
  String FilePath = Folder + Name + ".txt"; // construct filepath /Folder/Name.txt
//  if (Verbose) {Serial.println("Read Filepath = " +  FilePath);}
  File file = SD.open(FilePath);
  if(!file){Serial.println("Failed to open file for reading"); return;}

//  if (Verbose) {Serial.print("Reading file: ");}
  // data is stored as irrggbb or ohh - rhhhh pixel pointers, hex data after 2-byte headers
  // there are no delimeters like LF or DEL
  
  while(file.available()){
    SdBite = file.read();
    if (SdBite > 70) {  // this is a command value so decode it
      zDataType = 0;  // assume colour command
      switch(SdBite) {
        case 'i': // ink colour command, irrggbb
          // get the red colour
          zTemp = char(file.read()); zTemp+= char(file.read());
          MemRGB[0].r = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.print("i" + zTemp);
          // get the green colour
          zTemp = char(file.read()); zTemp+= char(file.read());
          MemRGB[0].g = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.print(zTemp);
          // get the blue colour
          zTemp = char(file.read()); zTemp+= char(file.read());
          MemRGB[0].b = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.println(zTemp);
          break;
        case 'o': // single hex digit pixel, oh
          zTemp = char(file.read());
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16); zDataType = 1;
//          Serial.println("o" + zTemp);
          break;
        case 'p': // 2-digit hex digit pixel, phh
          zTemp = char(file.read()); zTemp+= char(file.read());
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16); zDataType = 1;
//          Serial.println("p" + zTemp);
          break;
        case 'q': // 3-digit hex digit pixel, phhh
          zTemp = char(file.read()); zTemp+= char(file.read()); zTemp+= char(file.read());
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16); zDataType = 1;
//          Serial.println("q" + zTemp);
          break;
        case 'r': // 4-digit hex digit pixel, phhhh
          zTemp = char(file.read()); zTemp+= char(file.read());
          zTemp+= char(file.read()); zTemp+= char(file.read());
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16); zDataType = 1;
//          Serial.println("r" + zTemp);
          break;
      } SdBiteLast = SdBite;
      // now set the pixel in the corresponding zLvl array
      if ((zDataType > 0) && (PixPnt < NumLEDs)) {
        switch (zLvl) {
          case 0:    LED[PixPnt] = MemRGB[0]; break;  // LED frame is at level 0
          case 1: Forgnd[PixPnt] = MemRGB[0]; break;  // Forgnd frame is at level 1
          case 2: Bckgnd[PixPnt] = MemRGB[0]; break;  // Bckgnd frame is at level 2
          case 3: // Sprite frames are at level 3
            switch (SpritePnt) {
              case 0: Sprite0[PixPnt] = MemRGB[0]; break;
              case 1: Sprite1[PixPnt] = MemRGB[0]; break;
              case 2: Sprite2[PixPnt] = MemRGB[0]; break;
              case 3: Sprite3[PixPnt] = MemRGB[0]; break;
              case 4: Sprite4[PixPnt] = MemRGB[0]; break;
              case 5: Sprite5[PixPnt] = MemRGB[0]; break;
              case 6: Sprite6[PixPnt] = MemRGB[0]; break;
              case 7: Sprite7[PixPnt] = MemRGB[0]; break;
            }
        }
      }
    } else {  // this is a hex value so continue reading pixels in the current range
      // the first digit has already been loaded
      zTemp = char(SdBite);
      switch (SdBiteLast) {
        case 'o': // single hex digit pixel, oh
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.println(zTemp);
          break;
        case 'p': // 2-digit hex digit pixel, phh
          zTemp+= char(file.read()); PixPnt = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.println(zTemp);
          break;
        case 'q': // 3-digit hex digit pixel, phhh
          zTemp+= char(file.read()); zTemp+= char(file.read());
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.println(zTemp);
          break;
        case 'r': // 4-digit hex digit pixel, phhhh
          zTemp+= char(file.read()); zTemp+= char(file.read());
          zTemp+= char(file.read());
          PixPnt = (int)strtol(zTemp.c_str(),NULL,16);
//          Serial.println(zTemp);
          break;
      }
      // now set the pixel in the corresponding zLvl array
      if (PixPnt < NumLEDs) {
        switch (zLvl) {
          case 0:    LED[PixPnt] = MemRGB[0]; break;  // LED frame is at level 0
          case 1: Forgnd[PixPnt] = MemRGB[0]; break;  // Forgnd frame is at level 1
          case 2: Bckgnd[PixPnt] = MemRGB[0]; break;  // Bckgnd frame is at level 2
          case 3: // Sprite frames are at level 3
            switch (SpritePnt) {
              case 0: Sprite0[PixPnt] = MemRGB[0]; break;
              case 1: Sprite1[PixPnt] = MemRGB[0]; break;
              case 2: Sprite2[PixPnt] = MemRGB[0]; break;
              case 3: Sprite3[PixPnt] = MemRGB[0]; break;
              case 4: Sprite4[PixPnt] = MemRGB[0]; break;
              case 5: Sprite5[PixPnt] = MemRGB[0]; break;
              case 6: Sprite6[PixPnt] = MemRGB[0]; break;
              case 7: Sprite7[PixPnt] = MemRGB[0]; break;
            }
        }
      }
    }
  }
  file.close();
}

// --------------------------------------------------------------------------------

void readSerial() {
  // reads characters from the serial port and responds to commands
  // set UsbActive to lock OFF other functions
  keyVal = Serial.read();
  if (keyVal != -1) {decodeKey(keyVal); UsbActive = 100;}
}

// --------------------------------------------------------------------------------

void readSW0() {
  // read the button switch and respond accordingly, pressed == LOW
  // a button press will automatical drop the current task
  if (REBOOT) {return;}

  sw0State = digitalRead(sw0Pin); // record button state
  if (SW0_Nop) {return;}          // block this function whilst == true
  
  bool sw0Clr = false;
  if (sw0State == LOW) {
    //##############################################################################
    //
    // SW0 button is pressed down
    //
    //##############################################################################
//    if (FrameTask > 0) {waitFrameEnd(); return;}
    if (sw0LastState == HIGH) {
      // SW0 has just been pressed down
      Serial.println("SW0 LOW");
      LED[0].setRGB(0, 0, 12); if (FrameTask == 0) {LEDshow = true;}
      MenuFunc();
      sw0DwnTime = 0; // restart button down time counter
      sw0Cnt++;       // count on the falling edge
      sw0Timer = 0;   // reset the timer
    } else {
      sw0DwnTime++;   // track button pressed time
    }
    // if long press for >1.5 seconds then check to see if we are running a boot animation
    if (sw0DwnTime > 150) {
      if (BootAnim) {
        // we are running a boot animation so stop it
        AnimBuild = false; LEDshow = false; AnimMode = -1; BootAnim = false;
        while (digitalRead(sw0Pin) == LOW) {}
        ResetAnimEngines();
      }
    }
  } else {
    //##############################################################################
    //
    // SW0 button is released
    //
    //##############################################################################
    if (sw0LastState == LOW) {
      // SW0 has just been released
      Serial.println("SW0 HIGH");
      LED[0].setRGB(0, 0, 0); if (FrameTask == 0) {LEDshow = true;}
    }
    if (sw0Timer >= 50) {
      // button released for 500ms so assume valid button count
      sw0Cnt = 0; sw0Timer = 0; sw0DwnTime = 0;
    }
  }
  if (ESC0 && !sw0State) {return;}
  sw0LastState = sw0State; // record current state
  ESC0 = false;
  if (sw0Cnt > 0) {sw0Timer++;}
}

// ---------------------------------------------------------------------------------

void ResetAnimEngines() {
  // terminates all animation engines
//  Serial.println("ResetAnimEngines!");
  setAnimMode(-1);      // set default null task
  ClearAnimArray();     // clear animation array
  ClearAnimSchedule();  // clear scheduler array
  FastLED.clear(); FastLED.show(); // This sends the updated pixel color to the hardware.
//  setFrameTask(0);
  BootAnim = false;
}

// ---------------------------------------------------------------------------------

void RunAnimFile(String zNum) {
  // called from the menu system to run the selected animation file zNum
  if (AnimPLAY) {AnimSTOP = true; return;}  // stop the current animation
  
  ResetAnimEngines(); // prepare everything for running a file
  readSDAnim(zNum);   // loads /Anim/AnzNum.txt
  if (ErrFlg) {
    Serial.println("No animation script!");
  } else {
    setAnimMode(1); AnimPLAY = true; // start the animation engine
    Serial.println("Playing animation boot file from SD");
    BootAnim = true;  // flag running animation from boot up
    RepON = false;    // turn OFF any reporting when booting from SD
    Verbose = false;  // turn OFF verbose messaging
  }
}

// ---------------------------------------------------------------------------------

void SerialFlush() {
  // empry the serial input buffer
  keyVal = 0; while (keyVal != -1) {keyVal = Serial.read();}
}

// ---------------------------------------------------------------------------------

void setAnimEngine() {
  // set startup conditions for the animation engine
//  Serial.println("setAnimEngine()");
  AnimPnt = 0; AnimMode = 1; AnimPAUSE = false; AnimSTEP = false; AnimLoop = -1;
  AnimDel = 0; AnimErr = 0; AnimBuild = false; AnimFrame = -1; AnimSTOP = false;
  AnimPLAY = false; AnimWait = 0; LEDnoshow = false;
  BckGndFrame = -1; ForGndFrame = -1; IcOnce = true; LevelOld = 0;
  ForPnt = -1; GosubPnt = -1;
  if (!BootAnim) {RepON = true; Verbose = true;}
  ScrollTask = 0; // == 0 to cause scroll initialisation
  // clear LED buffers
  Gfx_Fill(0,PixWxPixH,0,0,0);  // sets LEDs S to F to R,G,B colour
  Gfx_FillB(0,PixWxPixH,0,0,0); // sets background LEDs S to F to R,G,B colour
  Gfx_FillF(0,PixWxPixH,0,0,0); // sets foreground LEDs S to F to R,G,B colour
  ClearSpriteArray();
  ClearVar();
}

// ---------------------------------------------------------------------------------

void setAnimMode(int16_t zM) {
  // called to set the frame building mode and reset anything that is in process
  // temporarily stop show events in Core 1
//  Serial.println("setAnimMode(" + String(zM) + ")");
  AnimLast = AnimMode;
  AnimPLAY = false; AnimBuild = false; LEDshow = false; AnimMode = -1; delay(5);
  // reset current mode
  if (AnimLast == 0) {FrameBuild = false; LEDshow = false; if (FrameTask != 0) {setFrameTask(0);}}
  if (AnimLast == 1) {AnimBuild = false; LEDshow = false; setAnimEngine();}
  
  // now set the new mode
  AnimMode = zM;
//  Serial.println("AnimMode " + String(AnimMode));
       if (AnimMode == 0) {
        if ((AnimLast != 0) && (FrameTask != 0)) {setFrameTask(0);}
        FrameBuild = true;}
  else if (AnimMode == 1) {setAnimEngine();}
}

// ---------------------------------------------------------------------------------

void setDispTask(int16_t zDT) {
  // called to initialise a special display task
  DispCnt = 2;        // set display update count
  DispOnce = true;    // set DispTask initial conditions if needed
  DispSub = 0;        // initialise sub task pointer
  DispSW0 = sw0State; // grab the state of SW0
  
  DispMode = zDT; // display the special screen
}

// ---------------------------------------------------------------------------------

void setFrameTask(int16_t zMT) {
  // change the frame task pointer
  // set default conditions for starting/ending all tasks
  AnimMode = 0;       // set mode to frame tasks
  Bright = BrightDef; // reset LED brightness level
  Col11 = 255;        // max colour
  Col34 = 192;        // 3/4 max colour
  Col12 = 128;        // 1/2 max colour
  Col14 = 64;         // 1/4 maxcolour
  FastLED.setBrightness(Bright);
  FrameBuild = false; // if == true then run a frame building task
  FrameCnt = 0;       // Frame task general counter/timer
  FrameDel = 0;       // Frame task delay counter, if > 0 holds off tasks
  FrameNext = 0;      // next FrameTask to be performed after a delay
  FramePeriod = 40;   // frame time period in ms, default = 40
  FrameSubTask = 0;   // reset sub task pointer
  LED_Blink = true;
  LED_Cnt = 0;
  LED_Col = 0;
  LED_Del = 0;
  LED_Inc = 1;
  LED_Num = 0;
  LED_Pnt = 0;
  LED_SubTask = 0;
  LED_Task = 0;
  LevelOld = 0;
  // if zMT == FrameTask then enter the default state after clearing the display
  // this is for use with the rotary encode and menu system, to stop the current task
  if (zMT == FrameTask) {zMT = 99;}
  FrameTask = abs(zMT);
  Serial.println("setFrameTask(" + String(FrameTask) + ")");
}

// --------------------------------------------------------------------------------

void WriteAnimToSD() {
  // writes the contents of the An??[] arrays to the SD card using preloaded data
  Serial.println("LW1."); // switch list on Matrix app to console mode
  String FilePath = Folder + Name + ".txt"; // construct filepath
  Serial.println("Filepath = " +  FilePath);
  if (!SD.exists(Folder.c_str())) {
    // if the folder does not exist, then create it
    createDir(SD,Folder.c_str()); Serial.println("Creating folder: " + Folder);
  }
  PrintEn = false;  // turn off verboswe messages
  String zData; int zNumLines = 0; String zError = "Write failed";
  // open file for writing
  File file = SD.open(FilePath.c_str(), FILE_WRITE);
  if(!file){
    Serial.println("Failed to open file for writing");
    return;
  }
  // check that there is some data in the animation arrays
  if (AnimMax < 1) {Serial.println("Failed - no Anim data!"); return;}
  
  // we do not bother writing AnimTitle data as it is never used here
  // we store the data in the same format as sent over the serial link
  int zDot = 0; // dot counter
  for (int zL = 0; zL < AnimMax;zL++) {
    zData = "AA." + getHex4(AnCmd[zL]) + "\n";
    if(!file.print(zData)){Serial.println(zError); return;}
    zData = String(AnD0[zL]) + "\n"; // send 1st data column as signed decimal numnber
    if(!file.print(zData)){Serial.println(zError); return;}
    zData = String(AnD1[zL]) + "\n"; // send 2nd data column as signed decimal numnber
    if(!file.print(zData)){Serial.println(zError); return;}
    zData = String(AnD2[zL]) + "\n"; // send 3rd data column as signed decimal numnber
    if(!file.print(zData)){Serial.println(zError); return;}
    zData = "$" + (AnStr[zL]) + "\n"; // send string data, may be null in most cases
    if(!file.print(zData)){Serial.println(zError); return;}

    zDot++; if (zDot >= 10) {Serial.println(".........."); zDot = 0;}
    zNumLines++;
  }
  file.flush();
  file.close();
  if (zDot > 0) {Serial.println(".....");}
  Serial.println("Stored " + String(zNumLines) + " lines");
  PrintEn = true;

  Serial.println("LW0."); // switch list on Matrix app from console mode
}

// --------------------------------------------------------------------------------

void WriteLedToSD() {
  // writes the contents of the LED matrix to the SD card using preloaded data
  Serial.println("LW1."); // switch list on Matrix app to console mode
  String FilePath = Folder + Name + ".txt"; // construct filepath
  Serial.println("Filepath = " +  FilePath);
  if (!SD.exists(Folder.c_str())) {
    // if the folder does not exist on the SD, then create it
    createDir(SD,Folder.c_str()); Serial.println("Creating folder: " + Folder);
  }
  PrintEn = false;  // turn off verboswe messages
  String zData; int zNumPix = 0; String zError = "Write failed";
  // open file for writing
  File file = SD.open(FilePath.c_str(), FILE_WRITE);
  if(!file){
    Serial.println("Failed to open file for writing");
    return;
  }
  // we do not bother writing FrameTitle and FrameMode data as it is never used here
  int zDot = 0; // dot counter
  int zW = 0;   // width of hex number in characters
  int zR = 0; int zG = 0; int zB = 0;  // temp colour reference
  for (int zL = 1; zL < NumLEDs;zL++) {
    if ((LED[zL].r > 0) || (LED[zL].g > 0) || (LED[zL].b > 0)) {
      // this LED is lit, but is it the same previous colour?
      if ((LED[zL].r == zR) && (LED[zL].g == zG) && (LED[zL].b == zB)) {
        // yes same colour so simply send point data
        // if the Hex value is of the same length we don't need to add a header
        zData = String(zL,HEX); zData.toUpperCase();
        if (zData.length() != zW) {
               if (zData.length() == 1) {zData = "o" + zData; zW = 1;}
          else if (zData.length() == 2) {zData = "p" + zData; zW = 2;}
          else if (zData.length() == 3) {zData = "q" + zData; zW = 3;}
          else if (zData.length() == 4) {zData = "r" + zData; zW = 4;}
        }
        if(!file.print(zData)){Serial.println(zError); return;}
//        Serial.println(zData);
      } else {
        // not the same as previous colour, so send new colour and point
        zData = "i" + getHexRGB(LED[zL].r,LED[zL].g,LED[zL].b);
        if(!file.print(zData)){Serial.println(zError); return;}
//        Serial.println(zData);
        // now send the pixel co-ordinate
        zData = String(zL,HEX); zData.toUpperCase();
             if (zData.length() == 1) {zData = "o" + zData; zW = 1;}
        else if (zData.length() == 2) {zData = "p" + zData; zW = 2;}
        else if (zData.length() == 3) {zData = "q" + zData; zW = 3;}
        else if (zData.length() == 4) {zData = "r" + zData; zW = 4;}
        if(!file.print(zData)){Serial.println(zError); return;}
//        Serial.println(zData);
        zR = LED[zL].r; zG = LED[zL].g; zB = LED[zL].b;
      }
      zNumPix++;
      zDot++; if (zDot >= 20) {Serial.println(".........."); zDot = 0;}
    }
  }
  file.flush();
  file.close();
  if (zDot > 0) {Serial.println("..........");}
  Serial.println("Stored " + String(zNumPix) + " pixels");
  PrintEn = true;

  Serial.println("LW0."); // switch list on Matrix app from console mode
}

// --------------------------------------------------------------------------------
