
/*
 Graphics functions for two serial 16x16 panels
 This tab contains a set of functions used in creating effects.
 The co-ordinate X,Y system is 0,0 at top left corner, 47,15 at bottom right.
 
 
 Gfx_Clear(S,F)                       - sets LEDs S to F to OFF
 Gfx_ClrXY(X,Y)                       - clears a pixel on the current face to the current colour
 Gfx_CopyInBck()                      - copies the background array into the LED array
 Gfx_CopyInBckWin(X0,Y0,X1,Y1,X2,zY2) - copies region zX0,zY0,zX1,zY1 in the background array into the LED frame array at zX2,zY2
 Gfx_CopyInFor()                      - copies the foreground array into the LED array
 Gfx_CopyInForWin(X0,Y0,X1,Y1,X2,zY2) - copies region zX0,zY0,zX1,zY1 in the foreground array into the LED frame array at zX2,zY2
 Gfx_DimBckgnd(S,F,D)                 - dim background LEDs in the range S to F by factor D
 Gfx_DimBckgndE2(S,F)                 - dim background LEDS in the range S to F by half
 Gfx_DimBckgndTo(S,F,D,R,G,B)         - dim background LEDs in the range S to F by factor D towards target R,G,B
 Gfx_DimForgnd(S,F,D)                 - dim foreground LEDs in the range S to F by factor D
 Gfx_DimForgndE2(S,F)                 - dim foreground LEDS in the range S to F by half
 Gfx_DimForgndTo(S,F,D,R,G,B)         - dim foreground LEDs in the range S to F by factor D towards target R,G,B
 Gfx_DimFrame(S,F,D)                  - dim LEDs in the range S to F by factor D
 Gfx_DimFrameE2(zS,zF)                - dim LEDS in the range S to F by half
 Gfx_DimFrameTo(S,F,D,R,G,B)          - dim LEDs in the range S to F by factor D towards target R,G,B
 Gfx_DrawBar(X,Y,W,H)                 - draws a filled box in the frame starting at X,Y with width W, height H
 Gfx_DrawBox(X,Y,W,H)                 - draws a box in the frame starting at X,Y with width W, height H
 Gfx_DrawChar(X,Y,C,M)                - draws a single character C at X,Y using mode M
 Gfx_DrawGC(X,Y,M)                    - draws a loaded graphics character at X,Y using mode M
 Gfx_DrawGCC(X,Y,C,R)                 - draws a column C of a loaded graphics character at X,Y with rotation R
 Gfx_DrawGFX(F,X,Y)                   - draws a loaded Gfx character at X,Y
 Gfx_DrawStr(T$,X,Y)                  - draws the text T$ starting at X,Y
 Gfx_DrawTxt(X,Y)                     - draws the characters in Text starting at X,Y
 Gfx_Fill(S,F,R,G,B)                  - sets Frame LEDs S to F to R,G,B colour
 Gfx_FillB(S,F,R,G,B)                 - sets background LEDs S to F to R,G,B colour
 Gfx_FillBReg(X0,Y0,X1,Y1,R,G,B)      - sets background LEDs in region X0,Y0 - X1,Y1 to R,G,B colour
 Gfx_FillF(S,F,R,G,B)                 - sets foreground LEDs S to F to R,G,B colour
 Gfx_FillFReg(X0,Y0,X1,Y1,R,G,B)      - sets foreground LEDs in region X0,Y0 - X1,Y1 to R,G,B colour
 Gfx_FillReg(X0,Y0,X1,Y1,R,G,B)       - sets Frame LEDs in region X0,Y0 - X1,Y1 to R,G,B colour
 Gfx_FillSp(Sp,S,F,R,G,B)             - sets sprite Sp LEDs S to F to R,G,B colour
 Gfx_FlipBckgndH(X0,Y0,X1,Y1)         - flip background region X0,Y0,X1,Y1 horizontally
 Gfx_FlipBckgndV(X0,Y0,X1,Y1)         - flip background region X0,Y0,X1,Y1 vertically
 Gfx_FlipForgndH(X0,Y0,X1,Y1)         - flip foreground region X0,Y0,X1,Y1 horizontally
 Gfx_FlipForgndV(X0,Y0,X1,Y1)         - flip foreground region X0,Y0,X1,Y1 vertically
 Gfx_FlipFrameH(X0,Y0,X1,Y1)          - flip frame region X0,Y0,X1,Y1 horizontally
 Gfx_FlipFrameV(X0,Y0,X1,Y1)          - flip frame region X0,Y0,X1,Y1 vertically
 Gfx_FlipGCRL()                       - flips all of the GC array bytes, bit by bit right to left
 Gfx_GetMatL(X,Y)                     - returns LED pointer L for X,Y for the whole frame
 Gfx_GetTextLen(T)                    - returns the length of string T in pixels
 Gfx_GetSpriteL(X,Y)                  - returns LED pointer L for X,Y for a 16x16 sprite
 Gfx_InvCol(R,G,B)                    - set LED colours to the inverse of R,G,B
 Gfx_Line(X0,Y0,X1,Y1)                - draw a line from X0,Y0 to X1,Y1
 Gfx_MemCol()                         - place the current RGB colours in a memory
 Gfx_MultCol(M)                       - multiply the ink colours by factor M
 Gfx_PicXY(X,Y)                       - gets a pixel colour from the current X,Y
 Gfx_RecCol()                         - recall the colours stored in a memory with Gfx_MemCol()
 Gfx_RndCol(M);                       - sets a random ink colour, max strength M
 Gfx_RotDwn()                         - rotate panel one pixel down
 Gfx_RotDwnBck()                      - rotate background one pixel down
 Gfx_RotDwnFor()                      - rotate foreground one pixel down
 Gfx_RotDwnWin(X0,Y0,X1,Y1)           - rotate panel window X0,Y0 to X1,Y1 one pixel down
 Gfx_RotDwnWinBck(X0,Y0,X1,Y1)        - rotate backgrouind window X0,Y0 to X1,Y1 one pixel down
 Gfx_RotDwnWinFor(X0,Y0,X1,Y1)        - rotate foregrouind window X0,Y0 to X1,Y1 one pixel down
 Gfx_RotGCL()                         - rotates the GC array bytes anti-clockwise, left
 Gfx_RotGCR()                         - rotates the GC array bytes clockwise, right
 Gfx_RotLft()                         - rotate panel one pixel to the left
 Gfx_RotLftBck()                      - rotate background one pixel to the left
 Gfx_RotLftFor()                      - rotate foreground one pixel to the left
 Gfx_RotLftWin(X0,Y0,X1,Y1)           - rotate panel window X0,Y0 to X1,Y1 one pixel to the left
 Gfx_RotLftWinBck(X0,Y0,X1,Y1)        - rotate background window X0,Y0 to X1,Y1 one pixel to the left
 Gfx_RotLftWinFor(X0,Y0,X1,Y1)        - rotate foreground window X0,Y0 to X1,Y1 one pixel to the left
 Gfx_RotRht()                         - rotate panel one pixel to the right
 Gfx_RotRhtBck()                      - rotate background one pixel to the right
 Gfx_RotRhtFor()                      - rotate foreground one pixel to the right
 Gfx_RotRhtWin(X0,Y0,X1,Y1)           - rotate panel window X0,Y0 to X1,Y1 one pixel to the right
 Gfx_RotRhtWinBck(X0,Y0,X1,Y1)        - rotate background window X0,Y0 to X1,Y1 one pixel to the right
 Gfx_RotRhtWinFor(X0,Y0,X1,Y1)        - rotate foreground window X0,Y0 to X1,Y1 one pixel to the right
 Gfx_RotUp()                          - rotate panel one pixel Up
 Gfx_RotUpBck()                       - rotate background one pixel Up
 Gfx_RotUpFor()                       - rotate foreground one pixel Up
 Gfx_RotUpWin(X0,Y0,X1,Y1)            - rotate panel window X0,Y0 to X1,Y1 one pixel to the up
 Gfx_RotUpWinBck(X0,Y0,X1,Y1)         - rotate background window X0,Y0 to X1,Y1 one pixel to the up
 Gfx_RotUpWinFor(X0,Y0,X1,Y1)         - rotate foreground window X0,Y0 to X1,Y1 one pixel to the up
 Gfx_ScrollTxt()                      - scroll Text with a window strip at Text_SY
 Gfx_SetBakXY(X,Y)                    - resets a pixel on the current face to the BAK colour
 Gfx_SetCol(C)                        - sets RGB colour variables from a palette based on C
 Gfx_SetColMax(M)                     - sets max palette colour 4 - 255
 Gfx_SetXY(X,Y)                       - sets a pixel to the current colour
 Gfx_SpriteFlipH(S,X0,Y0,X1,Y1)       - flip sprite S region X0,Y0,X1,Y1 horizontally
 Gfx_SpriteFlipV(S,X0,Y0,X1,Y1)       - flip sprite S region X0,Y0,X1,Y1 vertically
 Gfx_SpriteRotDwn(S)                  - rotate sprite S one pixel Down
 Gfx_SpriteRotLft(S)                  - rotate sprite S one pixel Left
 Gfx_SpriteRotUp(S)                   - rotate sprite S one pixel Up
 Gfx_SpriteRotRht(S)                  - rotate sprite S one pixel Right
 Gfx_SpriteLoad(S)                    - loads Sprite S into the output frame at SpriteS[X],SpriteS[Y]
 Gfx_StripRotUp()                     - rotate the entire contents of the strip from 1 to (NumLEDs - 1)
 Gfx_StripRotDwn()                    - rotate the entire contents of the strip from (NumLEDs - 1) to 1
 Gfx_TxtLft()                         - rotate text strip 1 pixel to the left
*/

// ---------------------------------------------------------------------------------

void Gfx_Clear(int zS,int zF) {
  // sets LEDs zS to zF to OFF
  for (int zP = zS;zP <= zF; zP++) {LED[zP].setRGB(0,0,0);}
}

// ---------------------------------------------------------------------------------

void Gfx_ClrXY(int zX, int zY) {
  // clears a pixel at zX,zY
  // co-ordinate system is 0,0 at top left corner, to Xmax,Ymax
  if ((zX < Xmin) || (zX > Xmax) || (zY < Ymin) || (zY > Ymax)) {return;}
  
  int zO = 0; int zY16 = zY/16;
  if (zY > 15) {zO = zY16 * 768; zY = zY - (zY16 * 16);}
  if ((zX % 2) > 0) {zY = 15 - zY;}   // column phase swap
  LED_PXY = zO + 1 + (16 * zX) + zY;  // determine pixel pointer
  LED[LED_PXY].setRGB(0,0,0);
}

// ---------------------------------------------------------------------------------

void Gfx_CopyInBck() {
  // copies the background array into the LED frame array
  // only LEDs that have colour are copied
  MemRGB[0].setRGB(0,0,0);
  for (int zP = 1;zP < NumLEDs; zP++) {
    if (Bckgnd[zP] != MemRGB[0]) {LED[zP] = Bckgnd[zP];}
  }
}

// ---------------------------------------------------------------------------------

void Gfx_CopyInBckWin(int zX0,int zY0,int zX1,int zY1,int zX2,int zY2) {
  // copies region zX0,zY0,zX1,zY1 in the background array into the LED frame array at zX2,zY2
  // only LEDs that have colour are copied
  int zT;
  // ensure source window co-ordinates are of the correct polarity
  if (zX1 < zX0) {zT = zX0; zX0 = zX1; zX1 = zT;}
  if (zY1 < zY0) {zT = zY0; zY0 = zY1; zY1 = zT;}
  MemRGB[0].setRGB(0,0,0);  // set the LED OFF test
  // test for single pixel
  if ((zX0 == zX1) && (zY0 == zY1)) {
    if (Bckgnd[Gfx_GetMatL(zX0,zY0)] != MemRGB[0]) {LED[Gfx_GetMatL(zX2,zY2)] = Bckgnd[Gfx_GetMatL(zX0,zY0)];}
    return;
  }
  // more than one pixel so copy row by row, working down the Y column
  int zYT = zY2;
  for (int zY = zY0;zY <= zY1;zY++) {
    int zXT = zX2;
    for (int zX = zX0;zX <= zX1;zX++) {
      int zLED = Gfx_GetMatL(zX,zY);
      if (Bckgnd[zLED] != MemRGB[0]) {LED[Gfx_GetMatL(zXT,zYT)] = Bckgnd[zLED];}
      zXT++;
    } zYT++;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_CopyInFor() {
  // copies the foreground array into the LED frame array
  MemRGB[0].setRGB(0,0,0);
  for (int zP = 1;zP < NumLEDs; zP++) {
    if (Forgnd[zP] != MemRGB[0]) {LED[zP] = Forgnd[zP];}
  }
}

// ---------------------------------------------------------------------------------

void Gfx_CopyInForWin(int zX0,int zY0,int zX1,int zY1,int zX2,int zY2) {
  // copies region zX0,zY0,zX1,zY1 in the foreground array into the LED frame array at zX2,zY2
  // only LEDs that have colour are copied
  int zT;
  // ensure source window co-ordinates are of the correct polarity
  if (zX1 < zX0) {zT = zX0; zX0 = zX1; zX1 = zT;}
  if (zY1 < zY0) {zT = zY0; zY0 = zY1; zY1 = zT;}
  MemRGB[0].setRGB(0,0,0);  // set the LED OFF test
  // test for single pixel
  if ((zX0 == zX1) && (zY0 == zY1)) {
    if (Forgnd[Gfx_GetMatL(zX0,zY0)] != MemRGB[0]) {LED[Gfx_GetMatL(zX2,zY2)] = Forgnd[Gfx_GetMatL(zX0,zY0)];}
    return;
  }
  // more than one pixel so copy row by row, working down the Y column
  int zYT = zY2;
  for (int zY = zY0;zY <= zY1;zY++) {
    int zXT = zX2;
    for (int zX = zX0;zX <= zX1;zX++) {
      int zLED = Gfx_GetMatL(zX,zY);
      if (Forgnd[zLED] != MemRGB[0]) {LED[Gfx_GetMatL(zXT,zYT)] = Forgnd[zLED];}
      zXT++;
    } zYT++;
  }
}

// ---------------------------------------------------------------------------------

bool Gfx_DimBckgnd(int zS,int zF,uint8_t zD) {
  // reduces the brightness of each background LED in the range zS to zF by zD
  // returns zDim = true if one or more pixels were dimmed
  bool zDim = false;
  for (int zP = zS;zP <= zF; zP++) {
    if (Bckgnd[zP].r > zD) {Bckgnd[zP].r -= zD; zDim = true;} else {Bckgnd[zP].r = 0;}
    if (Bckgnd[zP].g > zD) {Bckgnd[zP].g -= zD; zDim = true;} else {Bckgnd[zP].g = 0;}
    if (Bckgnd[zP].b > zD) {Bckgnd[zP].b -= zD; zDim = true;} else {Bckgnd[zP].b = 0;}
  }
  return zDim;
}

// ---------------------------------------------------------------------------------

bool Gfx_DimBckgndE2(int zS,int zF) {
  // reduces the brightness of each background LED in the range zS to zF half
  // returns zDim = true if one or more pixels were dimmed
  bool zDim = false;
  for (int zP = zS;zP <= zF; zP++) {
    if (Bckgnd[zP].r > 1) {Bckgnd[zP].r = Bckgnd[zP].r>>1; zDim = true;} else {Bckgnd[zP].r = 0;}
    if (Bckgnd[zP].g > 1) {Bckgnd[zP].g = Bckgnd[zP].g>>1; zDim = true;} else {Bckgnd[zP].g = 0;}
    if (Bckgnd[zP].b > 1) {Bckgnd[zP].b = Bckgnd[zP].b>>1; zDim = true;} else {Bckgnd[zP].b = 0;}
  }
  return zDim;
}

// ---------------------------------------------------------------------------------

void Gfx_DimBckgndTo(int zS,int zF,uint8_t zD,uint8_t zR,uint8_t zG,uint8_t zB) {
  // dim background LEDs in the range S to F by factor D towards target R,G,B
  // target can be brighter, hence dim up/down
  for (int zP = zS;zP <= zF; zP++) {
    if (Bckgnd[zP].r > zR) {
      if (zD < (Bckgnd[zP].r - zR)) {Bckgnd[zP].r -= zD;} else {Bckgnd[zP].r = zR;}
    } else if (Bckgnd[zP].r < zR) {
      if (zD < (zR - Bckgnd[zP].r)) {Bckgnd[zP].r += zD;} else {Bckgnd[zP].r = zR;}
    }
    if (Bckgnd[zP].g > zG) {
      if (zD < (Bckgnd[zP].g - zG)) {Bckgnd[zP].g -= zD;} else {Bckgnd[zP].g = zG;}
    } else if (Bckgnd[zP].g < zG) {
      if (zD < (zG - Bckgnd[zP].g)) {Bckgnd[zP].g += zD;} else {Bckgnd[zP].g = zG;}
    }
    if (Bckgnd[zP].r > zB) {
      if (zD < (Bckgnd[zP].b - zB)) {Bckgnd[zP].b -= zD;} else {Bckgnd[zP].b = zB;}
    } else if (Bckgnd[zP].b < zB) {
      if (zD < (zB - Bckgnd[zP].b)) {Bckgnd[zP].b += zD;} else {Bckgnd[zP].b = zB;}
    }
  }
}

// ---------------------------------------------------------------------------------

bool Gfx_DimForgnd(int zS,int zF,uint8_t zD) {
  // reduces the brightness of each foreground LED in the range zS to zF by zD
  // returns zDim = true if one or more pixels were dimmed
  bool zDim = false;
  for (int zP = zS;zP <= zF; zP++) {
    if (Forgnd[zP].r > zD) {Forgnd[zP].r -= zD; zDim = true;} else {Forgnd[zP].r = 0;}
    if (Forgnd[zP].g > zD) {Forgnd[zP].g -= zD; zDim = true;} else {Forgnd[zP].g = 0;}
    if (Forgnd[zP].b > zD) {Forgnd[zP].b -= zD; zDim = true;} else {Forgnd[zP].b = 0;}
  }
  return zDim;
}

// ---------------------------------------------------------------------------------

bool Gfx_DimForgndE2(int zS,int zF) {
  // reduces the brightness of each foreground LED in the range zS to zF half
  // returns zDim = true if one or more pixels were dimmed
  bool zDim = false;
  for (int zP = zS;zP <= zF; zP++) {
    if (Forgnd[zP].r > 1) {Forgnd[zP].r = Forgnd[zP].r>>1; zDim = true;} else {Forgnd[zP].r = 0;}
    if (Forgnd[zP].g > 1) {Forgnd[zP].g = Forgnd[zP].g>>1; zDim = true;} else {Forgnd[zP].g = 0;}
    if (Forgnd[zP].b > 1) {Forgnd[zP].b = Forgnd[zP].b>>1; zDim = true;} else {Forgnd[zP].b = 0;}
  }
  return zDim;
}

// ---------------------------------------------------------------------------------

void Gfx_DimForgndTo(int zS,int zF,uint8_t zD,uint8_t zR,uint8_t zG,uint8_t zB) {
  // dim foreground LEDs in the range S to F by factor D towards target R,G,B
  // target can be brighter, hence dim up/down
  for (int zP = zS;zP <= zF; zP++) {
    if (Forgnd[zP].r > zR) {
      if (zD < (Forgnd[zP].r - zR)) {Forgnd[zP].r -= zD;} else {Forgnd[zP].r = zR;}
    } else if (Forgnd[zP].r < zR) {
      if (zD < (zR - Forgnd[zP].r)) {Forgnd[zP].r += zD;} else {Forgnd[zP].r = zR;}
    }
    if (Forgnd[zP].g > zG) {
      if (zD < (Forgnd[zP].g - zG)) {Forgnd[zP].g -= zD;} else {Forgnd[zP].g = zG;}
    } else if (Forgnd[zP].g < zG) {
      if (zD < (zG - Forgnd[zP].g)) {Forgnd[zP].g += zD;} else {Forgnd[zP].g = zG;}
    }
    if (Forgnd[zP].r > zB) {
      if (zD < (Forgnd[zP].b - zB)) {Forgnd[zP].b -= zD;} else {Forgnd[zP].b = zB;}
    } else if (Forgnd[zP].b < zB) {
      if (zD < (zB - Forgnd[zP].b)) {Forgnd[zP].b += zD;} else {Forgnd[zP].b = zB;}
    }
  }
}

// ---------------------------------------------------------------------------------

bool Gfx_DimFrame(int zS,int zF,uint8_t zD) {
  // reduces the brightness of each LED in the range zS to zF by zD
  // returns zDim = true if one or more pixels were dimmed
  bool zDim = false;
  for (int zP = zS;zP <= zF; zP++) {
    if (LED[zP].r > zD) {LED[zP].r -= zD; zDim = true;} else {LED[zP].r = 0;}
    if (LED[zP].g > zD) {LED[zP].g -= zD; zDim = true;} else {LED[zP].g = 0;}
    if (LED[zP].b > zD) {LED[zP].b -= zD; zDim = true;} else {LED[zP].b = 0;}
  }
  return zDim;
}

// ---------------------------------------------------------------------------------

bool Gfx_DimFrameE2(int zS,int zF) {
  // reduces the brightness of each LED in the range zS to zF half
  // returns zDim = true if one or more pixels were dimmed
  bool zDim = false;
  for (int zP = zS;zP <= zF; zP++) {
    if (LED[zP].r > 1) {LED[zP].r = LED[zP].r>>1; zDim = true;} else {LED[zP].r = 0;}
    if (LED[zP].g > 1) {LED[zP].g = LED[zP].g>>1; zDim = true;} else {LED[zP].g = 0;}
    if (LED[zP].b > 1) {LED[zP].b = LED[zP].b>>1; zDim = true;} else {LED[zP].b = 0;}
  }
  return zDim;
}

// ---------------------------------------------------------------------------------

void Gfx_DimFrameTo(int zS,int zF,uint8_t zD,uint8_t zR,uint8_t zG,uint8_t zB) {
  // dim LEDs in the range S to F by factor D towards target R,G,B
  // target can be brighter, hence dim up/down
  for (int zP = zS;zP <= zF; zP++) {
    if (LED[zP].r > zR) {
      if (zD < (LED[zP].r - zR)) {LED[zP].r -= zD;} else {LED[zP].r = zR;}
    } else if (LED[zP].r < zR) {
      if (zD < (zR - LED[zP].r)) {LED[zP].r += zD;} else {LED[zP].r = zR;}
    }
    if (LED[zP].g > zG) {
      if (zD < (LED[zP].g - zG)) {LED[zP].g -= zD;} else {LED[zP].g = zG;}
    } else if (LED[zP].g < zG) {
      if (zD < (zG - LED[zP].g)) {LED[zP].g += zD;} else {LED[zP].g = zG;}
    }
    if (LED[zP].r > zB) {
      if (zD < (LED[zP].b - zB)) {LED[zP].b -= zD;} else {LED[zP].b = zB;}
    } else if (LED[zP].b < zB) {
      if (zD < (zB - LED[zP].b)) {LED[zP].b += zD;} else {LED[zP].b = zB;}
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_DrawBar(int zX,int zY,int zW,int zH) {
  // draws a filled box in the frame starting at X,Y with width W, height H
  zW--; zH--;
  // check for single point
  if ((zW == 0) && (zH == 0)) {Gfx_SetXY(zX,zY); return;}
  // draw top and bottom
  Gfx_Line(zX,zY,zX + zW,zY);
  Gfx_Line(zX,zY + zH,zX + zW,zY + zH);
  // draw horizontal fill in
  if (zH < 2) {return;}
  for (int zYY = zY+1;zYY < zY + zH;zYY++) {
    Gfx_Line(zX,zYY,zX + zW,zYY);
  }
}

// ---------------------------------------------------------------------------------

void Gfx_DrawBox(int zX,int zY,int zW,int zH) {
  // draws a box starting at zX,zY with width zW and height zH
  zW--; zH--;
  // check for single point
  if ((zW == 0) && (zH == 0)) {Gfx_SetXY(zX,zY); return;}
  // draw top and bottom
  Gfx_Line(zX,zY,zX + zW,zY);
  Gfx_Line(zX,zY + zH,zX + zW,zY + zH);
  // draw sides
  if (zH < 2) {return;}
  Gfx_Line(zX,zY + 1,zX,zY + zH - 1);
  Gfx_Line(zX + zW,zY + 1,zX + zW,zY + zH - 1);
}

// ---------------------------------------------------------------------------------

void Gfx_DrawChar(int zX,int zY,uint8_t zC,uint8_t zM) {
  // loads and draws a single character C at X,Y
  // zM is a modifying mask
  // XXXXXX00 - don't modify, left-justify from zX
  // XXXXXX11 - centre on zX
  // XXXXXX01 - right-justify from zX
  // XXXX00XX - don't reset non-set pixels
  // XXXX01XX - reset non-set pixels to OFF
  // XXXX10XX - set non-set pixels to background colour BAK_Red, etc
  // this code sets and resets pixels
  loadChar(zC);         // put character font into GC[] array
  Gfx_DrawGC(zX,zY,zM); // now draw it onto the panel
}

// ---------------------------------------------------------------------------------

void Gfx_DrawGC(int zX,int zY,uint8_t zM) {
  // draws a loaded graphics character at zX,zY using mode zM
  // zM is a modifying mask
  // XXXXXX00 - don't modify, left-justify from zX
  // XXXXXX11 - centre on zX
  // XXXXXX01 - right-justify from zX
  // XXXX00XX - don't reset non-set pixels
  // XXXX01XX - reset non-set pixels to OFF
  // XXXX10XX - set non-set pixels to background colour BAK_Red, etc
  // this code sets and resets pixels
  if ((zM & B00000011) > 0) {
    // perform character justification
         if ((zM & B00000011) == B00000011) {zX = zX - (GW/2);}  // centre on zX
    else if ((zM & B00000001) == B00000001) {zX = zX - GW + 1;}  // right justify from zX
  }
  for (int zA = 0;zA <= 7;zA++) {
    // go through each row of the character array
    if ((zM & B00001100) == 0) {
      if (GW > 0) {if (GC[zA] & B10000000) {Gfx_SetXY(zX  ,zY);}}
      if (GW > 1) {if (GC[zA] & B01000000) {Gfx_SetXY(zX+1,zY);}}
      if (GW > 2) {if (GC[zA] & B00100000) {Gfx_SetXY(zX+2,zY);}}
      if (GW > 3) {if (GC[zA] & B00010000) {Gfx_SetXY(zX+3,zY);}}
      if (GW > 4) {if (GC[zA] & B00001000) {Gfx_SetXY(zX+4,zY);}}
      if (GW > 5) {if (GC[zA] & B00000100) {Gfx_SetXY(zX+5,zY);}}
      if (GW > 6) {if (GC[zA] & B00000010) {Gfx_SetXY(zX+6,zY);}}
      if (GW > 7) {if (GC[zA] & B00000001) {Gfx_SetXY(zX+7,zY);}}
    } else if ((zM & B00001100) == B00000100) {
      if (GW > 0) {if (GC[zA] & B10000000) {Gfx_SetXY(zX  ,zY);} else {Gfx_ClrXY(zX,zY);}}
      if (GW > 1) {if (GC[zA] & B01000000) {Gfx_SetXY(zX+1,zY);} else {Gfx_ClrXY(zX+1,zY);}}
      if (GW > 2) {if (GC[zA] & B00100000) {Gfx_SetXY(zX+2,zY);} else {Gfx_ClrXY(zX+2,zY);}}
      if (GW > 3) {if (GC[zA] & B00010000) {Gfx_SetXY(zX+3,zY);} else {Gfx_ClrXY(zX+3,zY);}}
      if (GW > 4) {if (GC[zA] & B00001000) {Gfx_SetXY(zX+4,zY);} else {Gfx_ClrXY(zX+4,zY);}}
      if (GW > 5) {if (GC[zA] & B00000100) {Gfx_SetXY(zX+5,zY);} else {Gfx_ClrXY(zX+5,zY);}}
      if (GW > 6) {if (GC[zA] & B00000010) {Gfx_SetXY(zX+6,zY);} else {Gfx_ClrXY(zX+6,zY);}}
      if (GW > 7) {if (GC[zA] & B00000001) {Gfx_SetXY(zX+7,zY);} else {Gfx_ClrXY(zX+7,zY);}}
    } else if ((zM & B00001100) == B00001000) {
      if (GW > 0) {if (GC[zA] & B10000000) {Gfx_SetXY(zX  ,zY);} else {Gfx_SetBakXY(zX,zY);}}
      if (GW > 1) {if (GC[zA] & B01000000) {Gfx_SetXY(zX+1,zY);} else {Gfx_SetBakXY(zX+1,zY);}}
      if (GW > 2) {if (GC[zA] & B00100000) {Gfx_SetXY(zX+2,zY);} else {Gfx_SetBakXY(zX+2,zY);}}
      if (GW > 3) {if (GC[zA] & B00010000) {Gfx_SetXY(zX+3,zY);} else {Gfx_SetBakXY(zX+3,zY);}}
      if (GW > 4) {if (GC[zA] & B00001000) {Gfx_SetXY(zX+4,zY);} else {Gfx_SetBakXY(zX+4,zY);}}
      if (GW > 5) {if (GC[zA] & B00000100) {Gfx_SetXY(zX+5,zY);} else {Gfx_SetBakXY(zX+5,zY);}}
      if (GW > 6) {if (GC[zA] & B00000010) {Gfx_SetXY(zX+6,zY);} else {Gfx_SetBakXY(zX+6,zY);}}
      if (GW > 7) {if (GC[zA] & B00000001) {Gfx_SetXY(zX+7,zY);} else {Gfx_SetBakXY(zX+7,zY);}}
    }
    zY++;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_DrawGCC(int zX,int zY,uint8_t zC,uint8_t zR) {
  // draws a column zC of a loaded graphics character at X,Y with rotation zR
  // set bit 1's apply colour, 0's clear pixels
  // zR = 0 - normal
  // zR = 1 - rotate right
  // zR = 2 - rotate right twice
  // zR = 3 - rotate left
  // set mask for zC, 0 = MSB, 7 = LSB
       if (zC == 0) {zC = B10000000;}
  else if (zC == 1) {zC = B01000000;}
  else if (zC == 2) {zC = B00100000;}
  else if (zC == 3) {zC = B00010000;}
  else if (zC == 4) {zC = B00001000;}
  else if (zC == 5) {zC = B00000100;}
  else if (zC == 6) {zC = B00000010;}
  else if (zC == 7) {zC = B00000001;}
  
  switch(zR) {
    case 0: // normal
      for (int zA = 0;zA <= 7;zA++) {
        // go through each row of the character array
        if (GC[zA] & zC) {Gfx_SetXY(zX,zY);} else {Gfx_ClrXY(zX,zY);}
        zY++;
      } break;
    case 1: // character rotated right
      for (int zA = 0;zA <= 7;zA++) {
        // go through each row of the character array
        if (GC[zA] & zC) {Gfx_SetXY(zX,zY);} else {Gfx_ClrXY(zX,zY);}
        zX--;
      } break;
    case 2: // character rotated right twice
      for (int zA = 0;zA <= 7;zA++) {
        // go through each row of the character array
        if (GC[zA] & zC) {Gfx_SetXY(zX,zY);} else {Gfx_ClrXY(zX,zY);}
        zY--;
      } break;
    case 3: // character rotated left
      for (int zA = 0;zA <= 7;zA++) {
        // go through each row of the character array
        if (GC[zA] & zC) {Gfx_SetXY(zX,zY);} else {Gfx_ClrXY(zX,zY);}
        zX++;
      } break;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_DrawGfx(uint8_t zF,int zX,int zY) {
  // draws a loaded Gfx character at X,Y
  // this sets and resets pixels
  for (int zA = 0;zA <= 7;zA++) {
    // go through each row of the character array
    if (GC[zA] & B10000000) {Gfx_SetXY(zX  ,zY);} else {Gfx_ClrXY(zX,zY);}
    if (GC[zA] & B01000000) {Gfx_SetXY(zX+1,zY);} else {Gfx_ClrXY(zX+1,zY);}
    if (GC[zA] & B00100000) {Gfx_SetXY(zX+2,zY);} else {Gfx_ClrXY(zX+2,zY);}
    if (GC[zA] & B00010000) {Gfx_SetXY(zX+3,zY);} else {Gfx_ClrXY(zX+3,zY);}
    if (GC[zA] & B00001000) {Gfx_SetXY(zX+4,zY);} else {Gfx_ClrXY(zX+4,zY);}
    if (GC[zA] & B00000100) {Gfx_SetXY(zX+5,zY);} else {Gfx_ClrXY(zX+5,zY);}
    if (GC[zA] & B00000010) {Gfx_SetXY(zX+6,zY);} else {Gfx_ClrXY(zX+6,zY);}
    if (GC[zA] & B00000001) {Gfx_SetXY(zX+7,zY);} else {Gfx_ClrXY(zX+7,zY);}
    zY++;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_DrawStr(String zT,int16_t zX,int16_t zY) {
  // draws the characters in text zT starting at zX,zY
  if (zT.length() < 1) {return;}

  for (int zP = 0;zP < zT.length();zP++) {
    // draw each character in turn using colours Txt_R,Txt_G,Txt_B
    char zC = zT.charAt(zP); Gfx_DrawChar(zX,zY,zC,0);
    zX+= GW + 1;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_DrawTxt(int16_t zX,int16_t zY) {
  // draws the characters in Text starting at zX,zY
  if (Text.length() < 1) {return;}

  for (int zP = 0;zP < Text.length();zP++) {
    // draw each character in turn using colours Txt_R,Txt_G,Txt_B
    char zC = Text.charAt(zP); Gfx_DrawChar(zX,zY,zC,0);
    zX+= GW + 1;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_Fill(int zS,int zF,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets Frame LEDs S to F to R,G,B colour
  for (int zP = zS;zP <= zF; zP++) {LED[zP].setRGB(zR,zG,zB);}
}

// ---------------------------------------------------------------------------------

void Gfx_FillReg(int zX0,int zY0,int zX1,int zY1,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets Frame LEDs in region zX0,zY0 - zX1,zY1 to zR,zG,zB colour
  int zX; int zY;
  // ensure region is specified correctly
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  // ensure region is within the frame
  if (zX0 < Xmin) {zX0 = Xmin;} if (zX0 > Xmax) {zX0 = Xmax;}
  if (zY0 < Ymin) {zY0 = Ymin;} if (zY0 > Ymax) {zY0 = Ymax;}
  if (zX1 < Xmin) {zX1 = Xmin;} if (zX1 > Xmax) {zX1 = Xmax;}
  if (zY1 < Ymin) {zY1 = Ymin;} if (zY1 > Ymax) {zY1 = Ymax;}
  // now colour the pixels in that region
  if ((zX0 == zX1) && (zY0 == zY1)){LED[Gfx_GetMatL(zX0,zY0)].setRGB(zR,zG,zB); return;} // single point
  // do row by row, working down the colum
  for (zY = zY0;zY <= zY1;zY++) {
    for (zX = zX0;zX <= zX1; zX++) {LED[Gfx_GetMatL(zX,zY)].setRGB(zR,zG,zB);}
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FillB(int zS,int zF,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets background S to F to R,G,B colour
  for (int zP = zS;zP <= zF; zP++) {Bckgnd[zP].setRGB(zR,zG,zB);}
}

// ---------------------------------------------------------------------------------

void Gfx_FillBReg(int zX0,int zY0,int zX1,int zY1,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets background LEDs in region zX0,zY0 - zX1,zY1 to zR,zG,zB colour
  int zX; int zY;
  // ensure region is specified correctly
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  // ensure region is within the frame
  if (zX0 < Xmin) {zX0 = Xmin;} if (zX0 > Xmax) {zX0 = Xmax;}
  if (zY0 < Ymin) {zY0 = Ymin;} if (zY0 > Ymax) {zY0 = Ymax;}
  if (zX1 < Xmin) {zX1 = Xmin;} if (zX1 > Xmax) {zX1 = Xmax;}
  if (zY1 < Ymin) {zY1 = Ymin;} if (zY1 > Ymax) {zY1 = Ymax;}
  // now colour the pixels in that region
  if ((zX0 == zX1) && (zY0 == zY1)){Bckgnd[Gfx_GetMatL(zX0,zY0)].setRGB(zR,zG,zB); return;} // single point
  // do row by row, working down the colum
  for (zY = zY0;zY <= zY1;zY++) {
    for (zX = zX0;zX <= zX1; zX++) {Bckgnd[Gfx_GetMatL(zX,zY)].setRGB(zR,zG,zB);}
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FillF(int zS,int zF,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets foreground S to F to R,G,B colour
  for (int zP = zS;zP <= zF; zP++) {Forgnd[zP].setRGB(zR,zG,zB);}
}

// ---------------------------------------------------------------------------------

void Gfx_FillFReg(int zX0,int zY0,int zX1,int zY1,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets foreground LEDs in region zX0,zY0 - zX1,zY1 to zR,zG,zB colour
  int zX; int zY;
  // ensure region is specified correctly
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  // ensure region is within the frame
  if (zX0 < Xmin) {zX0 = Xmin;} if (zX0 > Xmax) {zX0 = Xmax;}
  if (zY0 < Ymin) {zY0 = Ymin;} if (zY0 > Ymax) {zY0 = Ymax;}
  if (zX1 < Xmin) {zX1 = Xmin;} if (zX1 > Xmax) {zX1 = Xmax;}
  if (zY1 < Ymin) {zY1 = Ymin;} if (zY1 > Ymax) {zY1 = Ymax;}
  // now colour the pixels in that region
  if ((zX0 == zX1) && (zY0 == zY1)){Forgnd[Gfx_GetMatL(zX0,zY0)].setRGB(zR,zG,zB); return;} // single point
  // do row by row, working down the colum
  for (zY = zY0;zY <= zY1;zY++) {
    for (zX = zX0;zX <= zX1; zX++) {Forgnd[Gfx_GetMatL(zX,zY)].setRGB(zR,zG,zB);}
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FillSp(int zSp,int zS,int zF,uint8_t zR,uint8_t zG,uint8_t zB) {
  // sets sprite zSp LEDs zS to zF to zR,zG,zB colour
  for (int zP = zS;zP <= zF; zP++) {
    switch(zSp) {
      case 0: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 1: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 2: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 3: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 4: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 5: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 6: Sprite0[zP].setRGB(zR,zG,zB); break;
      case 7: Sprite0[zP].setRGB(zR,zG,zB); break;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipBckgndH(int zX0,int zY0,int zX1,int zY1) {
  // flip background region X0,Y0,X1,Y1 horizontally
  // process is row by row, swapping zX = 0 to (zX1 - zX0)/2
  int zX;   // column pointer
  int zC;   // centre column pointer
  int zXR;  // right most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zX0 + ((zX1 - zX0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zY = zY0;zY <= zY1; zY++) {
    // work vertically down the column from zY0 - zY1
    zXR = zX1;
    for (zX = zX0;zX <= zC;zX++) {
      // swap pixels in this row
      MemRGB[0] = Bckgnd[Gfx_GetMatL(zX,zY)];  // get the left LED
      Bckgnd[Gfx_GetMatL(zX,zY)] = Bckgnd[Gfx_GetMatL(zXR,zY)];
      Bckgnd[Gfx_GetMatL(zXR,zY)] = MemRGB[0]; // put the right LED
      zXR--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipBckgndV(int zX0,int zY0,int zX1,int zY1) {
  // flip background region X0,Y0,X1,Y1 vertically
  // process is column by column, swapping zY = 0 to (zY1 - zY0)/2
  int zX;   // column pointer
  int zC;   // centre row pointer
  int zYB;  // bottom most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zY0 + ((zY1 - zY0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zX = zX0;zX <= zX1; zX++) {
    // work horizontally across each row from zX0 - zX1
    zYB = zY1;
    for (zY = zY0;zY <= zC;zY++) {
      // swap pixels in this column
      MemRGB[0] = Bckgnd[Gfx_GetMatL(zX,zY)];  // get the left LED
      Bckgnd[Gfx_GetMatL(zX,zY)] = Bckgnd[Gfx_GetMatL(zX,zYB)];
      Bckgnd[Gfx_GetMatL(zX,zYB)] = MemRGB[0];
      zYB--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipForgndH(int zX0,int zY0,int zX1,int zY1) {
  // flip foreground region X0,Y0,X1,Y1 horizontally
  // process is row by row, swapping zX = 0 to (zX1 - zX0)/2
  int zX;   // column pointer
  int zC;   // centre column pointer
  int zXR;  // right most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zX0 + ((zX1 - zX0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zY = zY0;zY <= zY1; zY++) {
    // work vertically down the column from zY0 - zY1
    zXR = zX1;
    for (zX = zX0;zX <= zC;zX++) {
      // swap pixels in this row
      MemRGB[0] = Forgnd[Gfx_GetMatL(zX,zY)];  // get the left LED
      Forgnd[Gfx_GetMatL(zX,zY)] = Forgnd[Gfx_GetMatL(zXR,zY)];
      Forgnd[Gfx_GetMatL(zXR,zY)] = MemRGB[0]; // put the right LED
      zXR--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipForgndV(int zX0,int zY0,int zX1,int zY1) {
  // flip foreground region X0,Y0,X1,Y1 vertically
  // process is column by column, swapping zY = 0 to (zY1 - zY0)/2
  int zX;   // column pointer
  int zC;   // centre row pointer
  int zYB;  // bottom most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zY0 + ((zY1 - zY0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zX = zX0;zX <= zX1; zX++) {
    // work horizontally across each row from zX0 - zX1
    zYB = zY1;
    for (zY = zY0;zY <= zC;zY++) {
      // swap pixels in this column
      MemRGB[0] = Forgnd[Gfx_GetMatL(zX,zY)];  // get the left LED
      Forgnd[Gfx_GetMatL(zX,zY)] = Forgnd[Gfx_GetMatL(zX,zYB)];
      Forgnd[Gfx_GetMatL(zX,zYB)] = MemRGB[0];
      zYB--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipFrameH(int zX0,int zY0,int zX1,int zY1) {
  // flip frame region X0,Y0,X1,Y1 horizontally
  // process is row by row, swapping zX = 0 to (zX1 - zX0)/2
  int zX;   // column pointer
  int zC;   // centre column pointer
  int zXR;  // right most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zX0 + ((zX1 - zX0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zY = zY0;zY <= zY1; zY++) {
    // work vertically down the column from zY0 - zY1
    zXR = zX1;
    for (zX = zX0;zX <= zC;zX++) {
      // swap pixels in this row
      MemRGB[0] = LED[Gfx_GetMatL(zX,zY)];  // get the left LED
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zXR,zY)];
      LED[Gfx_GetMatL(zXR,zY)] = MemRGB[0]; // put the right LED
      zXR--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipFrameV(int zX0,int zY0,int zX1,int zY1) {
  // flip frame region X0,Y0,X1,Y1 vertically
  // process is column by column, swapping zY = 0 to (zY1 - zY0)/2
  int zX;   // column pointer
  int zC;   // centre row pointer
  int zYB;  // bottom most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zY0 + ((zY1 - zY0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zX = zX0;zX <= zX1; zX++) {
    // work horizontally across each row from zX0 - zX1
    zYB = zY1;
    for (zY = zY0;zY <= zC;zY++) {
      // swap pixels in this column
      MemRGB[0] = LED[Gfx_GetMatL(zX,zY)];  // get the left LED
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zX,zYB)];
      LED[Gfx_GetMatL(zX,zYB)] = MemRGB[0];
      zYB--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_FlipGCRL() {
  // flips all of the GC array bytes, bit by bit right to left
  uint8_t zFB;  // temp flip byte
  for (int zA = 0;zA <= 7;zA++) {
    zFB = 0;
    if (GC[zA] & B10000000) {zFB+= B00000001;}
    if (GC[zA] & B01000000) {zFB+= B00000010;}
    if (GC[zA] & B00100000) {zFB+= B00000100;}
    if (GC[zA] & B00010000) {zFB+= B00001000;}
    if (GC[zA] & B00001000) {zFB+= B00010000;}
    if (GC[zA] & B00000100) {zFB+= B00100000;}
    if (GC[zA] & B00000010) {zFB+= B01000000;}
    if (GC[zA] & B00000001) {zFB+= B10000000;}
    GC[zA] = zFB;
  }
}

// ---------------------------------------------------------------------------------

int16_t Gfx_GetMatL(int zX,int zY) {
  // returns LED pointer zL for point zX,zY in a frame
  // panel size is 48xNNN, but also works for 16x16
  if ((zX < Xmin) || (zX > Xmax) || (zY < Ymin) || (zY > Ymax)) {return 0;}
  
  int zO = 0; int zY16 = zY/16;
  if (zY > 15) {zO = zY16 * 768; zY = zY - (zY16 * 16);}
  if ((zX % 2) > 0) {zY = 15 - zY;}   // column phase swap
  int zL = zO + 1 + (16 * zX) + zY;   // determine pixel pointer
  return zL;
}

// ---------------------------------------------------------------------------------

int16_t Gfx_GetSpriteL(int zX,int zY) {
  // returns LED pointer zL for point zX,zY in a sprite
  // sprite size is 16x16
  if ((zX < 0) || (zX > 15) || (zY < 0) || (zY > 15)) {return 0;}

  if ((zX % 2) > 0) {zY = 15 - zY;} // column phase swap
  int zL = 1 + (16 * zX) + zY;      // determine pixel pointer
  return zL;
}

// ---------------------------------------------------------------------------------

int16_t Gfx_GetTextLen(String zT) {
  // returns the length of zT in pixels
  int zL = 0; if (zT.length() < 1) {return zL;}

  for (int zP = 0;zP < zT.length();zP++) {
    // load each character in turn and count its width
    char zC = zT.charAt(zP); loadChar(zC);  // put character font into GC[] array
    zL+= GW + 1;
  }
  if (zL > 0) {zL--;} // remove inter-character space from end
  return zL;
}

// ---------------------------------------------------------------------------------

void Gfx_InvCol(uint8_t zR,uint8_t zG,uint8_t zB) {
  // set LED colours to the inverse of zR,zG,zB based on LED_Max
//  ColRGB[0].r = LED_Max - zR; ColRGB[0].g = LED_Max - zG; ColRGB[0].b = LED_Max - zB;
  ColRGB[0].r = zR ^ 0xFF; ColRGB[0].g = zG ^ 0xFF; ColRGB[0].b = zB ^ 0xFF;
}

// ---------------------------------------------------------------------------------

void Gfx_Line(int zX0,int zY0,int zX1,int zY1) {
  // draw a line from zX0,zY0 to zX1,zY1
  int zO = 0; int zR2; int zX; int zY; int zYN; int zY16;
  
  if ((abs(zX1 - zX0)== 0) && (abs(zY1 - zY0) == 0)) {
    // single point
    if ((zX0 >= Xmin) && (zX0 <= Xmax) && (zY0 >= Ymin) && (zY0 <= Ymax)) {
      zY16 = zY0/16;
      if (zY0 > 15) {zO = zY16 * 768; zY0 = zY0 - (zY16 * 16);}
      if ((zX0 % 2) > 0) {zY0 = 15 - zY0;}  // column phase swap
      LED_PXY = zO + 1 + (16 * zX0) + zY0;  // determine pixel pointer
      LED[LED_PXY] = ColRGB[0];
    }
  } else if (abs(zX1 - zX0) > abs(zY1 - zY0)) {
    // line is longer in X than Y, so increment X
    // ensure zX1 > zX0
    if (zX1 < zX0) {zX = zX1; zX1 = zX0; zX0 = zX; zY = zY1; zY1 = zY0; zY0 = zY;} // ensure zY1 is largest
    zR2 = (zX1 - zX0)/2; if (zY1 < zY0) {zR2 = -zR2;}
    for (zX = zX0;zX <= zX1;zX++) {
      if ((zX >= Xmin) && (zX <= Xmax)) {
        // only set point if within panel area
        zYN = zY0 + (((zY1 - zY0) * (zX - zX0)) + zR2)/(zX1 - zX0);
        if ((zYN >= Ymin) && (zYN <= Ymax)) {
          zY16 = zYN/16;
          if (zYN > 15) {zO = zY16 * 768; zYN = zYN - (zY16 * 16);}
          if ((zX % 2) > 0) {zYN = 15 - zYN;} // column phase swap
          LED_PXY = zO + 1 + (16 * zX) + zYN; // determine pixel pointer
          LED[LED_PXY] = ColRGB[0];
        }
      }
    }
  } else {
    // line is longer in Y than X
    // ensure zY1 > zY0
    if (zY1 < zY0) {zY = zY1; zY1 = zY0; zY0 = zY; zX = zX1; zX1 = zX0; zX0 = zX;} // ensure zX1 is largest
    zR2 = (zY1 - zY0)/2; if (zX1 < zX0) {zR2 = -zR2;}
    for (zY = zY0;zY <= zY1;zY++) {
      if ((zY >= Ymin) && (zY <= Ymax)) {
        // only set point if within panel area
        zX = zX0 + (((zX1 - zX0) * (zY - zY0))/(zY1 - zY0));
        if ((zX >= Xmin) && (zX <= Xmax)) {
          zYN = zY; zY16 = zYN/16;
          if (zYN > 15) {zO = zY16 * 768; zYN = zYN - (zY16 * 16);}
          if ((zX % 2) > 0) {zYN = 15 - zYN;} // column phase swap
          LED_PXY = zO + 1 + (16 * zX) + zYN; // determine pixel pointer
          LED[LED_PXY] = ColRGB[0];
        }
      }
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_MemCol() {
  // place the current RGB colours in a memory
  MemRGB[0] = ColRGB[0];
}

// ---------------------------------------------------------------------------------

void Gfx_MultCol(uint8_t zM) {
  // multiply the ink colours by factor zM
  ColRGB[0].r*= zM; ColRGB[0].g*= zM; ColRGB[0].b*= zM;
}

// ---------------------------------------------------------------------------------

void Gfx_PicXY(int zX,int zY) {
  // gets a pixel colour from X,Y
  // co-ordinate system is 0,0 at top left corner, to Xmax,Ymax
  if ((zX < Xmin) || (zX > Xmax) || (zY < Ymin) || (zY > Ymax)) {return;}
  
  int zO = 0; int zY16 = zY/16;
  if (zY > 15) {zO = zY16 * 768; zY = zY - (zY16 * 16);}
  if ((zX % 2) > 0) {zY = 15 - zY;} // column phase swap
  LED_PXY = GfxP0 - zO - (16 * zX) - zY; // determine pixel pointer
  ColRGB[0] = LED[LED_PXY];
}

// ---------------------------------------------------------------------------------

void Gfx_RecCol() {
  // recall the colours stored in a memory
  ColRGB[0] = MemRGB[0];
}

// ---------------------------------------------------------------------------------

void Gfx_RndCol(int zM) {
  // sets a random ink colour, max strength M
  int z12 = zM/2; int z14 = zM/4; int z34 = zM - z14;
Again:
  uint8_t zR = random8(12);
  if (zR == RndCol) {goto Again;}
  switch(zR) {
    case  0: ColRGB[0].r =  zM; ColRGB[0].g =   0; ColRGB[0].b =   0; break;
    case  1: ColRGB[0].r = z34; ColRGB[0].g = z14; ColRGB[0].b =   0; break;
    case  2: ColRGB[0].r = z12; ColRGB[0].g = z12; ColRGB[0].b =   0; break;
    case  3: ColRGB[0].r = z14; ColRGB[0].g = z34; ColRGB[0].b =   0; break;
    case  4: ColRGB[0].r =   0; ColRGB[0].g =  zM; ColRGB[0].b =   0; break;
    case  5: ColRGB[0].r =   0; ColRGB[0].g = z34; ColRGB[0].b = z14; break;
    case  6: ColRGB[0].r =   0; ColRGB[0].g = z12; ColRGB[0].b = z12; break;
    case  7: ColRGB[0].r =   0; ColRGB[0].g = z14; ColRGB[0].b = z34; break;
    case  8: ColRGB[0].r =   0; ColRGB[0].g =   0; ColRGB[0].b =  zM; break;
    case  9: ColRGB[0].r = z14; ColRGB[0].g =   0; ColRGB[0].b = z34; break;
    case 10: ColRGB[0].r = z12; ColRGB[0].g =   0; ColRGB[0].b = z12; break;
    case 11: ColRGB[0].r = z34; ColRGB[0].g =   0; ColRGB[0].b = z14; break;
  } RndCol = zR;
}

// ---------------------------------------------------------------------------------

void Gfx_RotDwn() {
  // rotate panel Frame one pixel down on a 48 x NNN panel array
  Gfx_RotDwnWin(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotDwnBck() {
  // rotate background one pixel down on a 48 x NNN panel array
  Gfx_RotDwnWinBck(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotDwnFor() {
  // rotate foreground one pixel down on a 48 x NNN panel array
  Gfx_RotDwnWinFor(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotDwnWin(int zX0,int zY0,int zX1,int zY1) {
  // rotate panel window X0,Y0 to X1,Y1 one pixel down
  // process is column by column zX = 0 to Xmax
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zX = zX0;zX <= zX1;zX++) {
    // shift this column down
    MemRGB[0] = LED[Gfx_GetMatL(zX,zY1)];  // get the bottom LED
    for (zY = zY1;zY > zY0; zY--) {
      // work vertically up the column
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zX,zY-1)];
    }
    LED[Gfx_GetMatL(zX,zY0)] = MemRGB[0]; // replace the top LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotDwnWinBck(int zX0,int zY0,int zX1,int zY1) {
  // rotate background window X0,Y0 to X1,Y1 one pixel down
  // process is column by column zX = 0 to Xmax
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zX = zX0;zX <= zX1;zX++) {
    // shift this column down
    MemRGB[0] = Bckgnd[Gfx_GetMatL(zX,zY1)];  // get the bottom LED
    for (zY = zY1;zY > zY0; zY--) {
      // work vertically up the column
      Bckgnd[Gfx_GetMatL(zX,zY)] = Bckgnd[Gfx_GetMatL(zX,zY-1)];
    }
    Bckgnd[Gfx_GetMatL(zX,zY0)] = MemRGB[0]; // replace the top LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotDwnWinFor(int zX0,int zY0,int zX1,int zY1) {
  // rotate foreground window X0,Y0 to X1,Y1 one pixel down
  // process is column by column zX = 0 to Xmax
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zX = zX0;zX <= zX1;zX++) {
    // shift this column down
    MemRGB[0] = Forgnd[Gfx_GetMatL(zX,zY1)];  // get the bottom LED
    for (zY = zY1;zY > zY0; zY--) {
      // work vertically up the column
      Forgnd[Gfx_GetMatL(zX,zY)] = Forgnd[Gfx_GetMatL(zX,zY-1)];
    }
    Forgnd[Gfx_GetMatL(zX,zY0)] = MemRGB[0]; // replace the top LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotGCL() {
  // rotates the GC array bytes anti-clockwise, left
  // start by clearing buffer GB[]
  GB[0] = 0; GB[1] = 0; GB[2] = 0; GB[3] = 0; GB[4] = 0; GB[5] = 0; GB[6] = 0; GB[7] = 0;
  uint8_t zM = B00000001;  // set bit mask
  for (int zA = 7;zA >= 0;zA--) {
    byte zB = 7;  // reset buffer pointer
          if (GC[zA] & B10000000) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B01000000) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B00100000) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B00010000) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B00001000) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B00000100) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B00000010) {GB[zB]+= zM;}
    zB--; if (GC[zA] & B00000001) {GB[zB]+= zM;}
    zM += zM; // increment bit mask
  }
  // copy buffer GB[] back into GC[]
  for (int zA = 0;zA <= 7;zA++) {GC[zA] = GB[zA];}
  // swap width & height
  GW = 8; GH = 8;
}

// ---------------------------------------------------------------------------------

void Gfx_RotGCR() {
  // rotates the GC array bytes clockwise, right
  // start by clearing buffer GB[]
  GB[0] = 0; GB[1] = 0; GB[2] = 0; GB[3] = 0; GB[4] = 0; GB[5] = 0; GB[6] = 0; GB[7] = 0;
  uint8_t zM = B00000001;  // set bit mask
  for (int zA = 0;zA <= 7;zA++) {
    byte zB = 0;  // reset buffer pointer
          if (GC[zA] & B10000000) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B01000000) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B00100000) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B00010000) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B00001000) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B00000100) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B00000010) {GB[zB]+= zM;}
    zB++; if (GC[zA] & B00000001) {GB[zB]+= zM;}
    zM += zM; // increment bit mask
  }
  // copy buffer GB[] back into GC[]
  for (int zA = 0;zA <= 7;zA++) {GC[zA] = GB[zA];}
  // swap width & height
  GW = 8; GH = 8;
}

// ---------------------------------------------------------------------------------

void Gfx_RotLft() {
  // rotate whole panel one pixel left
  Gfx_RotLftWin(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotLftBck() {
  // rotate whole background array one pixel left
  Gfx_RotLftWinBck(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotLftFor() {
  // rotate whole foreground array one pixel left
  Gfx_RotLftWinFor(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotLftWin(int zX0,int zY0,int zX1,int zY1) {
  // rotate panel window X0,Y0 to X1,Y1 one pixel to the left
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zY = zY0;zY <= zY1;zY++) {
    // shift this row left
    MemRGB[0] = LED[Gfx_GetMatL(zX0,zY)];  // get the left LED
    for (zX = zX0;zX < zX1; zX++) {
      // work horizontally along the column
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zX+1,zY)];
    }
    LED[Gfx_GetMatL(zX1,zY)] = MemRGB[0];  // replace the right LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotLftWinBck(int zX0,int zY0,int zX1,int zY1) {
  // rotate background window X0,Y0 to X1,Y1 one pixel to the left
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zY = zY0;zY <= zY1;zY++) {
    // shift this row left
    MemRGB[0] = Bckgnd[Gfx_GetMatL(zX0,zY)];  // get the left LED
    for (zX = zX0;zX < zX1; zX++) {
      // work horizontally along the column
      Bckgnd[Gfx_GetMatL(zX,zY)] = Bckgnd[Gfx_GetMatL(zX+1,zY)];
    }
    Bckgnd[Gfx_GetMatL(zX1,zY)] = MemRGB[0];  // replace the right LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotLftWinFor(int zX0,int zY0,int zX1,int zY1) {
  // rotate foreground window X0,Y0 to X1,Y1 one pixel to the left
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zY = zY0;zY <= zY1;zY++) {
    // shift this row left
    MemRGB[0] = Forgnd[Gfx_GetMatL(zX0,zY)];  // get the left LED
    for (zX = zX0;zX < zX1; zX++) {
      // work horizontally along the column
      Forgnd[Gfx_GetMatL(zX,zY)] = Forgnd[Gfx_GetMatL(zX+1,zY)];
    }
    Forgnd[Gfx_GetMatL(zX1,zY)] = MemRGB[0];  // replace the right LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotRht() {
  // rotate whole panel array one pixel right
  Gfx_RotRhtWin(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotRhtBck() {
  // rotate whole background array one pixel right
  Gfx_RotRhtWinBck(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotRhtFor() {
  // rotate whole foreground array one pixel right
  Gfx_RotRhtWinFor(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotRhtWin(int zX0,int zY0,int zX1,int zY1) {
  // rotate panel window X0,Y0 to X1,Y1 one pixel to the right
  // copy column Xmax, rotate whole panel right, paste column Xmin
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zY = zY0;zY <= zY1;zY++) {
    // shift this row left
    MemRGB[0] = LED[Gfx_GetMatL(zX1,zY)];  // get the right LED
    for (zX = zX1;zX > zX0; zX--) {
      // work horizontally along the column, right to left
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zX-1,zY)];
    }
    LED[Gfx_GetMatL(zX0,zY)] = MemRGB[0];  // replace the left LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotRhtWinBck(int zX0,int zY0,int zX1,int zY1) {
  // rotate background window X0,Y0 to X1,Y1 one pixel to the right
  // copy column Xmax, rotate whole panel right, paste column Xmin
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zY = zY0;zY <= zY1;zY++) {
    // shift this row left
    MemRGB[0] = Bckgnd[Gfx_GetMatL(zX1,zY)];  // get the right LED
    for (zX = zX1;zX > zX0; zX--) {
      // work horizontally along the column, right to left
      Bckgnd[Gfx_GetMatL(zX,zY)] = Bckgnd[Gfx_GetMatL(zX-1,zY)];
    }
    Bckgnd[Gfx_GetMatL(zX0,zY)] = MemRGB[0];  // replace the left LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotRhtWinFor(int zX0,int zY0,int zX1,int zY1) {
  // rotate background window X0,Y0 to X1,Y1 one pixel to the right
  // copy column Xmax, rotate whole panel right, paste column Xmin
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zY = zY0;zY <= zY1;zY++) {
    // shift this row left
    MemRGB[0] = Forgnd[Gfx_GetMatL(zX1,zY)];  // get the right LED
    for (zX = zX1;zX > zX0; zX--) {
      // work horizontally along the column, right to left
      Forgnd[Gfx_GetMatL(zX,zY)] = Forgnd[Gfx_GetMatL(zX-1,zY)];
    }
    Forgnd[Gfx_GetMatL(zX0,zY)] = MemRGB[0];  // replace the left LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotUp() {
  // rotate panel Frame one pixel up on a 48 x NNN panel array
  Gfx_RotUpWin(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotUpBck() {
  // rotate panel background one pixel up on a 48 x NNN panel array
  Gfx_RotUpWinBck(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotUpFor() {
  // rotate panel foreground one pixel up on a 48 x NNN panel array
  Gfx_RotUpWinFor(Xmin,Ymin,Xmax,Ymax);
}

// ---------------------------------------------------------------------------------

void Gfx_RotUpWin(int zX0,int zY0,int zX1,int zY1) {
  // rotate panel window X0,Y0 to X1,Y1 one pixel to the up
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zX = zX0;zX <= zX1;zX++) {
    // shift this column down
    MemRGB[0] = LED[Gfx_GetMatL(zX,zY0)];  // get the top LED
    for (zY = zY0;zY < zY1; zY++) {
      // work vertically down the column
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zX,zY+1)];
    }
    LED[Gfx_GetMatL(zX,zY1)] = MemRGB[0]; // replace the bottom LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotUpWinBck(int zX0,int zY0,int zX1,int zY1) {
  // rotate background window X0,Y0 to X1,Y1 one pixel to the up
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zX = zX0;zX <= zX1;zX++) {
    // shift this column down
    MemRGB[0] = Bckgnd[Gfx_GetMatL(zX,zY0)];  // get the top LED
    for (zY = zY0;zY < zY1; zY++) {
      // work vertically down the column
      Bckgnd[Gfx_GetMatL(zX,zY)] = Bckgnd[Gfx_GetMatL(zX,zY+1)];
    }
    Bckgnd[Gfx_GetMatL(zX,zY1)] = MemRGB[0]; // replace the bottom LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_RotUpWinFor(int zX0,int zY0,int zX1,int zY1) {
  // rotate foreground window X0,Y0 to X1,Y1 one pixel to the up
  int zX;  // column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  for (zX = zX0;zX <= zX1;zX++) {
    // shift this column down
    MemRGB[0] = Forgnd[Gfx_GetMatL(zX,zY0)];  // get the top LED
    for (zY = zY0;zY < zY1; zY++) {
      // work vertically down the column
      Forgnd[Gfx_GetMatL(zX,zY)] = Forgnd[Gfx_GetMatL(zX,zY+1)];
    }
    Forgnd[Gfx_GetMatL(zX,zY1)] = MemRGB[0]; // replace the bottom LED
  }
}

// ---------------------------------------------------------------------------------

void Gfx_ScrollTxt() {
  // scroll Text within a full window strip at Text_SY
  if (Verbose) {Serial.println("ScrollTask " + String(ScrollTask));}
  if (Text.length() < 1) {return;}

  if (ScrollTask == 0) {
    // initialises the scrolling function
    Txt_Pnt = 0;  // point at first character
    Txt_Num = Text.length();
    // get the character from LED_Text
    Txt_Char = Text.charAt(Txt_Pnt);  // get the character from the string
    loadChar(Txt_Char); Txt_Cnt = 0;  // copy in charàcter bitmap and set column to 0
    ScrollTask++;
  }
  switch (ScrollTask) {
    case 1: // scroll text left 1 pixel at a time
      Gfx_TxtLft();  // move the text strip left one pixel
      Gfx_DrawGCC(Xmax,Text_SY,Txt_Cnt,0); // draw one column of the character
      Txt_Cnt++; if (Txt_Cnt > GW) {
        // move onto the next character
        if (Txt_Pnt < (Txt_Num - 1)) {
          Txt_Pnt++;
          Txt_Char = Text.charAt(Txt_Pnt);
          loadChar(Txt_Char); Txt_Cnt = 0;
        } else {ScrollTask = 2;}
      } break;
    case 2: // end of text string reached, so fill in a blank section
      Gfx_TxtLft();  // move the text strip left one pixel
      MemRGB[0] = ColRGB[0];  // remember the current text colour
      ColRGB[0].setRGB(0,0,0);
      Gfx_Line(Xmax,Text_SY,Xmax,Text_SY + 7); // remove the wrapped character line
      ColRGB[0] = MemRGB[0];  // recall the current text colour
      break;
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SetBakXY(int zX,int zY) {
  // resets a pixel at zX,zY to the BAK colour
  // co-ordinate system is 0,0 at top left corner, to Xmax,Ymax
  if ((zX < Xmin) || (zX > Xmax) || (zY < Ymin) || (zY > Ymax)) {return;}

  int zO = 0; int zY16 = zY/16;
  if (zY > 15) {zO = zY16 * 768; zY = zY - (zY16 * 16);}
  if ((zX % 2) > 0) {zY = 15 - zY;}   // column phase swap
  LED_PXY = zO + 1 + (16 * zX) + zY;  // determine pixel pointer
  LED[LED_PXY] = BakRGB[0];
}

// ---------------------------------------------------------------------------------

void Gfx_SetCol(uint8_t zC) {
  // sets RGB colour variables from a palette based on zC
  switch(zC) {
    case  0: ColRGB[0].r = Col11; ColRGB[0].g =     0; ColRGB[0].b =     0; break; // red
    case  1: ColRGB[0].r = Col34; ColRGB[0].g = Col14; ColRGB[0].b =     0; break; // red/yellow
    case  2: ColRGB[0].r = Col12; ColRGB[0].g = Col12; ColRGB[0].b =     0; break; // yellow
    case  3: ColRGB[0].r = Col14; ColRGB[0].g = Col34; ColRGB[0].b =     0; break; // green/yellow
    case  4: ColRGB[0].r =     0; ColRGB[0].g = Col11; ColRGB[0].b =     0; break; // green
    case  5: ColRGB[0].r =     0; ColRGB[0].g = Col34; ColRGB[0].b = Col14; break; // green/magenta
    case  6: ColRGB[0].r =     0; ColRGB[0].g = Col12; ColRGB[0].b = Col12; break; // magenta
    case  7: ColRGB[0].r =     0; ColRGB[0].g = Col14; ColRGB[0].b = Col34; break; // blue/magenta
    case  8: ColRGB[0].r =     0; ColRGB[0].g =     0; ColRGB[0].b = Col11; break; // blue
    case  9: ColRGB[0].r = Col14; ColRGB[0].g =     0; ColRGB[0].b = Col34; break; // blue/purple
    case 10: ColRGB[0].r = Col12; ColRGB[0].g =     0; ColRGB[0].b = Col12; break; // purple
    case 11: ColRGB[0].r = Col34; ColRGB[0].g =     0; ColRGB[0].b = Col14; break; // red/purple
    case 12: ColRGB[0].r =     0; ColRGB[0].g =     0; ColRGB[0].b =     0; break; // OFF
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SetColMax(int16_t zM) {
  // sets the maximum value of colour palette, 4 - 255
  if (zM < 4) {zM = 4;};
  if (zM > 255) {zM = 255;}
  Col11 = zM; Col34 = (zM * 3)/4; Col12 = zM/2; Col14 = zM/4;
}

// ---------------------------------------------------------------------------------

void Gfx_SetXY(int zX,int zY) {
  // sets a pixel at zX,zY to the current colour on three 16x16 panels
  // co-ordinate system is 0,0 at top left corner, to Xmax,Ymax
  if ((zX < Xmin) || (zX > Xmax) || (zY < Ymin) || (zY > Ymax)) {return;}
  
  int zO = 0; int zY16 = zY/16;
  if (zY > 15) {zO = zY16 * 768; zY = zY - (zY16 * 16);}
  if ((zX % 2) > 0) {zY = 15 - zY;}   // column phase swap
  LED_PXY = zO + 1 + (16 * zX) + zY;  // determine pixel pointer
  LED[LED_PXY] = ColRGB[0];
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteFlipH(int zS,int zX0,int zY0,int zX1,int zY1) {
  // flip sprite S region X0,Y0,X1,Y1 horizontally
  // process is row by row, swapping zX = 0 to (zX1 - zX0)/2
  int zX;   // column pointer
  int zC;   // centre column pointer
  int zXR;  // right most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zX0 + ((zX1 - zX0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zY = zY0;zY <= zY1; zY++) {
    // work vertically down the column from zY0 - zY1
    zXR = zX1;
    for (zX = zX0;zX <= zC;zX++) {
      // swap pixels in this row
      switch (zS) {
        case 0 : MemRGB[0] = Sprite0[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite0[Gfx_GetSpriteL(zX,zY)] = Sprite0[Gfx_GetSpriteL(zXR,zY)];
          Sprite0[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 1 : MemRGB[0] = Sprite1[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite1[Gfx_GetSpriteL(zX,zY)] = Sprite1[Gfx_GetSpriteL(zXR,zY)];
          Sprite1[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 2 : MemRGB[0] = Sprite2[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite2[Gfx_GetSpriteL(zX,zY)] = Sprite2[Gfx_GetSpriteL(zXR,zY)];
          Sprite2[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 3 : MemRGB[0] = Sprite3[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite3[Gfx_GetSpriteL(zX,zY)] = Sprite3[Gfx_GetSpriteL(zXR,zY)];
          Sprite3[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 4 : MemRGB[0] = Sprite4[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite4[Gfx_GetSpriteL(zX,zY)] = Sprite4[Gfx_GetSpriteL(zXR,zY)];
          Sprite4[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 5 : MemRGB[0] = Sprite5[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite5[Gfx_GetSpriteL(zX,zY)] = Sprite5[Gfx_GetSpriteL(zXR,zY)];
          Sprite5[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 6 : MemRGB[0] = Sprite6[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite6[Gfx_GetSpriteL(zX,zY)] = Sprite6[Gfx_GetSpriteL(zXR,zY)];
          Sprite6[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
        case 7 : MemRGB[0] = Sprite7[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite7[Gfx_GetSpriteL(zX,zY)] = Sprite7[Gfx_GetSpriteL(zXR,zY)];
          Sprite7[Gfx_GetSpriteL(zXR,zY)] = MemRGB[0]; break;
      } zXR--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteFlipV(int zS,int zX0,int zY0,int zX1,int zY1) {
  // flip sprite S region X0,Y0,X1,Y1 vertically
  // process is column by column, swapping zY = 0 to (zY1 - zY0)/2
  int zX;   // column pointer
  int zC;   // centre row pointer
  int zYB;  // bottom most column pointer
  if (zX1 < zX0) {zX = zX0; zX0 = zX1; zX1 = zX;}
  zC = zY0 + ((zY1 - zY0)/2);
  int zY;  // row pointer
  if (zY1 < zY0) {zY = zY0; zY0 = zY1; zY1 = zY;}
  
  for (zX = zX0;zX <= zX1; zX++) {
    // work horizontally across each row from zX0 - zX1
    zYB = zY1;
    for (zY = zY0;zY <= zC;zY++) {
      // swap pixels in this column
      switch (zS) {
        case 0 : MemRGB[0] = Sprite0[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite0[Gfx_GetSpriteL(zX,zY)] = Sprite0[Gfx_GetSpriteL(zX,zYB)];
          Sprite0[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 1 : MemRGB[0] = Sprite1[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite1[Gfx_GetSpriteL(zX,zY)] = Sprite1[Gfx_GetSpriteL(zX,zYB)];
          Sprite1[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 2 : MemRGB[0] = Sprite2[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite2[Gfx_GetSpriteL(zX,zY)] = Sprite2[Gfx_GetSpriteL(zX,zYB)];
          Sprite2[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 3 : MemRGB[0] = Sprite3[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite3[Gfx_GetSpriteL(zX,zY)] = Sprite3[Gfx_GetSpriteL(zX,zYB)];
          Sprite3[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 4 : MemRGB[0] = Sprite4[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite4[Gfx_GetSpriteL(zX,zY)] = Sprite4[Gfx_GetSpriteL(zX,zYB)];
          Sprite4[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 5 : MemRGB[0] = Sprite5[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite5[Gfx_GetSpriteL(zX,zY)] = Sprite5[Gfx_GetSpriteL(zX,zYB)];
          Sprite5[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 6 : MemRGB[0] = Sprite6[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite6[Gfx_GetSpriteL(zX,zY)] = Sprite6[Gfx_GetSpriteL(zX,zYB)];
          Sprite6[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
        case 7 : MemRGB[0] = Sprite7[Gfx_GetSpriteL(zX,zY)];  // get the left LED
          Sprite7[Gfx_GetSpriteL(zX,zY)] = Sprite7[Gfx_GetSpriteL(zX,zYB)];
          Sprite7[Gfx_GetSpriteL(zX,zYB)] = MemRGB[0]; break;
      } zYB--;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteLoad(int zS) {
  // loads Sprite zS into the output frame at SpriteX[zS],SpriteY[zS]
  int zL; MemRGB[0].setRGB(0,0,0);
//  Serial.println(String(zS) + "\t" + String(SpriteX[zS]) + "\t" + String(SpriteY[zS]));
  for (int zY = 0;zY <= 15;zY++) {
    for (int zX = 0;zX <= 15;zX++) {
      zL = Gfx_GetSpriteL(zX,zY);
      switch(zS) {
        case 0:
          if (Sprite0[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite0[zL];}
          break;
        case 1:
          if (Sprite1[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite1[zL];}
          break;
        case 2:
          if (Sprite2[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite2[zL];}
          break;
        case 3:
          if (Sprite3[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite3[zL];}
          break;
        case 4:
          if (Sprite4[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite4[zL];}
          break;
        case 5:
          if (Sprite5[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite5[zL];}
          break;
        case 6:
          if (Sprite6[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite6[zL];}
          break;
        case 7:
          if (Sprite7[zL] != MemRGB[0]) {LED[Gfx_GetMatL(SpriteX[zS] + zX,SpriteY[zS] + zY)] = Sprite7[zL];}
          break;
      }
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteRotDwn(int zS) {
  // rotate sprite zS one pixel Down, column by column
  for (int zX = 0;zX <= 15;zX++) {
    switch(zS) {
      case 0: MemRGB[0] = Sprite0[Gfx_GetSpriteL(zX,15)]; break;
      case 1: MemRGB[0] = Sprite1[Gfx_GetSpriteL(zX,15)]; break;
      case 2: MemRGB[0] = Sprite2[Gfx_GetSpriteL(zX,15)]; break;
      case 3: MemRGB[0] = Sprite3[Gfx_GetSpriteL(zX,15)]; break;
      case 4: MemRGB[0] = Sprite4[Gfx_GetSpriteL(zX,15)]; break;
      case 5: MemRGB[0] = Sprite5[Gfx_GetSpriteL(zX,15)]; break;
      case 6: MemRGB[0] = Sprite6[Gfx_GetSpriteL(zX,15)]; break;
      case 7: MemRGB[0] = Sprite7[Gfx_GetSpriteL(zX,15)]; break;
    }
    for (int zY = 15;zY > 0;zY--) {
      switch(zS) {
        case 0: Sprite0[Gfx_GetSpriteL(zX,zY)] = Sprite0[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 1: Sprite1[Gfx_GetSpriteL(zX,zY)] = Sprite1[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 2: Sprite2[Gfx_GetSpriteL(zX,zY)] = Sprite2[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 3: Sprite3[Gfx_GetSpriteL(zX,zY)] = Sprite3[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 4: Sprite4[Gfx_GetSpriteL(zX,zY)] = Sprite4[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 5: Sprite5[Gfx_GetSpriteL(zX,zY)] = Sprite5[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 6: Sprite6[Gfx_GetSpriteL(zX,zY)] = Sprite6[Gfx_GetSpriteL(zX,zY-1)]; break;
        case 7: Sprite7[Gfx_GetSpriteL(zX,zY)] = Sprite7[Gfx_GetSpriteL(zX,zY-1)]; break;
      }
    }
    switch(zS) {
      case 0: Sprite0[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 1: Sprite1[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 2: Sprite2[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 3: Sprite3[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 4: Sprite4[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 5: Sprite5[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 6: Sprite6[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
      case 7: Sprite7[Gfx_GetSpriteL(zX,0)] = MemRGB[0]; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteRotLft(int zS) {
  // rotate sprite zS one pixel Left, row by row
  for (int zY = 0;zY <= 15;zY++) {
    switch(zS) {
      case 0: MemRGB[0] = Sprite0[Gfx_GetSpriteL(0,zY)]; break;
      case 1: MemRGB[0] = Sprite1[Gfx_GetSpriteL(0,zY)]; break;
      case 2: MemRGB[0] = Sprite2[Gfx_GetSpriteL(0,zY)]; break;
      case 3: MemRGB[0] = Sprite3[Gfx_GetSpriteL(0,zY)]; break;
      case 4: MemRGB[0] = Sprite4[Gfx_GetSpriteL(0,zY)]; break;
      case 5: MemRGB[0] = Sprite5[Gfx_GetSpriteL(0,zY)]; break;
      case 6: MemRGB[0] = Sprite6[Gfx_GetSpriteL(0,zY)]; break;
      case 7: MemRGB[0] = Sprite7[Gfx_GetSpriteL(0,zY)]; break;
    }
    for (int zX = 0;zX < 15;zX++) {
      switch(zS) {
        case 0: Sprite0[Gfx_GetSpriteL(zX,zY)] = Sprite0[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 1: Sprite1[Gfx_GetSpriteL(zX,zY)] = Sprite1[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 2: Sprite2[Gfx_GetSpriteL(zX,zY)] = Sprite2[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 3: Sprite3[Gfx_GetSpriteL(zX,zY)] = Sprite3[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 4: Sprite4[Gfx_GetSpriteL(zX,zY)] = Sprite4[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 5: Sprite5[Gfx_GetSpriteL(zX,zY)] = Sprite5[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 6: Sprite6[Gfx_GetSpriteL(zX,zY)] = Sprite6[Gfx_GetSpriteL(zX+1,zY)]; break;
        case 7: Sprite7[Gfx_GetSpriteL(zX,zY)] = Sprite7[Gfx_GetSpriteL(zX+1,zY)]; break;
      }
    }
    switch(zS) {
      case 0: Sprite0[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 1: Sprite1[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 2: Sprite2[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 3: Sprite3[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 4: Sprite4[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 5: Sprite5[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 6: Sprite6[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
      case 7: Sprite7[Gfx_GetSpriteL(15,zY)] = MemRGB[0]; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteRotUp(int zS) {
  // rotate sprite zS one pixel Up, column by column
  for (int zX = 0;zX <= 15;zX++) {
    switch(zS) {
      case 0: MemRGB[0] = Sprite0[Gfx_GetSpriteL(zX,0)]; break;
      case 1: MemRGB[0] = Sprite1[Gfx_GetSpriteL(zX,0)]; break;
      case 2: MemRGB[0] = Sprite2[Gfx_GetSpriteL(zX,0)]; break;
      case 3: MemRGB[0] = Sprite3[Gfx_GetSpriteL(zX,0)]; break;
      case 4: MemRGB[0] = Sprite4[Gfx_GetSpriteL(zX,0)]; break;
      case 5: MemRGB[0] = Sprite5[Gfx_GetSpriteL(zX,0)]; break;
      case 6: MemRGB[0] = Sprite6[Gfx_GetSpriteL(zX,0)]; break;
      case 7: MemRGB[0] = Sprite7[Gfx_GetSpriteL(zX,0)]; break;
    }
    for (int zY = 0;zY < 15;zY++) {
      switch(zS) {
        case 0: Sprite0[Gfx_GetSpriteL(zX,zY)] = Sprite0[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 1: Sprite1[Gfx_GetSpriteL(zX,zY)] = Sprite1[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 2: Sprite2[Gfx_GetSpriteL(zX,zY)] = Sprite2[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 3: Sprite3[Gfx_GetSpriteL(zX,zY)] = Sprite3[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 4: Sprite4[Gfx_GetSpriteL(zX,zY)] = Sprite4[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 5: Sprite5[Gfx_GetSpriteL(zX,zY)] = Sprite5[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 6: Sprite6[Gfx_GetSpriteL(zX,zY)] = Sprite6[Gfx_GetSpriteL(zX,zY+1)]; break;
        case 7: Sprite7[Gfx_GetSpriteL(zX,zY)] = Sprite7[Gfx_GetSpriteL(zX,zY+1)]; break;
      }
    }
    switch(zS) {
      case 0: Sprite0[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 1: Sprite1[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 2: Sprite2[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 3: Sprite3[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 4: Sprite4[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 5: Sprite5[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 6: Sprite6[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
      case 7: Sprite7[Gfx_GetSpriteL(zX,15)] = MemRGB[0]; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_SpriteRotRht(int zS) {
  // rotate sprite zS one pixel Right, row by row
  for (int zY = 0;zY <= 15;zY++) {
    switch(zS) {
      case 0: MemRGB[0] = Sprite0[Gfx_GetSpriteL(15,zY)]; break;
      case 1: MemRGB[0] = Sprite1[Gfx_GetSpriteL(15,zY)]; break;
      case 2: MemRGB[0] = Sprite2[Gfx_GetSpriteL(15,zY)]; break;
      case 3: MemRGB[0] = Sprite3[Gfx_GetSpriteL(15,zY)]; break;
      case 4: MemRGB[0] = Sprite4[Gfx_GetSpriteL(15,zY)]; break;
      case 5: MemRGB[0] = Sprite5[Gfx_GetSpriteL(15,zY)]; break;
      case 6: MemRGB[0] = Sprite6[Gfx_GetSpriteL(15,zY)]; break;
      case 7: MemRGB[0] = Sprite7[Gfx_GetSpriteL(15,zY)]; break;
    }
    for (int zX = 15;zX > 0;zX--) {
      switch(zS) {
        case 0: Sprite0[Gfx_GetSpriteL(zX,zY)] = Sprite0[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 1: Sprite1[Gfx_GetSpriteL(zX,zY)] = Sprite1[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 2: Sprite2[Gfx_GetSpriteL(zX,zY)] = Sprite2[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 3: Sprite3[Gfx_GetSpriteL(zX,zY)] = Sprite3[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 4: Sprite4[Gfx_GetSpriteL(zX,zY)] = Sprite4[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 5: Sprite5[Gfx_GetSpriteL(zX,zY)] = Sprite5[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 6: Sprite6[Gfx_GetSpriteL(zX,zY)] = Sprite6[Gfx_GetSpriteL(zX-1,zY)]; break;
        case 7: Sprite7[Gfx_GetSpriteL(zX,zY)] = Sprite7[Gfx_GetSpriteL(zX-1,zY)]; break;
      }
    }
    switch(zS) {
      case 0: Sprite0[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 1: Sprite1[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 2: Sprite2[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 3: Sprite3[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 4: Sprite4[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 5: Sprite5[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 6: Sprite6[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
      case 7: Sprite7[Gfx_GetSpriteL(0,zY)] = MemRGB[0]; break;
    }
  }
}

// ---------------------------------------------------------------------------------

void Gfx_StripRotDwn() {
  // rotate the entire contents of the strip from (NumLEDs - 1) to 1
  int zF = 1; // set the from pointer
  Block[0] = LED[zF]; // temp copy
  zF++; // adjust the from pointer
  for (int zT = 1;zT > (NumLEDs - 2);zT++) {
    // cycle backwards down the strip
    LED[zT] = LED[zF]; zF++;
  }
  LED[zF] = Block[0];  // overwrite with temp copy
}

// ---------------------------------------------------------------------------------

void Gfx_StripRotUp() {
  // rotate the entire contents of the strip from 1 to (NumLEDs - 1)
  int zF = NumLEDs - 1; // set the from pointer
  Block[0] = LED[zF]; // temp copy
  zF--; // adjust the from pointer
  for (int zT = (NumLEDs - 1);zT > 1;zT--) {
    // cycle backwards down the strip
    LED[zT] = LED[zF]; zF--;
  }
  LED[1] = Block[0];  // overwrite with temp copy
}

// ---------------------------------------------------------------------------------

void Gfx_TxtLft() {
  // rotate text strip, 8 pixels deep, 1 pixel to the left
  // copy Colum 0,Y rotate whole panel left, paste column Xmax,Y
  int zX; // column pointer
  int zY; // row pointer
  for (zY = Text_SY;zY <= (Text_SY + 7);zY++) {
    // shift this row left
    MemRGB[0] = LED[Gfx_GetMatL(Xmin,zY)];  // get the left LED
    for (zX = Xmin;zX < Xmax; zX++) {
      // work horizontally along the column
      LED[Gfx_GetMatL(zX,zY)] = LED[Gfx_GetMatL(zX+1,zY)];
    }
    LED[Gfx_GetMatL(Xmax,zY)] = MemRGB[0];  // replace the right LED
  }
}

// ---------------------------------------------------------------------------------
