  //##############################################################################
  //
  //  Panel 48x16 Frame Task Functions
  //
  //##############################################################################


void doFrameTasks() {
  // called every 20ms (variable) to progress task actions
  // tasks functions can be delayed several cycles by setting FrameDel > 0
  // tasks can be blocked until buttons are released by WaitUp == true
  // task variables can be reset to their defaults by setting FrameTask < 0
  // the 2nd core can stop tasks by setting FrameRESET = true
  if (FrameDel > 0) {FrameDel--; FrameBuild = false; return;} // set > 0 to reduce/delay task rate
  if (WaitUp) {if (sw0State == LOW) {return;} else {WaitUp = false;}}
  if (FrameRESET) {setFrameTask(0); FrameRESET = false;}
  
  // if FrameTask is -ve we set variables to their default values before starting it
  if (FrameTask < 0) {setFrameTask(-FrameTask);}
  
  // FrameTasks with low numbers are LED pattern generators.
  // Whilst FrameBuild == true these tasks are called repeatedly. However, once a
  // task has done what it needs to do, it will set FrameBuild = false.
  // If you shorten or extend this list you must change the FrameMax definition.
//  Serial.println("doFrameTasks " + String(FrameTask));
  switch(FrameTask) {
    case  0: FrameBuild = false; break; // do nothing, default state
    case  1: Frame_Default(); break;    // a walking coloured dots
    case  2: Frame_Rainbow(); break;    // a scrolling rainbow
    case  3: Frame_RndChar(); break;    // random fading characters
    case  4: Frame_RndUTC(); break;     // random fading UTC letters rotated
    case  5: Frame_ScrollUTC(); break;  // scrolling left UTC banner
    case  6: Frame_Lines(); break;      // moving random fading lines
    case 999: Test_Task(); break;       // dedbug test task
    default:  // catch-all for FrameTask
      // occurs when FrameTask is out of range, ie. = 99 or doesn't exist
      // clear the LED display and enter the do nothing state
      FastLED.clear(); LEDshow = true; FrameBuild = false; FrameTask = 0;
      break;
  }
  // check SW0 status
  if (sw0State == LOW) {LED[0].setRGB(0,0,12);}
}

// ---------------------------------------------------------------------------------

void Frame_Default() {
  // default frame simply walks a pixel pattern through the two panels
  switch(LED_Task) {
    case 0:
      // initialise task
      FramePeriod = 20; // frame rate 50Hz
      if (FramePeriodDef > 0) {FramePeriod = FramePeriodDef;}
      ColRGB[0].setRGB(255,0,0); Col_Pnt = 0;
      LED_Cnt = 0; LED_Pnt = 0;
      // clear the display and set the 1st pixel
      Bright = BrightDef; FastLED.setBrightness(Bright);
      FastLED.clear(); LED[1] = ColRGB[0];
      LED_Task++; break;
    case 1:
      // cycle through pixels and colours
      Gfx_StripRotUp();
      // increment pointer or reset
      if (LED_Cnt < NumLEDs) {LED_Cnt ++;}
      else {
        LED_Pnt++;
        if (LED_Pnt >= 16) {
          // restarting the frame
          LED_Pnt = 1;
          Col_Pnt++; if (Col_Pnt > 5) {Col_Pnt = 0;}
          switch(Col_Pnt) {
            case 0: ColRGB[0].r = 255; ColRGB[0].g =   0; ColRGB[0].b =   0; break;
            case 1: ColRGB[0].r = 128; ColRGB[0].g = 128; ColRGB[0].b =   0; break;
            case 2: ColRGB[0].r =   0; ColRGB[0].g = 255; ColRGB[0].b =   0; break;
            case 3: ColRGB[0].r =   0; ColRGB[0].g = 128; ColRGB[0].b = 128; break;
            case 4: ColRGB[0].r =   0; ColRGB[0].g =   0; ColRGB[0].b = 255; break;
            case 5: ColRGB[0].r = 128; ColRGB[0].g =   0; ColRGB[0].b = 128; break;
          }
          LED[1] = ColRGB[0];
        }
      }
      LEDshow = true; FrameBuild = false;
      break;
  }
}

// ---------------------------------------------------------------------------------

void Frame_Lines() {
  // moving random fading stars
  switch(LED_Task) {
    case 0:
      // initialise task
      FramePeriod = 40; // frame rate 25Hz
      if (FramePeriodDef > 0) {FramePeriod = FramePeriodDef;}
      Bright = BrightDef; FastLED.setBrightness(Bright);
      FastLED.clear();
      LED_Inc = -1;
      LEDshow = true; FrameBuild = false;
      LED_Task++; break;
    case 1:
      Gfx_RotLft();         // move the whole panel up one pixel
      Gfx_SetCol(12);       // clear colour
      Gfx_Line(Xmax,Ymin,Xmax,Ymax); // line
      Gfx_RndCol(Col11);    // random colour
      Gfx_Line(Xmax,Ymin,Xmax,Ymax); // line
      LEDshow = true; FrameBuild = false;
      if (LED_Inc < 1) if (FramePeriod > 20) {FramePeriod--;} else {LED_Inc = 1;}
      else if (FramePeriod < 100) {FramePeriod++;} else {LED_Inc = -1;}
      LED_Del = 8; LED_Task++; break;
     case 2:
      Gfx_RotLft();         // move the whole panel up one pixel
      Gfx_SetCol(12);       // clear colour
      Gfx_Line(Xmax,Ymin,Xmax,Ymax); // line
      LEDshow = true; FrameBuild = false;
      LED_Del--; if (LED_Del < 1) {LED_Task = 1;} // start over again
      break;
 }
}

// ---------------------------------------------------------------------------------

void Frame_Rainbow() {
  // a scrolling rainbow
  switch(LED_Task) {
    case 0:
      // initialise task
      FramePeriod = 40; // frame rate 25Hz
      if (FramePeriodDef > 0) {FramePeriod = FramePeriodDef;}
      Bright = BrightDef; FastLED.setBrightness(Bright);
      FastLED.clear(); LED_Col = 0;
      LED_Task++; break;
    case 1:
      // scroll all pixels up one row
      Gfx_RotRht();  // move the whole panel right one pixel
      if (LED_Col < 13) {
        // set the left line colour
        Gfx_SetCol(LED_Col);
        Gfx_Line(0,Ymin,0,Ymax);
        LED_Col++;
      } else {LED_Task++; LED_Flag = true;}
      LEDshow = true; FrameBuild = false;
      break;
    case 2:
      // seed random scroll left/right
      LED_Flag = !LED_Flag; LED_Cnt = 4 + random(Xmax - 3);
      if (LED_Flag) {LED_Task = 3;} else {LED_Task = 4;} 
      break;
    case 3:
      Gfx_RotRht();  // move the whole panel up one pixel
      LED_Cnt--; if (LED_Cnt < 1) {LED_Task = 2;}
      LEDshow = true; FrameBuild = false;
      break;
    case 4:
      Gfx_RotLft();  // move the whole panel left one pixel
      LED_Cnt--; if (LED_Cnt < 1) {LED_Task = 2;}
      LEDshow = true; FrameBuild = false;
      break;
  }
}

// ---------------------------------------------------------------------------------

void Frame_RndChar() {
  // random fading characters
  switch(LED_Task) {
    case 0:
      // initialise task
      FramePeriod = 80; // frame rate 12.5Hz
      if (FramePeriodDef > 0) {FramePeriod = FramePeriodDef;}
      Bright = BrightDef; FastLED.setBrightness(Bright);
      FastLED.clear(); LED_Pnt = 0; LED_Cnt = 20 + random(40);
      LED_Task++; break;
    case 1:
      Gfx_DimFrame(1,768,1 + (Col11 / 8));
      Gfx_SetCol(random(12));
      loadChar(65 + random(26));  // get random char A - Z into GC[] array
           if (LED_Pnt == 1) {Gfx_RotGCR();}
      else if (LED_Pnt == 2) {Gfx_RotGCR(); Gfx_RotGCR();}
      else if (LED_Pnt == 3) {Gfx_RotGCL();}
      Gfx_DrawGC(random(Xmax - 4),random(Ymax - 6),0);
      LEDshow = true; FrameBuild = false;
      LED_Cnt--; if (LED_Cnt < 1) {LED_Cnt = 8; LED_Task = 9;} else {LED_Task++;}
      break;
    case 2:
      Gfx_DimFrame(1,768,1 + (Col11 / 8));
      LEDshow = true; FrameBuild = false;
      LED_Task++; break;
    case 3:
      Gfx_DimFrame(1,768,1 + (Col11 / 8));
      LEDshow = true; FrameBuild = false;
      LED_Task = 1; break;
    case 9:
      // change rotation mode and count
      Gfx_DimFrame(1,768,1 + (Col11 / 8));
      LEDshow = true; FrameBuild = false;
      LED_Cnt--; if (LED_Cnt < 1) {
        LED_Last = LED_Pnt;
        while (LED_Last == LED_Pnt) {LED_Pnt = random(4);}
        LED_Cnt = 20 + random(40);
        LED_Task = 1;
      }
      break;
  }
}

// ---------------------------------------------------------------------------------

void Frame_RndUTC(){
  // random fading UTC letters rotated
  uint8_t zX; uint8_t zY;
  switch(LED_Task) {
    case 0:
      // initialise task
      FramePeriod = 40; // frame rate 25Hz
      if (FramePeriodDef > 0) {FramePeriod = FramePeriodDef;}
      Bright = BrightDef; FastLED.setBrightness(Bright);
      FastLED.clear();
      LED_Task++; break;
    case 1:
      Gfx_SetCol(random(12));
      zX = random(Xmax - 14); zY = random(Ymax - 6);
      loadChar('U'); Gfx_DrawGC(zX,zY,0);
      zX+= 6;
      loadChar('T'); Gfx_DrawGC(zX,zY,0);
      zX+= 6;
      loadChar('C'); Gfx_DrawGC(zX,zY,0);
      LEDshow = true; FrameBuild = false;
      LED_Task++; break;
    case 2:
      // loop until dimming has ended
      if (!Gfx_DimFrame(1,768,1 + (Col11 / 8))) {LED_Task = 1;}
      LEDshow = true; FrameBuild = false;
      break;
  }
}

// ---------------------------------------------------------------------------------

void Frame_ScrollUTC() {
  // scrolling left UTC banner
  // text is rotated to scroll up the 32 pixel panel
  switch(LED_Task) {
    case 0:
      // initialise task
      FramePeriod = 75; // frame rate 25Hz
      if (FramePeriodDef > 0) {FramePeriod = FramePeriodDef;}
      Bright = BrightDef; FastLED.setBrightness(Bright);
      FastLED.clear();
      LED_Task++; break;
    case 1:
      LED_Pnt = 0;  // point at first character
      LED_Text = "UTC South Durham";
      LED_Num = LED_Text.length();
      Gfx_RndCol(255);
      LED_Task++; break;
    case 2:
      // get the character from LED_Text
      LED_Char = LED_Text.charAt(LED_Pnt);  // get the character from the string
      loadChar(LED_Char); LED_Cnt = 0;      // copy in charàcter bitmap and set column to 0
      LED_Task++; break;
    case 3:
      // draw each column of the current character
//      t0 = micros();
      Gfx_RotLft();  // move the whole panel left one pixel
//      t1 = micros(); Serial.println(t1 - t0);
//      t0 = micros();
      Gfx_DrawGCC(Xmax,4,LED_Cnt,0); // draw one column of the character
//      t1 = micros(); Serial.println(t1 - t0);
      LED_Cnt++; if (LED_Cnt > 5) {
        // move onto the next character
        if (LED_Pnt < (LED_Num - 1)) {LED_Pnt++; LED_Task = 2;}
        else {LED_Del = 30; LED_Task = 4; Gfx_SetCol(12);}
      }
      LEDshow = true; FrameBuild = false;
      break;
    case 4:
      // end of text string reached, so fill in a blank section
      Gfx_RotLft();          // move the whole panel up one pixel
      Gfx_Line(Xmax,4,Xmax,11); // remove the wrapped character line
      LEDshow = true; FrameBuild = false;
      LED_Del--; if (LED_Del < 1) {LED_Task = 1;} // start over again
      break;
  }
}

// ---------------------------------------------------------------------------------

void Test_Task() {
  // put a test here to run from POST
   return;
  
  Serial.println("Text_Task()");
  Bright = BrightDef; FastLED.setBrightness(Bright);
  FastLED.clear();
  Text = "UTC";
  ColRGB[0].r = Txt_Red; ColRGB[0].g = Txt_Grn; ColRGB[0].b = Txt_Blu;
  Gfx_DrawTxt(0,0);
  LEDshow = true; FrameBuild = false;
}

// ---------------------------------------------------------------------------------
